/*
 *  BasicTermByDocumentMatrix.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.documentSpace;

import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

import org.sesalab.IR.documentSpace.exception.ObjectNotFoundException;
import org.sesalab.IR.documentSpace.storage.IDocumentSpaceStorage;
import org.sesalab.IR.documentSpace.weigher.BasicWeigher;
import org.sesalab.IR.documentSpace.weigher.IWeigherSparseMatrix;
import org.sesalab.IR.matrix.BasicMatrix;
import org.sesalab.IR.matrix.CSCSparseMatrix;
import org.sesalab.IR.matrix.ISparseMatrix;
import org.sesalab.IR.matrix.exception.IncorrectDimensionException;
import org.sesalab.IR.matrix.exception.InvalidIndexException;

/**
 * Fornisce un'implementazione per la matrice termini-documenti
 * 
 * @param <DocumentType>
 *        Tipo di dato utilizzato per modellare la descrizione dei documenti
 *        della collezione
 * @param <TermType>
 *        Tipo di dato utilizzato per modellare la descrizione dei termini della
 *        collezione
 * @author Paola Sgueglia
 */
public class BasicTermByDocumentMatrix<DocumentType, TermType> implements ITermByDocumentMatrix<DocumentType, TermType> {

	/**
	 * Crea una nuova istanza della matrice termini-documenti. L'algoritmo di
	 * pesatura  dato in input.
	 * 
	 * @param pWeigherSparseMatrix
	 *        L'algoritmo di pesatura
	 * @param pDocumentSpaceStorage
	 * 		  La strategia di memorizzazione dello spazio dei documenti   
	 */
	public BasicTermByDocumentMatrix(IWeigherSparseMatrix pWeigherSparseMatrix, IDocumentSpaceStorage<DocumentType, TermType> pDocumentSpaceStorage) {
		this.init(pWeigherSparseMatrix, pDocumentSpaceStorage);
	}

	/**
	 * Crea una nuova istanza della matrice termini-documenti. L'algoritmo di
	 * pesatura  quello di default
	 * @param pDocumentSpaceStorage
	 * 		  La strategia di memorizzazione dello spazio dei documenti  
	 */
	public BasicTermByDocumentMatrix(IDocumentSpaceStorage<DocumentType, TermType> pDocumentSpaceStorage) {
		this.init(new BasicWeigher(), pDocumentSpaceStorage);
	}

	/**
	 * @see org.sesaLab.IR.documentSpace.ITermByDocumentMatrix#addDocument(DocumentType,
	 *      java.util.Map)
	 */
	@SuppressWarnings("javadoc")
	public void addDocument(DocumentType pDocument, Map<TermType, Integer> pTerms) {
		try {
			// Si incrementa il numero di documenti e si aggiunge una nuova
			// colonna alla matrice termini-documenti
			int documentColumn = ++this.documentCount;
			this.matrix.increaseColumns(1);

			// Si aggiunge il documento alla collezione
			this.columnToDocument.put(new Integer(documentColumn), pDocument);
			this.documentToColumn.put(pDocument, new Integer(documentColumn));
			
			// Si aggiungono i termini alla collezione
			Iterator<TermType> termsIterator = pTerms.keySet().iterator();
			while (termsIterator.hasNext()) {
				TermType term = termsIterator.next();
				Integer occurrences = pTerms.get(term);
				Integer row = this.termToRow.get(term);
				int termRow;
				if (row != null)
					// Il termine  gia presente
					termRow = row.intValue();
				else {
					// E' stato trovato un nuovo termine
					// Si incrementa il numero di termini e si aggiunge una
					// nuova riga alla matrice termini-documenti
					termRow = ++this.termCount;
					this.matrix.increaseRows(1);
					// Si aggiunge il termine alla collezione
					this.termToRow.put(term, new Integer(termRow));
					this.rowToTerm.put(new Integer(termRow), term);
				}

				// Si aggiorna la matrice termini-documenti
				this.matrix.set(termRow, documentColumn, occurrences.doubleValue());
			}
		} catch (IncorrectDimensionException e) {
			// Eccezione che non si pu verificare
		} catch (InvalidIndexException e) {
			// Eccezione che non si pu verificare
		}
	}

	/**
	 * @see org.sesaLab.IR.documentSpace.ITermByDocumentMatrix#contains(DocumentType)
	 */
	@SuppressWarnings("javadoc")
	public boolean contains(DocumentType pDocument) {
		return this.columnToDocument.containsValue(pDocument);
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#containsAll(java.util.Collection)
	 */
	public boolean containsAll(Collection<DocumentType> pDocuments) {
		return this.documentToColumn.keySet().containsAll(pDocuments);
	}

	/**
	 * @see org.sesaLab.IR.documentSpace.ITermByDocumentMatrix#getColumn(DocumentType)
	 */
	@SuppressWarnings("javadoc")
	public int getColumn(DocumentType pDocument) throws ObjectNotFoundException {
		Integer column = this.documentToColumn.get(pDocument);
		if (column == null)
			throw new ObjectNotFoundException("Document not found");
		return column.intValue();
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#getDocument(int)
	 */
	public DocumentType getDocument(int pColumn) {
		return this.columnToDocument.get(pColumn);
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#getDocumentCount()
	 */
	public int getDocumentCount() {
		return this.documentCount;
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#getIndexedDocuments()
	 */
	public Collection<DocumentType> getIndexedDocuments() {
		return this.documentToColumn.keySet();
	}

	/**
	 * @see org.sesaLab.IR.documentSpace.ITermByDocumentMatrix#getRow(TermType)
	 */
	@SuppressWarnings("javadoc")
	public int getRow(TermType pTerm) throws ObjectNotFoundException {
		Integer row = this.termToRow.get(pTerm);
		if (row == null)
			throw new ObjectNotFoundException("Term not found");
		return row.intValue();
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#getTerm(int)
	 */
	public TermType getTerm(int pRow) {
		return this.rowToTerm.get(pRow);
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#getTermCount()
	 */
	public int getTermCount() {
		return this.termCount;
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#getWeigher()
	 */
	public IWeigherSparseMatrix getWeigher() {
		return this.weigherMatrix;
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#getWeightedMatrix()
	 */
	public ISparseMatrix getWeightedMatrix() {
		return this.weigherMatrix.weigh(this.matrix);
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#isEmpty()
	 */
	public boolean isEmpty() {
		return this.columnToDocument.isEmpty();
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#load()
	 */
	public void load() throws IOException, ClassNotFoundException {
		// Caricamento matrice
		this.matrix = this.documentSpaceStorage.loadMatrix();

		// Caricamento termini
		this.rowToTerm = this.documentSpaceStorage.loadTerms();
		this.termToRow.clear();
		Iterator<Integer> rowIterator = this.rowToTerm.keySet().iterator();
		while (rowIterator.hasNext()) {
			Integer row = rowIterator.next();
			TermType term = this.rowToTerm.get(row);
			this.termToRow.put(term, row);
		}
		this.termCount = this.termToRow.size();
		
		// Caricamento documenti
		this.columnToDocument = this.documentSpaceStorage.loadDocuments();
		this.documentToColumn.clear();
		Iterator<Integer> columnIterator = this.columnToDocument.keySet().iterator();
		while (columnIterator.hasNext()) {
			Integer column = columnIterator.next();
			DocumentType document = this.columnToDocument.get(column);
			this.documentToColumn.put(document, column);
		}
		this.documentCount = this.columnToDocument.size();
	}

	/**
	 * @see org.sesaLab.IR.documentSpace.ITermByDocumentMatrix#removeDocument(DocumentType)
	 */
	@SuppressWarnings("javadoc")
	public void removeDocument(DocumentType pDocument) throws ObjectNotFoundException {
		Integer column = this.documentToColumn.get(pDocument);
		if (column == null)
			throw new ObjectNotFoundException();

		// Si rimuove il documento dalla collezione
		this.columnToDocument.remove(column);
		this.documentToColumn.remove(pDocument);

		int documentColumn = column.intValue();
		try {
			this.matrix.removeColumn(documentColumn);
		} catch (InvalidIndexException e) {
			// Eccezione che non si pu verificare
		}

		this.fillDocumentBlanks(documentColumn);
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#save()
	 */
	public void save() throws IOException {
		this.documentSpaceStorage.saveMatrix(this.matrix);
		this.documentSpaceStorage.saveTerms(this.rowToTerm);
		this.documentSpaceStorage.saveDocuments(this.columnToDocument);
	}

	/**
	 * @see org.sesalab.IR.documentSpace.ITermByDocumentMatrix#setWeigher(org.sesalab.IR.documentSpace.weigher.IWeigherSparseMatrix)
	 */
	public void setWeigher(IWeigherSparseMatrix pWeigherMatrix) {
		this.weigherMatrix = pWeigherMatrix;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		String strMatrix = "";
		if (this.isEmpty())
			return (strMatrix);
		BasicMatrix matrix = null;
		try {
			matrix = new BasicMatrix(this.matrix.getRowCount(), this.matrix.getColumnCount());
			for (int i = 0; i < this.matrix.getNotNullValues().length; i++)
				matrix.set(this.matrix.getNotNullRowIndices()[i], this.matrix.getNotNullColumnIndices()[i], this.matrix.getNotNullValues()[i]);
		} catch (IncorrectDimensionException e) {
			// Eccezione che non si pu generare
		} catch (InvalidIndexException e) {
			// Eccezione che non si pu generare
		}
		Iterator<DocumentType> documentIterator = this.columnToDocument.values().iterator();
		strMatrix = "\t\t";
		while (documentIterator.hasNext()) {
			String aux = documentIterator.next().toString();
			strMatrix += aux + "\t\t";
		}
		strMatrix += "\n";
		Object[] terms = this.rowToTerm.values().toArray();
		for (int i = 1; i <= matrix.getRowCount(); i++) {
			String aux = terms[i - 1].toString();
			if (aux.length() <= 5)
				strMatrix += aux + "\t\t";
			else
				strMatrix += aux + "\t";
			for (int j = 1; j <= matrix.getColumnCount(); j++)
				try {
					strMatrix += matrix.elementAt(i, j) + "\t\t";
				} catch (InvalidIndexException e) {
					// Eccezione che non si pu mai generare
				}
			strMatrix += "\n";
		}
		return strMatrix;
	}

	/*
	 * Elimina eventuali buchi causati dall'eliminazione di un documento 
	 * @param pDocumentColumn 
	 * 		  La colonna della matrice termini-documenti relativa al documento 
	 * 		  eliminato
	 */
	private void fillDocumentBlanks(int pDocumentColumn) {
		Map<Integer, DocumentType> newColumnToDocument = new TreeMap<Integer, DocumentType>();
		Map<DocumentType, Integer> newDocumentToColumn = new TreeMap<DocumentType, Integer>();

		Iterator<Integer> columnIterator = this.columnToDocument.keySet().iterator();

		while (columnIterator.hasNext()) {
			Integer column = columnIterator.next();
			DocumentType document = this.columnToDocument.get(column);
			int documentColumn = column.intValue();
			if (documentColumn > pDocumentColumn)
				column = new Integer(column.intValue() - 1);
			newColumnToDocument.put(column, document);
			newDocumentToColumn.put(document, column);
		}

		this.documentToColumn = newDocumentToColumn;
		this.columnToDocument = newColumnToDocument;

		int i = 1;
		while (i <= this.matrix.getRowCount()) {
			if (this.matrix.isNullRow(i)) {
				try {
					this.matrix.removeRow(i);
				} catch (InvalidIndexException e) {
					// Eccezione che non si verifica mai
				}
				this.rowToTerm.remove(new Integer(i));
				this.fillTermBlanks(i);
			} else {
				i++;
			}
		}
		this.documentCount--;
	}

	/*
	 * Elimina eventuali buchi causati dall'eliminazione di un termine 
	 * @param pRowTerm 
	 * 	      La riga della matrice termini-documenti relativa al termine da
	 		  eliminare
	 */
	private void fillTermBlanks(int pRowTerm) {
		Map<Integer, TermType> newRowToTerm = new TreeMap<Integer, TermType>();
		Map<TermType, Integer> newTermToRow = new TreeMap<TermType, Integer>();

		Iterator<Integer> rowIterator = this.rowToTerm.keySet().iterator();

		while (rowIterator.hasNext()) {
			Integer row = rowIterator.next();
			TermType term = this.rowToTerm.get(row);
			int rowTerm = row.intValue();
			if (rowTerm > pRowTerm)
				row = new Integer(row.intValue() - 1);
			newRowToTerm.put(row, term);
			newTermToRow.put(term, row);
		}

		this.termToRow = newTermToRow;
		this.rowToTerm = newRowToTerm;
		
		this.termCount--;
	}

	/*
	 * Inizializza le strutture dati utilizzate per gestire la matrice
	 * termini-documenti 
	 * @param pWeigherSparseMatrix 
	 *        L'algoritmo di pesatura da utilizzare per pesare la matrice
	 * @param pDocumentSpaceStorage
	 * 		  La strategia di memorizzazione dello spazio dei documenti   
	 */
	private void init(IWeigherSparseMatrix pWeigherSparseMatrix, IDocumentSpaceStorage<DocumentType, TermType> pDocumentSpaceStorage) {
		this.matrix = new CSCSparseMatrix();
		this.weigherMatrix = pWeigherSparseMatrix;
		this.documentSpaceStorage = pDocumentSpaceStorage;
		this.columnToDocument = new TreeMap<Integer, DocumentType>();
		this.documentToColumn = new TreeMap<DocumentType, Integer>();
		this.rowToTerm = new TreeMap<Integer, TermType>();
		this.termToRow = new TreeMap<TermType, Integer>();
		this.documentCount = 0;
		this.termCount = 0;
	}

	// Mappa utilizzata per ottenere data una colonna il relativo documento
	private Map<Integer, DocumentType> columnToDocument;

	// Numero documenti della collezione
	private int documentCount;

	// Utilizzato per memorizzare la matrice termini-documenti
	private IDocumentSpaceStorage<DocumentType, TermType> documentSpaceStorage;

	// Mappa utilizzata per ottenere dato un documento la relativa colonna
	private Map<DocumentType, Integer> documentToColumn;

	// Matrice sparsa utilizzata per modellare la matrice termini documenti
	private ISparseMatrix matrix;

	// Mappa utilizzata per ottenere data una riga il relativo termine
	private Map<Integer, TermType> rowToTerm;

	// Numero termini della collezione
	private int termCount;

	// Mappa utilizzata per ottenere dato un termine il relativo indice di riga
	private Map<TermType, Integer> termToRow;

	// Utilizzato per pesare la matrice termini-documenti
	private IWeigherSparseMatrix weigherMatrix;

	/**
	 * @return La strategia di memorizzazione della matrice termini-documenti
	 */
	public IDocumentSpaceStorage<DocumentType, TermType> getDocumentSpaceStorage() {
		return this.documentSpaceStorage;
	}

	/**
	 * Setta la strategia di memorizzazione della matrice termini-documenti
	 * @param pDocumentSpaceStorage La strategia di memorizzazione della matrice termini-documenti
	 */
	public void setDocumentSpaceStorage(
			IDocumentSpaceStorage<DocumentType, TermType> pDocumentSpaceStorage) {
		this.documentSpaceStorage = pDocumentSpaceStorage;
	}
}
