/*
 *  ITermByDocumentMatrix.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.documentSpace;

import java.io.IOException;
import java.util.Collection;
import java.util.Map;

import org.sesalab.IR.documentSpace.exception.ObjectNotFoundException;
import org.sesalab.IR.documentSpace.weigher.IWeigherSparseMatrix;
import org.sesalab.IR.matrix.ISparseMatrix;

/**
 * Definisce l'interfaccia della matrice termini-documenti
 * 
 * @param <DocumentType>
 *        Tipo di dato utilizzato per modellare la descrizione dei documenti
 *        della collezione
 * @param <TermType>
 *        Tipo di dato utilizzato per modellare la descrizione dei termini della
 *        collezione
 * @author Paola Sgueglia
 */
public interface ITermByDocumentMatrix<DocumentType, TermType> {

	/**
	 * Consente di aggiungere un documento alla matrice termini-documenti
	 * 
	 * @param pDocument
	 *        La descrizione del documento da aggiungere
	 * @param pTerms
	 *        Termini, con relative occorrenze, presenti all'interno del
	 *        documento
	 */
	public void addDocument(DocumentType pDocument, Map<TermType, Integer> pTerms);

	/**
	 * Verifica se un documento  presente all'interno della matrice
	 * termini-documenti
	 * 
	 * @param pDocument
	 *        Il documento da ricercare
	 * @return <code>true</code> nel caso in cui il documento  presente,
	 *         <code>false</code> altrimenti
	 */
	public boolean contains(DocumentType pDocument);

	/**
	 * Verifica se una collezione di documenti  presente all'interno della
	 * matrice termini-documenti
	 * 
	 * @param pDocuments
	 *        La collezione di documenti da ricercare
	 * @return <code>true</code> nel caso in cui tutti i documenti sono
	 *         presenti, <code>false</code> altrimenti
	 */
	public boolean containsAll(Collection<DocumentType> pDocuments);

	/**
	 * Consente di ottenere la colonna relativa al documento pDocument
	 * 
	 * @param pDocument
	 *        Il documento di cui reperire la colonna
	 * @return La colonna relativa al documento pDocument
	 * @throws ObjectNotFoundException
	 *         Generata quando si tenta di accedere ad un documento non presente
	 *         nella collezione
	 */
	public int getColumn(DocumentType pDocument) throws ObjectNotFoundException;

	/**
	 * Consente di ottenere la descrizione del pColumn documento della
	 * collezione
	 * 
	 * @param pColumn
	 *        La colonna della matrice termini-documenti cui il documento fa
	 *        riferimento
	 * @return La descrizione del documento
	 */
	public DocumentType getDocument(int pColumn);

	/**
	 * Restituisce il numero di documenti presenti nella collezione
	 * 
	 * @return Il numero di documenti presenti nella collezione
	 */
	public int getDocumentCount();

	/**
	 * @return La lista dei documenti indicizzati 
	 */
	public Collection<DocumentType> getIndexedDocuments();

	/**
	 * Consente di ottenere la riga relativa al termine pTerm
	 * 
	 * @param pTerm
	 *        Il termine di cui reperire la riga
	 * @return La riga relativa al termine pTerm
	 * @throws ObjectNotFoundException
	 *         Generata quando si tenta di accedere ad un termine non presente
	 *         nella collezione
	 */
	public int getRow(TermType pTerm) throws ObjectNotFoundException;

	/**
	 * Consente di ottenere il termini p-Row della collezione
	 * 
	 * @param pRow
	 *        La riga della matrice termini-documenti cui il termine fa
	 *        riferimento
	 * @return Il termine della collezione
	 */
	public TermType getTerm(int pRow);

	/**
	 * Restituisce il numero di termini presenti nella collezione
	 * 
	 * @return Il numero di termini presenti nella collezione
	 */
	public int getTermCount();

	/**
	 * Resituisce la classe che implementa l'algoritmo di pesatura della matrice
	 * 
	 * @return La classe che implementa l'algoritmo di pesatura della matrice
	 */
	public IWeigherSparseMatrix getWeigher();

	/**
	 * Restituisce la matrice dei termini-documenti pesata
	 * 
	 * @return La matrice dei termini-documenti pesata
	 */
	public ISparseMatrix getWeightedMatrix();

	/**
	 * Verifica se la matrice termini-documenti  vuota
	 * 
	 * @return <code>true</code> se la matrice  vuota, <code>false</code>
	 *         altrimenti
	 */
	public boolean isEmpty();

	/**
	 * Carica la matrice termini-documenti
	 * 
	 * @throws ClassNotFoundException
	 * @throws IOException
	 */
	public void load() throws IOException, ClassNotFoundException;

	/**
	 * Consente di rimuovere un documento dalla matrice termini-documenti
	 * 
	 * @param pDocument
	 *        La descrizione del documento da rimuovere
	 * @throws ObjectNotFoundException
	 *         Generata quando si tenta di rimuovere un documento non presente
	 *         nella collezione
	 */
	public void removeDocument(DocumentType pDocument) throws ObjectNotFoundException;

	/**
	 * Salva la matrice termini-documenti
	 * 
	 * @throws IOException
	 */
	public void save() throws IOException;
	
	/**
	 * Setta la classe che implementa l'algoritmo di pesatura della matrice
	 * 
	 * @param pWeigherMatrix
	 *        La classe che implementa l'algoritmo di pesatura della matrice
	 */
	public void setWeigher(IWeigherSparseMatrix pWeigherMatrix);
}
