/*
 *  LogEntropyWeigher.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.documentSpace.weigher;

import org.sesalab.IR.matrix.CSCSparseMatrix;
import org.sesalab.IR.matrix.ISparseMatrix;
import org.sesalab.IR.matrix.exception.InvalidIndexException;

/**
 * Classe che implementa l'algoritmo di pesatura della matrice termini-documenti
 * denominata LogEntropy
 * 
 * @author Paola Sgueglia
 */
public class LogEntropyWeigher implements IWeigherSparseMatrix {

	/**
	 * Crea una nuova istanza della classe
	 */
	public LogEntropyWeigher() {
		super();
	}

	/**
	 * @see org.sesalab.IR.documentSpace.weigher.IWeigherSparseMatrix#weigh(org.sesalab.IR.matrix.ISparseMatrix)
	 */
	public ISparseMatrix weigh(ISparseMatrix pSparseMatrix) {
		// Variabili di appoggio
		int index;
		double logLocalArgument, logGlobalArgument, den, num;

		// Dichiarazione dei vettore necessari per il calcolo dei pesi
		double[] globalFrequency, relativeFrequency, entropy;

		// Inizializzazione
		int rows[] = pSparseMatrix.getNotNullRowIndices();
		int columns[] = pSparseMatrix.getNotNullColumnIndices();
		double values[] = pSparseMatrix.getNotNullValues();
		double weightedValues[] = new double[values.length];

		// Inizializzazione del vettore contenente la frequenza globale dei
		// termini
		globalFrequency = new double[rows.length];
		for (index = 0; index < rows.length; index++)
			globalFrequency[index] = 0;
		// Calcolo della frequenza globale dei termini
		for (index = 0; index < values.length; index++) {
			globalFrequency[rows[index] - 1] += values[index];
		}

		// Inizializzazione del vettore contenente le frequenze relative
		relativeFrequency = new double[values.length];
		for (index = 0; index < values.length; index++)
			relativeFrequency[index] = 0;
		// Calcolo della frequenza relativa dei termini
		for (index = 0; index < values.length; index++)
			relativeFrequency[index] = values[index] / globalFrequency[rows[index] - 1];

		// Inizializzazione del vettore contenente l'entropia di ogni termine
		entropy = new double[rows.length];
		for (index = 0; index < rows.length; index++)
			entropy[index] = 0;
		den = Math.log(columns.length);
		
		// Calcolo della frequenza relativa dei termini
		for (index = 0; index < values.length; index++) {
			num = relativeFrequency[index] * Math.log(relativeFrequency[index]);
			entropy[rows[index] - 1] += num / den;
		}

		// Aggiornamento del vettore delle entropie
		for (index = 0; index < rows.length; index++)
			entropy[index] = 1 - entropy[index];

		// Definizione del peso dei termini
		for (index = 0; index < values.length; index++) {
			logLocalArgument = (values[index] + 1);
			logGlobalArgument = (entropy[rows[index] - 1]);
			weightedValues[index] = Math.log(logLocalArgument * logGlobalArgument);
		}

		// Creazione matrice sparsa
		ISparseMatrix weightedMatrix = new CSCSparseMatrix();
		((CSCSparseMatrix) weightedMatrix).init(pSparseMatrix.getRowCount(), pSparseMatrix.getColumnCount());
		for (index = 0; index < values.length; index++) {
			try {
				weightedMatrix.set(rows[index], columns[index], weightedValues[index]);
			} catch (InvalidIndexException e) {
				// Eccezione che non si pu mai verificare
			}
		}
		return weightedMatrix;
	}
}