/*
 *  BasicMatrix.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.matrix;

import org.sesalab.IR.matrix.exception.IncorrectDimensionException;
import org.sesalab.IR.matrix.exception.InvalidIndexException;

/**
 * Fornisce l'implementazione di default per una matrice
 * 
 * @author Paola Sgueglia
 */
public class BasicMatrix implements IMatrix {
	private static final long serialVersionUID = -8636786777075686607L;

	/**
	 * Costruisce una matrice di vuota
	 * 
	 * @param pRows
	 *        Il numero di righe della matrice
	 * @param pColumns
	 *        Il numero di colonne
	 * @throws IncorrectDimensionException
	 *         Generata quando il numero di righe o di colonne non  valido
	 */
	public BasicMatrix(int pRows, int pColumns) throws IncorrectDimensionException {
		if (pRows <= 0)
			throw new IncorrectDimensionException("Invalid row number");
		if (pColumns <= 0)
			throw new IncorrectDimensionException("Invalid column number");
		// Creazione e inizializzazione della matrice
		this.matrix = new double[pRows][pColumns];
		for (int i = 0; i < pRows; i++)
			for (int j = 0; j < pColumns; j++)
				this.matrix[i][j] = 0;
		this.rows = pRows;
		this.columns = pColumns;
	}

	/**
	 * Costruisce una matrice a partire da un array bi-dimensionale
	 * 
	 * @param pRows
	 *        Il numero di righe della matrice
	 * @param pColumns
	 *        Il numero di colonne
	 * @param pArrayMatrix
	 *        L'array bi-dimensionale contenente la matrice
	 * @throws IncorrectDimensionException
	 *         Generata quando il numero di righe o di colonne non  valido
	 */
	public BasicMatrix(int pRows, int pColumns, double[][] pArrayMatrix) throws IncorrectDimensionException {
		if (pRows <= 0)
			throw new IncorrectDimensionException("Invalid row number");
		if (pColumns <= 0)
			throw new IncorrectDimensionException("Invalid column number");

		// Creazione e inizializzazione della matrice
		this.matrix = pArrayMatrix;
		this.rows = pRows;
		this.columns = pColumns;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#elementAt(int, int)
	 */
	public double elementAt(int pRowIndex, int pColumnIndex) throws InvalidIndexException {
		this.checkRowIndex(pRowIndex);
		this.checkColumnIndex(pColumnIndex);
		return this.matrix[pRowIndex - 1][pColumnIndex - 1];
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#getColumn(int)
	 */
	public double[] getColumn(int pColumnIndex) throws InvalidIndexException {
		this.checkColumnIndex(pColumnIndex);
		double column[] = new double[this.rows];
		for (int i = 0; i < this.rows; i++)
			column[i] = this.matrix[i][pColumnIndex - 1];
		return column;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#getColumnCount()
	 */
	public int getColumnCount() {
		return this.columns;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#getRow(int)
	 */
	public double[] getRow(int pRowIndex) throws InvalidIndexException {
		this.checkRowIndex(pRowIndex);
		double row[] = new double[this.columns];
		for (int i = 0; i < this.columns; i++)
			row[i] = this.matrix[pRowIndex - 1][i];
		return row;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#getRowCount()
	 */
	public int getRowCount() {
		return this.rows;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#increaseColumns(int)
	 */
	public void increaseColumns(int pAddedColumns) throws IncorrectDimensionException {
		if (pAddedColumns < 0)
			throw new IllegalArgumentException();
		int newColumnNumber = this.columns + pAddedColumns;
		double newMatrix[][] = new double[this.rows][newColumnNumber];
		for (int i = 0; i < this.rows; i++)
			for (int j = 0; j < this.columns; j++)
				newMatrix[i][j] = this.matrix[i][j];
		for (int i = 0; i < this.rows; i++)
			for (int j = this.columns; j < newColumnNumber; j++)
				newMatrix[i][j] = 0;
		this.matrix = newMatrix;
		this.columns = newColumnNumber;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#increaseRows(int)
	 */
	public void increaseRows(int pAddedRows) throws IncorrectDimensionException {
		if (pAddedRows < 0)
			throw new IncorrectDimensionException();
		int newRowNumber = this.rows + pAddedRows;
		double newMatrix[][] = new double[newRowNumber][this.columns];
		for (int i = 0; i < this.rows; i++)
			for (int j = 0; j < this.columns; j++)
				newMatrix[i][j] = this.matrix[i][j];
		for (int i = this.rows; i < newRowNumber; i++)
			for (int j = 0; j < this.columns; j++)
				newMatrix[i][j] = 0;
		this.matrix = newMatrix;
		this.rows = newRowNumber;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#mul(org.sesalab.IR.matrix.IMatrix)
	 */
	public IMatrix mul(IMatrix pToMultiply) throws IncorrectDimensionException {
		if (this.columns != pToMultiply.getRowCount())
			throw new IncorrectDimensionException("Inner matrix dimensions must agree");

		IMatrix result = new BasicMatrix(this.rows, pToMultiply.getColumnCount());

		try {
			for (int i = 1; i <= this.rows; i++)
				for (int j = 1; j <= pToMultiply.getColumnCount(); j++) {
					double val = 0;
					for (int k = 1; k <= this.columns; k++)
						val += this.elementAt(i, k) * pToMultiply.elementAt(k, j);
					result.set(i, j, val);
				}
		} catch (InvalidIndexException e) {
			// Eccezione che non si verifica mai
		}
		return (result);
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#removeColumn(int)
	 */
	public void removeColumn(int pColumn) throws InvalidIndexException {
		this.checkColumnIndex(pColumn);
		double newMatrix[][] = new double[this.rows][this.columns - 1];
		for (int i = 0; i < this.rows; i++)
			for (int j = 0; j < pColumn - 1; j++)
				newMatrix[i][j] = this.matrix[i][j];
		for (int i = 0; i < this.rows; i++)
			for (int j = pColumn; j < this.columns; j++)
				newMatrix[i][j - 1] = this.matrix[i][j];
		this.matrix = newMatrix;
		this.columns--;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#removeRow(int)
	 */
	public void removeRow(int pRow) throws InvalidIndexException {
		this.checkRowIndex(pRow);
		double newMatrix[][] = new double[this.rows - 1][this.columns];
		for (int i = 0; i < pRow - 1; i++)
			for (int j = 0; j < this.columns; j++)
				newMatrix[i][j] = this.matrix[i][j];
		for (int i = pRow; i < this.rows; i++)
			for (int j = 0; j < this.columns; j++)
				newMatrix[i - 1][j] = this.matrix[i][j];
		this.matrix = newMatrix;
		this.rows--;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#set(int, int, double)
	 */
	public void set(int pRowIndex, int pColumnIndex, double pElement) throws InvalidIndexException {
		this.checkRowIndex(pRowIndex);
		this.checkColumnIndex(pColumnIndex);
		this.matrix[pRowIndex - 1][pColumnIndex - 1] = pElement;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#subMatrix(int, int)
	 */
	public IMatrix subMatrix(int pRows, int pColumns) throws InvalidIndexException {
		this.checkRowIndex(pRows);
		this.checkColumnIndex(pColumns);
		BasicMatrix newMatrix = null;
		try {
			newMatrix = new BasicMatrix(pRows, pColumns);
			for (int i = 0; i < pRows; i++)
				for (int j = 0; j < pColumns; j++)
					newMatrix.matrix[i][j] = this.matrix[i][j];
		} catch (IncorrectDimensionException e) {
			// Eccezione che non si pu verificares
		}
		return newMatrix;
	}
	
	/**
	 * @see org.sesalab.IR.matrix.IMatrix#toArray()
	 */
	public double[][] toArray() {
		return this.matrix.clone();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		String strMatrix = "";
		for (int i = 0; i < this.rows; i++) {
			for (int j = 0; j < this.columns; j++)
				strMatrix += this.matrix[i][j] + "\t";
			strMatrix += "\n";
		}
		return strMatrix;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#transpose()
	 */
	public IMatrix transpose() {
		IMatrix result = null;
		try {
			result = new BasicMatrix(this.columns, this.rows);
			for (int i = 1; i <= this.rows; i++)
				for (int j = 1; j <= this.columns; j++) {
					result.set(j, i, this.elementAt(i, j));
				}
		} catch (IncorrectDimensionException e1) {
			// Eccezione che non si verifica mai
		} catch (InvalidIndexException e) {
			// Eccezione che non si verifica mai
		}
		return (result);
	}

	/*
	 * Verifica se l'indice di colonna  valido @param pColumnIndex L'indice di
	 * colonna @throws InvalidIndexException Generata quando l'indice non 
	 * valido
	 */
	private void checkColumnIndex(int pColumnIndex) throws InvalidIndexException {
		if ((pColumnIndex <= 0) || (pColumnIndex > (this.columns + 1)))
			throw new InvalidIndexException("Invalid column index");
	}

	/*
	 * Verifica se l'indice di riga  valido @param pRowIndex L'indice di riga
	 * @throws InvalidIndexException Generata quando l'indice non  valido
	 */
	private void checkRowIndex(int pRowIndex) throws InvalidIndexException {
		if ((pRowIndex <= 0) || (pRowIndex > (this.rows + 1)))
			throw new InvalidIndexException("Invalid row index");
	}

	// Il numero di colonne
	int columns;

	// Il numero di righe
	int rows;

	// La matrice rappresentata attraverso un array bi-dimensionale
	private double matrix[][];
}