/*
 *  CSCSparseMatrix.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.matrix;

import java.util.ArrayList;

import org.sesalab.IR.matrix.exception.IncorrectDimensionException;
import org.sesalab.IR.matrix.exception.InvalidIndexException;

/**
 * Fornisce un'implementazione di una matrice sparsa secondo il formato
 * Compressed Sparse Column (CSC)
 * 
 * @author Paola Sgueglia
 */
public class CSCSparseMatrix implements ISparseMatrix {
	// Usata per etichettare un elemento come ultimo elemento della riga e della
	// colonna
	private static int LAST_IN_BOTH_DIRECTIONS = 0;

	// Usata per etichettare un elemento come ultimo elemento della colonna
	private static int LAST_IN_COLUMN = 2;

	// Usata per etichettare un elemento come ultimo elemento della riga
	private static int LAST_IN_ROW = 1;

	private static final long serialVersionUID = 4337893516801773022L;

	/**
	 * Crea una nuova istanza della classe
	 */
	public CSCSparseMatrix() {
		this.init(0, 0);
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#elementAt(int, int)
	 */
	public double elementAt(int pRow, int pColumn) throws InvalidIndexException {
		this.checkRowIndex(pRow);
		this.checkColumnIndex(pColumn);
		return this.getElement(pRow - 1, pColumn - 1);
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#getColumn(int)
	 */
	public double[] getColumn(int pColumn) throws InvalidIndexException {
		this.checkColumnIndex(pColumn);
		double[] column = new double[this.rows];
		for (int i = this.columnPointers[pColumn - 1]; i < this.columnPointers[pColumn]; i++)
			column[this.rowIndices[i]] = this.values[i];
		return (column);
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#getColumnCount()
	 */
	public int getColumnCount() {
		return this.columns;
	}

	/**
	 * @return Il vettore contenente gli indici di inizio delle colonne
	 */
	public int[] getColumnPointers() {
		return this.columnPointers;
	}

	/**
	 * @return Il vettore contenente gli indici delle colonne
	 */
	public int getColumns() {
		return this.columns;
	}

	/**
	 * @see org.sesalab.IR.matrix.ISparseMatrix#getNotNullColumnIndices()
	 */
	public int[] getNotNullColumnIndices() {
		int[] nonZeroColumnIndexes = new int[this.rowIndices.length];
		int k = 0;
		for (int i = 0; i < this.columnPointers.length - 1; i++) {
			int columnElements = (this.columnPointers[i + 1] - this.columnPointers[i]);
			while (columnElements > 0) {
				nonZeroColumnIndexes[k] = i + 1;
				k++;
				columnElements--;
			}
		}
		return nonZeroColumnIndexes;
	}

	/**
	 * @see org.sesalab.IR.matrix.ISparseMatrix#getNotNullRowIndices()
	 */
	public int[] getNotNullRowIndices() {
		int[] auxRowIndices = this.rowIndices.clone();
		for (int i = 0; i < auxRowIndices.length; i++)
			auxRowIndices[i]++;
		return auxRowIndices;
	}

	/**
	 * @see org.sesalab.IR.matrix.ISparseMatrix#getNotNullValues()
	 */
	public double[] getNotNullValues() {
		return this.values;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#getRow(int)
	 */
	public double[] getRow(int pRow) {
		BasicMatrix matrix = null;
		try {
			matrix = new BasicMatrix(this.rows, this.columns);
			for (int i = 0; i < this.columnPointers.length - 1; i++)
				for (int j = this.columnPointers[i]; j < this.columnPointers[i + 1]; j++) {
					matrix.set(this.rowIndices[i] + 1, i + 1, this.values[i]);
				}
			return matrix.getRow(pRow);
		} catch (IncorrectDimensionException e) {
			// Eccezione che non si pu generare
		} catch (InvalidIndexException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#getRowCount()
	 */
	public int getRowCount() {
		return this.rows;
	}

	/**
	 * @return Il vettore degli indici delle righe
	 */
	public int[] getRowIndices() {
		return this.rowIndices;
	}

	/**
	 * @return Le righe della matrice
	 */
	public int getRows() {
		return this.rows;
	}

	/**
	 * @return Il vettore dei valori non nulli
	 */
	public double[] getValues() {
		return this.values;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#increaseColumns(int)
	 */
	public void increaseColumns(int pAddedColumns) {
		int[] aux = new int[this.columns + pAddedColumns + 1];
		int i = 0;
		for (i = 0; i <= this.columns; i++) {
			aux[i] = this.columnPointers[i];
		}
		for (; i < aux.length; i++)
			aux[i] = this.columnPointers[this.columns];

		this.columnPointers = aux;
		this.columns += pAddedColumns;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#increaseRows(int)
	 */
	public void increaseRows(int pAddedRows) {
		this.rows += pAddedRows;
	}

	/**
	 * Inizializza le strutture dati
	 * 
	 * @param pValues
	 *            Valori non nulli
	 * @param pRowIndices
	 *            Indici di riga degli elementi non nulli
	 * @param pColumnPointers
	 *            Indici di inizio di ciascuna colonna in values e rowIndices
	 * @param pRows
	 *            Numero di righe
	 * @param pColumns
	 *            Numero di colonne
	 */
	public void init(double[] pValues, int[] pRowIndices,
			int[] pColumnPointers, int pRows, int pColumns) {
		this.values = pValues;
		this.rowIndices = pRowIndices;
		this.columnPointers = pColumnPointers;
		this.rows = pRows;
		this.columns = pColumns;
	}

	/**
	 * Inizializza le strutture dati
	 * 
	 * @param pRows
	 *            Numero di righe
	 * @param pColumns
	 *            Numero di colonne
	 */
	public void init(int pRows, int pColumns) {
		this.init(new double[0], new int[0], new int[pColumns + 1], pRows, pColumns);
	}

	/**
	 * @see org.sesalab.IR.matrix.ISparseMatrix#isNullColumn(int)
	 */
	public boolean isNullColumn(int pColumn) {
		int columnItems = this.columnPointers[pColumn]
				- this.columnPointers[pColumn - 1];
		if (columnItems != 0)
			return false;
		return true;
	}

	/**
	 * @see org.sesalab.IR.matrix.ISparseMatrix#isNullRow(int)
	 */
	public boolean isNullRow(int pRow) {
		int i;
		for (i = 0; i < this.rowIndices.length; i++) {
			if (this.rowIndices[i] == (pRow - 1))
				break;
		}
		if (i < this.rowIndices.length)
			return false;
		return true;
	}

	/**
	 * Not implemented method
	 * 
	 * @see org.sesalab.IR.matrix.IMatrix#mul(org.sesalab.IR.matrix.IMatrix)
	 */
	public IMatrix mul(IMatrix pToMultiply) throws IncorrectDimensionException {
		return null;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#removeColumn(int)
	 */
	public void removeColumn(int pColumn) {
		this.putColumnToZero(pColumn - 1);
		this.shrinkColumn(pColumn - 1);
		this.columns--;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#removeRow(int)
	 */
	public void removeRow(int pRow) {
		this.putRowToZero(pRow - 1);
		for (int i = 0; i < this.rowIndices.length; i++) {
			if (this.rowIndices[i] >= pRow - 1)
				this.rowIndices[i]--;
		}
		this.rows--;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#set(int, int, double)
	 */
	public void set(int pRow, int pColumn, double pElement)
			throws InvalidIndexException {
		this.checkRowIndex(pRow);
		this.checkColumnIndex(pColumn);
		if (pElement != 0)
			this.addNotNull(pRow - 1, pColumn - 1, pElement);
		else
			this.addNull(pRow - 1, pColumn - 1);
	}

	/**
	 * Setta il vettore contenente gli indici di inizio delle colonne
	 * 
	 * @param pColumnPointers
	 *            Il vettore contenente gli indici di inizio delle colonne
	 */
	public void setColumnPointers(int[] pColumnPointers) {
		this.columnPointers = pColumnPointers;
	}

	/**
	 * @param pColumns
	 *            Il vettore degli indici delle colonne
	 */
	public void setColumns(int pColumns) {
		this.columns = pColumns;
	}

	/**
	 * Setta il vettore degli indici delle righe
	 * 
	 * @param pRowIndices
	 *            Il vettore degli indici delle righe
	 */
	public void setRowIndices(int[] pRowIndices) {
		this.rowIndices = pRowIndices;
	}

	/**
	 * @param pRows
	 *            Il vettore delle righe
	 */
	public void setRows(int pRows) {
		this.rows = pRows;
	}

	/**
	 * @param pValues
	 *            Il vettore dei valori non nulli
	 */
	public void setValues(double[] pValues) {
		this.values = pValues;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#subMatrix(int, int)
	 */
	public IMatrix subMatrix(int pRows, int pColumns) {
		return null;
	}

	/**
	 * @see org.sesalab.IR.matrix.IMatrix#toArray()
	 */
	public double[][] toArray() {
		double[][] matrix = new double[this.rows][this.columns];
		for (int i = 0; i < this.columnPointers.length - 1; i++)
			for (int j = this.columnPointers[i]; j < this.columnPointers[i + 1]; j++)
				matrix[this.rowIndices[j]][i] = this.values[j];
		return matrix;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		BasicMatrix matrix = null;
		try {
			matrix = new BasicMatrix(this.rows, this.columns);
			for (int i = 0; i < this.columnPointers.length - 1; i++)
				for (int j = this.columnPointers[i]; j < this.columnPointers[i + 1]; j++)
					matrix.set(this.rowIndices[j] + 1, i + 1, this.values[j]);
		} catch (IncorrectDimensionException e) {
			// Eccezione che non si pu generare
			e.printStackTrace();
		} catch (InvalidIndexException e) {
			e.printStackTrace();
		}
		return (matrix != null) ? matrix.toString() : "";
	}

	/**
	 * Not implemented method
	 * 
	 * @see org.sesalab.IR.matrix.IMatrix#transpose()
	 */
	public IMatrix transpose() {
		return null;
	}

	/*
	 * Aggiunge un elemento non nullo alla matrice
	 * 
	 * @param pRow
	 *            Indice di riga
	 * @param pColumn
	 *            Indice di colonna
	 * @param pElement
	 *            Elemento da aggiungere
	 */
	private void addNotNull(int pRow, int pColumn, double pElement) {
		// Calcola il numero di elementi nella colonna pColumn
		int columnElements = this.columnPointers[pColumn + 1]
				- this.columnPointers[pColumn];
		try {
			if (this.getElement(pRow, pColumn) == 0) {
				int k = 0;
				// Cerca tra gli elementi nella colonna pColumn il primo indice
				// non inferiore a pRow
				for (k = this.columnPointers[pColumn]; k < this.columnPointers[pColumn]
						+ columnElements; k++) {
					if (this.rowIndices[k] < pRow)
						;
				}
				// Il nuovo elemento va alla posizione k. Incrementa values e
				// rowIndices
				this.strechValues(k, pElement);
				this.strechRowIndexes(k, pRow);

				for (int index = pColumn + 1; index < this.columnPointers.length; index++) {
					this.columnPointers[index] += 1;
				}
			} else {
				for (int k = this.columnPointers[pColumn]; k < this.columnPointers[pColumn]
						+ columnElements; k++) {
					if (this.rowIndices[k] == pRow)
						this.values[k] = pElement;
				}
			}
		} catch (InvalidIndexException e) {
			// Eccezione che non si pu verificare
		}
	}

	/*
	 * Aggiunge un elemento nullo alla matrice
	 * 
	 * @param pRow
	 *            Indice di riga
	 * @param pColumn
	 *            Indice di colonna
	 */
	private void addNull(int pRow, int pColumn) {
		try {
			if (this.getElement(pRow, pColumn) == 0)
				return;
			int ret = this.isLastElement(pRow, pColumn);
			if (ret == CSCSparseMatrix.LAST_IN_ROW)
				this.putRowToZero(pRow);
			else if (ret == CSCSparseMatrix.LAST_IN_COLUMN) {
				this.putColumnToZero(pColumn);
			} else if (ret == CSCSparseMatrix.LAST_IN_BOTH_DIRECTIONS) {
				this.putRowToZero(pRow);
				this.putColumnToZero(pColumn);
			} else {
				ArrayList<Integer> valuesToDelete = new ArrayList<Integer>();
				ArrayList<Integer> columnToDecrease = new ArrayList<Integer>();

				for (int j = this.columnPointers[pColumn]; j < this.columnPointers[pColumn + 1]; j++) {
					if (this.rowIndices[j] == pRow) {
						valuesToDelete.add(j);
						columnToDecrease.add(pColumn + 1);
						break;
					}
				}
				this.shrink(valuesToDelete);
				this.decrease(columnToDecrease);
			}
		} catch (InvalidIndexException e) {
			// Eccezione che non si pu verificare
		}
	}

	/*
	 * Verifica se l'indice di colonna  valido
	 * 
	 * @param pColumnIndex
	 *             L'indice di colonna
	 * @throws InvalidIndexException
	 *             Generata quando l'indice non  valido
	 */
	private void checkColumnIndex(int pColumnIndex)
			throws InvalidIndexException {
		if ((pColumnIndex < 0) || (pColumnIndex >= (this.columns + 1)))
			throw new InvalidIndexException("Invalid column index");
	}

	/*
	 * Verifica se l'indice di riga  valido 
	 * @param pRowIndex 
	 * 				L'indice di riga
	 * @throws InvalidIndexException 
	 * 				Generata quando l'indice non  valido
	 */
	private void checkRowIndex(int pRowIndex) throws InvalidIndexException {
		if ((pRowIndex < 0) || (pRowIndex >= (this.rows + 1)))
			throw new InvalidIndexException("Invalid row index");
	}

	/*
	 * Taglia una colonna della matrice
	 * 
	 * @param pColumnPointer
	 *            Puntatore alla colonna
	 * @param pColumnElements
	 *            Elementi della colonna
	 */
	private void cutMatrix(int pColumnPointer, int pColumnElements) {
		double[] auxValues = new double[this.values.length - pColumnElements];
		int[] auxRowIndices = new int[auxValues.length];
		int j = 0;
		for (; j < pColumnPointer; j++) {
			auxValues[j] = this.values[j];
			auxRowIndices[j] = this.rowIndices[j];
		}
		j += pColumnElements;
		for (; j < this.values.length; j++) {
			auxValues[j - pColumnElements] = this.values[j];
			auxRowIndices[j - pColumnElements] = this.rowIndices[j];
		}

		this.values = auxValues;
		this.rowIndices = auxRowIndices;
	}

	/*
	 * Decrementa i puntatori alle colonne
	 * 
	 * @param pColumnToDecrease
	 *            Puntatori da decrementare
	 */
	private void decrease(ArrayList<Integer> pColumnToDecrease) {
		for (int i = 0; i < this.columnPointers.length; i++) {
			if (pColumnToDecrease.contains(i))
				for (int j = i; j < this.columnPointers.length; j++) {
					this.columnPointers[j] -= 1;
				}
		}
	}

	/*
	 * Restituisce l'elemento di indice (pRowIndex,pColumnIndex)
	 * 
	 * @param pRow
	 *            L'indice di riga
	 * @param pColumn
	 *            L'indice di colonna
	 * @return L'elemento alla posizione (pRowIndex,pColumnIndex)
	 * @throws InvalidIndexException
	 *             Generata quando gli indici eccedono la dimensione della
	 *             matrice
	 */
	private double getElement(int pRow, int pColumn)
			throws InvalidIndexException {
		this.checkRowIndex(pRow);
		this.checkColumnIndex(pColumn);
		int columnElements = this.columnPointers[pColumn + 1]
				- this.columnPointers[pColumn];
		for (int k = this.columnPointers[pColumn]; k < this.columnPointers[pColumn]
				+ columnElements; k++) {
			if (this.rowIndices[k] == pRow)
				return this.values[k];
		}
		return 0;
	}

	/*
	 * Verifica se si sta mettendo a 0 l'ultimo elemento di una riga o di una
	 * colonna o di entrambe
	 * 
	 * @param pRow
	 *            Indice di riga
	 * @param pColumn
	 *            Indice di colonna
	 * @return CSCSparseMatrix.LAST_IN_BOTH_DIRECTIONS se l'elemento  l'ultimo
	 *         elemento della riga e della colonna,
	 *         CSCSparseMatrix.LAST_IN_COLUMN se  l'ultimo della colonna,
	 *         CSCSparseMatrix.LAST_IN_ROW se  l'ultimo della riga
	 * @throws InvalidIndexException
	 *             Generata quando gli indici eccedono la dimensione della
	 *             matrice
	 */
	private int isLastElement(int pRow, int pColumn)
			throws InvalidIndexException {
		this.checkRowIndex(pRow);
		this.checkColumnIndex(pColumn);
		boolean lastInColumn = false;
		boolean lastInRow = false;
		// Se il numero di elementi della colonna pColumn  1 allora
		// (pRow,pColumn)  l'ultimo della colonna
		int columnElements = this.columnPointers[pColumn + 1]
				- this.columnPointers[pColumn];
		if (columnElements == 1) {
			lastInColumn = true;
		}
		// se non ci sono altri elementi di indice pRow allora (pRow,pColumn) 
		// l'ultimo della riga
		int rowElements = 0;
		for (int k = 0; (k < this.rowIndices.length) && (rowElements < 2); k++) {
			if (this.rowIndices[k] == pRow)
				rowElements += 1;
		}
		if (rowElements == 1) {
			lastInRow = true;
		}

		if (lastInColumn && lastInRow)
			return CSCSparseMatrix.LAST_IN_BOTH_DIRECTIONS;
		if (lastInColumn)
			return CSCSparseMatrix.LAST_IN_COLUMN;
		if (lastInRow)
			return CSCSparseMatrix.LAST_IN_ROW;
		return (-1);
	}

	/*
	 * Mette a zero la colonna pColumn
	 * 
	 * @param pColumn
	 *            Indice della colonna da mettere a zero
	 */
	private void putColumnToZero(int pColumn) {
		int columnElements = this.columnPointers[pColumn + 1]
				- this.columnPointers[pColumn];
		for (int k = pColumn + 1; k < this.columnPointers.length; k++) {
			this.columnPointers[k] = this.columnPointers[k] - columnElements;
		}
		cutMatrix(this.columnPointers[pColumn], columnElements);
	}

	/*
	 * Mette a zero la riga pRow
	 * 
	 * @param pRow
	 *            Indice della riga da mettere a zero
	 */
	private void putRowToZero(int pRow) {
		ArrayList<Integer> valuesToDelete = new ArrayList<Integer>();
		ArrayList<Integer> columnToDecrease = new ArrayList<Integer>();

		for (int i = 0; i < this.columnPointers.length - 1; i++) {
			for (int j = this.columnPointers[i]; j < this.columnPointers[i + 1]; j++) {
				if (this.rowIndices[j] == pRow) {
					valuesToDelete.add(j);
					columnToDecrease.add(i + 1);
				}
			}
		}
		this.shrink(valuesToDelete);
		this.decrease(columnToDecrease);
	}

	/*
	 * Restringe i valori non nulli della matrice
	 * 
	 * @param pValuesToDelete
	 *            Valori non nulli da eliminare
	 */
	private void shrink(ArrayList<Integer> pValuesToDelete) {
		if (pValuesToDelete.size() == 0)
			return;
		double[] auxValues = new double[this.values.length
				- pValuesToDelete.size()];
		int[] auxRowIndices = new int[auxValues.length];
		int k = 0;

		// Copia solo i valori che non sono da cancellare
		for (int i = 0; i < this.values.length; i++) {
			if (!pValuesToDelete.contains(i)) {
				auxValues[k] = this.values[i];
				auxRowIndices[k] = this.rowIndices[i];
				k++;
			}
		}

		this.values = auxValues;
		this.rowIndices = auxRowIndices;
	}

	/*
	 * Restringe le colonne
	 * 
	 * @param pColumn
	 *            Indice della colonna da eliminare
	 */
	private void shrinkColumn(int pColumn) {
		int[] auxColumnPointers = new int[this.columnPointers.length - 1];
		int i = 0;
		for (; i < pColumn; i++) {
			auxColumnPointers[i] = this.columnPointers[i];
		}
		i++;
		for (; i < this.columnPointers.length; i++) {
			auxColumnPointers[i - 1] = this.columnPointers[i];
		}
		this.columnPointers = auxColumnPointers;
	}

	/*
	 * Restringe gli indici di riga
	 * 
	 * @param pRow
	 *            Indice di riga da eliminare
	 * @param pNewRow
	 *            Nuovo indice di riga da inserire
	 */
	private void strechRowIndexes(int pRow, int pNewRow) {
		int[] auxRowIndices = new int[this.rowIndices.length + 1];
		int i = 0;
		for (; i < pRow; i++) {
			auxRowIndices[i] = this.rowIndices[i];
		}
		auxRowIndices[i++] = pNewRow;

		for (; i < auxRowIndices.length; i++) {
			auxRowIndices[i] = this.rowIndices[i - 1];
		}
		this.rowIndices = auxRowIndices;
	}

	/*
	 * Allarga l'array contenente i valori non nulli
	 * 
	 * @param pIndex
	 *            Indice dove inserire l'elemento
	 * @param pElement
	 *            L'elemento da inserire
	 */
	private void strechValues(int pIndex, double pElement) {
		double[] auxValues = new double[this.values.length + 1];
		int i = 0;
		for (; i < pIndex; i++) {
			auxValues[i] = this.values[i];
		}
		auxValues[i++] = pElement;

		for (; i < auxValues.length; i++) {
			auxValues[i] = this.values[i - 1];
		}
		this.values = auxValues;
	}

	// Contiene gli indici di inizio di ciascuna colonna in values e rowIndices
	private int[] columnPointers;

	// Numero di colonne della matrice
	private int columns;

	// Contiene gli indici di riga degli elementi non nulli
	private int[] rowIndices;

	// Numero di righe
	private int rows;

	// Contiene i valori non nulli
	private double[] values;
}
