/*
 *  IMatrix.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.matrix;

import java.io.Serializable;

import org.sesalab.IR.matrix.exception.IncorrectDimensionException;
import org.sesalab.IR.matrix.exception.InvalidIndexException;

/**
 * Definisce l'interfaccia per una matrice
 * 
 * @author Paola Sgueglia
 */
public interface IMatrix extends Serializable {

	/**
	 * Restituisce l'elemento di indice <code>(pRowIndex,pColumnIndex)</code>
	 * 
	 * @param pRowIndex
	 *        L'indice di riga
	 * @param pColumnIndex
	 *        L'indice di colonna
	 * @return L'elemento alla posizione <code>(pRowIndex,pColumnIndex)</code>
	 * @throws InvalidIndexException
	 *         Generata quando gli indici eccedono la dimensione della matrice
	 */
	public double elementAt(int pRowIndex, int pColumnIndex) throws InvalidIndexException;

	/**
	 * Restituisce la colonna di indice <code>pColomunIndex</code>
	 * 
	 * @param pColomunIndex
	 *        L'indice di colonna
	 * @return La colonna specificata
	 * @throws InvalidIndexException
	 *         Generata quando l'indice di colonna eccede la dimensione della
	 *         matrice
	 */
	public double[] getColumn(int pColomunIndex) throws InvalidIndexException;

	/**
	 * Restituisce il numero di colonne
	 * 
	 * @return Il numero di colonne
	 */
	public int getColumnCount();

	/**
	 * Restituisce la riga di indice <code>pRowIndex</code>
	 * 
	 * @param pRowIndex
	 *        L'indice di riga
	 * @return La riga specificata
	 * @throws InvalidIndexException
	 *         Generata quando l'indice di riga eccede la dimensione della
	 *         matrice
	 */
	public double[] getRow(int pRowIndex) throws InvalidIndexException;

	/**
	 * Restituisce il numero di righe
	 * 
	 * @return Il numero di righe
	 */
	public int getRowCount();

	/**
	 * Incrementa il numero di colonne. Se prima di invocare il metodo la
	 * matrice ha dimensione [i,j], dopo avr dimensione [i,j + pAddedColumns]
	 * 
	 * @param pAddedColumns
	 *        Il numero di colonne da aggiungere
	 * @throws IncorrectDimensionException
	 *         Generata quando pAddedColumns  minore di 0
	 */
	public void increaseColumns(int pAddedColumns) throws IncorrectDimensionException;

	/**
	 * Incrementa il numero di righe. Se prima di invocare il metodo la matrice
	 * ha dimensione [i,j], dopo avr dimensione [i + pAddedRows,j]
	 * 
	 * @param pAddedRows
	 *        Il numero di righe da aggiungere
	 * @throws IncorrectDimensionException
	 *         Generata quando pAddedRows  minore di 0
	 */
	public void increaseRows(int pAddedRows) throws IncorrectDimensionException;

	/**
	 * Moltiplica la matrice in esame con una data in input
	 * 
	 * @param pToMultiply
	 *        La matrice con cui moltiplicare
	 * @return La matrice risultante
	 * @throws IncorrectDimensionException 
	 *         Generata quando il numero di colonne della matrice non coincide con
	 *         il numero di righe della matrice data in input
	 */
	public IMatrix mul(IMatrix pToMultiply) throws IncorrectDimensionException;

	/**
	 * Elimina una colonna della matrice
	 * 
	 * @param pColumn
	 *        Indice della colonna da eliminare
	 * @throws InvalidIndexException
	 *         Generata quando l'indice di colonna eccede la dimensione della
	 *         matrice
	 */
	public void removeColumn(int pColumn) throws InvalidIndexException;

	/**
	 * Elimina una riga della matrice
	 * 
	 * @param pRow
	 *        Indice della riga da eliminare
	 * @throws InvalidIndexException
	 *         Generata quando l'indice di riga eccede la dimensione della
	 *         matrice
	 */
	public void removeRow(int pRow) throws InvalidIndexException;

	/**
	 * Aggiunge l'elemento <code>pElement</code> in posizione
	 * <code>(pRowIndex,pColumnIndex)</code>
	 * 
	 * @param pRowIndex
	 *        L'indice di riga
	 * @param pColumnIndex
	 *        L'indice di colonna
	 * @param pElement
	 *        elemento da aggiungere
	 * @throws InvalidIndexException
	 *         Generata quando gli indici eccedono la dimensione della matrice
	 */
	public void set(int pRowIndex, int pColumnIndex, double pElement) throws InvalidIndexException;

	/**
	 * Restituisce la sottomatrice di dimensione pRows x pColumns costruita a
	 * partire dall'angolo superiore sinistro della matrice
	 * 
	 * @param pRows
	 *        Il numero di righe della sottomatrice
	 * @param pColumns
	 *        Il numero di colonne della sottomatrice
	 * @return La sottomatrice di dimensione pRows x pColumns
	 * @throws InvalidIndexException
	 *         Generata quando gli indici eccedono la dimensione della matrice
	 */
	public IMatrix subMatrix(int pRows, int pColumns) throws InvalidIndexException;
	
	/**
	 * Restuisce la transposta della matrice in esame
	 * @return La matrice trasposta
	 */
	public IMatrix transpose();
	
	/**
	 * @return La matrice sottoforma di array bi-dimensionale
	 */
	public double[][] toArray();
}