/*
 *  IIRModel.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.model;

import java.util.Collection;
import java.util.Map;

import org.sesalab.IR.model.exception.InvalidQueryException;
import org.sesalab.IR.model.vectorFeedbackPerformer.IVectorFeedbackPerformer;
import org.sesalab.IR.rankedList.IRankedList;
import org.sesalab.IR.rankedList.cutter.ICutter;

/**
 * Definisce l'interfaccia di un modello di Information Retrieval
 * 
 * @author Paola Sgueglia
 * @param <DocumentType>
 *        Tipo dei documenti
 * @param <TermType>
 *        Tipo dei termini
 */
public interface IIRModel<DocumentType, TermType> {
	/**
	 * @return La strategia per l'appicazione di feedback forniti dall'utente
	 */
	public IVectorFeedbackPerformer<DocumentType, TermType> getFeedbacksPerformer();

	/**
	 * Restituisce una ranked list in ordine decrescente di somiglianza.
	 * Consente di filtrare i documenti target.
	 * 
	 * @param pQueries
	 *        Documenti usati come query
	 * @param pTargets
	 *        Documenti usati come target
	 * @param pThreshold
	 *        Soglia utilizzata per tagliare la ranked list
	 * @return La ranked list ordinata in ordine decrescente
	 * @throws InvalidQueryException
	 *         Generata quando si tenta di usare come query un documento non
	 *         presente nella collezione
	 */
	public IRankedList getRankedList(Collection<DocumentType> pQueries, Collection<DocumentType> pTargets, Double pThreshold) throws InvalidQueryException;

	/**
	 * Restituisce una ranked list in ordine decrescente di somiglianza
	 * 
	 * @param pQueries
	 *        Documenti usati come query
	 * @param pThreshold
	 *        Soglia utilizzata per tagliare la ranked list
	 * @return La ranked list ordinata in ordine decrescente
	 * @throws InvalidQueryException
	 *         Generata quando si tenta di usare come query un documento non
	 *         presente nella collezione
	 */
	public IRankedList getRankedList(Collection<DocumentType> pQueries, Double pThreshold) throws InvalidQueryException;

	/**
	 * Restituisce una ranked list in ordine decrescente di somiglianza.
	 * Consente di filtrare i documenti target e l'applicazione di feedback
	 * forniti dall'utente
	 * 
	 * @param pUserFeedbacks
	 *        Documenti usati come query e relativi feedback forniti dall'utente
	 * @param pTargets
	 *        Documenti usati come target
	 * @param pThreshold
	 *        Soglia utilizzata per tagliare la ranked list
	 * @return La ranked list ordinata in ordine decrescente
	 * @throws InvalidQueryException
	 *         Generata quando si tenta di usare come query un documento non
	 *         presente nella collezione
	 */
	public IRankedList getRankedList(Map<DocumentType, Feedbacks<DocumentType>> pUserFeedbacks, Collection<DocumentType> pTargets, Double pThreshold) throws InvalidQueryException;

	/**
	 * Restituisce una ranked list in ordine decrescente di somiglianza.
	 * Consente l'applicazione di feedback forniti dall'utente
	 * 
	 * @param pUserFeedbacks
	 *        Documenti usati come query e relativi feedback forniti dall'utente
	 * @param pThreshold
	 *        Soglia utilizzata per tagliare la ranked list
	 * @return La ranked list ordinata in ordine decrescente
	 * @throws InvalidQueryException
	 *         Generata quando si tenta di usare come query un documento non
	 *         presente nella collezione
	 */
	public IRankedList getRankedList(Map<DocumentType, Feedbacks<DocumentType>> pUserFeedbacks, Double pThreshold) throws InvalidQueryException;

	/**
	 * Setta la strategia per l'appicazione di feedback forniti dall'utente
	 * 
	 * @param pVectorFeedbackPerformer
	 *        La strategia per l'appicazione di feedback forniti dall'utente
	 */
	public void setFeedbacksPerformer(IVectorFeedbackPerformer<DocumentType, TermType> pVectorFeedbackPerformer);
	
	/**
	 * Setta l'algoritmo di taglio da usare per tagliare la ranked list
	 * @param pCutter L'algoritmo di taglio
	 */
	public void setRankedListCutter(ICutter pCutter);
}
