/*
 *  LatentSemanticIndexing.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.model.LDA;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.rosuda.JRI.REXP;
import org.rosuda.JRI.Rengine;
import org.sesalab.IR.documentSpace.ITermByDocumentMatrix;
import org.sesalab.IR.documentSpace.exception.ObjectNotFoundException;
import org.sesalab.IR.documentSpace.weigher.BasicWeigher;
import org.sesalab.IR.matrix.BasicMatrix;
import org.sesalab.IR.matrix.IMatrix;
import org.sesalab.IR.matrix.ISparseMatrix;
import org.sesalab.IR.matrix.exception.IncorrectDimensionException;
import org.sesalab.IR.matrix.exception.InvalidIndexException;
import org.sesalab.IR.model.Feedbacks;
import org.sesalab.IR.model.LSI.SSVD.SSVDManager;
import org.sesalab.IR.model.LSI.exception.InvalidSubspaceDimensionException;
import org.sesalab.IR.model.VSM.VectorSpaceModel;
import org.sesalab.IR.model.exception.IncorrectVectorSpaceException;
import org.sesalab.IR.model.exception.InvalidQueryException;
import org.sesalab.IR.model.vectorSimilarityPerformer.IVectorSimilarityPerformer;
import org.sesalab.IR.rankedList.BasicRankedList;
import org.sesalab.IR.rankedList.IRankedList;
import org.sesalab.IR.rankedList.RankedListElement;

import Jama.SingularValueDecomposition;

/**
 * Classe che implementa il modello di Information Retrieval denominato Latent
 * Semantic Indexing
 * 
 * @author Paola Sgueglia
 * @param <DocumentType>
 *        Tipo dei documenti indicizzati
 * @param <TermType>
 *        Tipo dei termini indicizzati
 */
public class LDAIndexing<DocumentType extends Comparable, TermType> extends VectorSpaceModel<DocumentType, TermType> {
// TODO: no longer make this a subclass of vectorspacemodel? it isn't one.
	
	/** So:
	 * 	**	preprocess docs (already done?)
	 * 	**	convert to form we can pass to lda
	 * 			***** FIgure out where exactly we do that -- where should model get initialized??
	 * 		call lda
	 * 		get data back
	 * 		make the model
	 * 
	 * TODO: would it make sense to move LDA to it's own different plug-in?
	*/

	private static Rengine re = null;
	
	/**
	 * Variabile utilizzata per indicare che bisogna applicare i feedback sulla
	 * matrice termini-documenti
	 */
	public static final int APPLY_FEEDBACKS_THEN_REDUCE = 2;

	/**
	 * Nome del file che contiene la decomposizione ai valori singolari
	 */
	public static final String DECOMPOSITION_FILE_NAME = "decomposition.svd";

	/**
	 * Variabile utilizzata per indicare che bisogna applicare i feedback sulla
	 * matrice dei concetti
	 */
	public static final int REDUCE_THEN_APPLY_FEEDBACKS = 1;

	private double topicThreshold = .03;

	private boolean useHellingers;

	private HashMap<String,String> doclines;

	private int numberTopics;

	private HashMap<String, HashMap<String, Double>> lastComputedTopics;

	private double alpha;

	private double eta;

	private double beta;

	private int iterations;

	private HashMap<String, List<String>> links;

	private REXP model;

	private HashMap<String, HashMap<String, Double>> lastComputedTopicWords;

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * valore di somiglianza tra due documenti  calcolato attravero
	 * pSimilarityPerformer
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pVectorSimilarityPerformer
	 *        La strategia di calcolo del valore di somiglianza tra due
	 *        documenti
	 * @param pWorkingFolder
	 *        La directory in cui verr memorizzata la decomposizione ai valori
	 *        singolari della matrice termini-documenti
	 * @param pReducedSpaceDimension
	 *        La dimensione del sottospazio
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 * @throws InvalidSubspaceDimensionException
	 *         Generata quando si sceglie una dimensione per il sottospazio
	 *         latente non valida
	 */
//	public LDAIndexing(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, IVectorSimilarityPerformer pVectorSimilarityPerformer, String pWorkingFolder, int pReducedSpaceDimension, boolean pForceRebuilding, boolean useHellingers) throws InvalidSubspaceDimensionException {
//		super(pTermByDocumentMatrix, pVectorSimilarityPerformer);
//		if (pReducedSpaceDimension < 2)
//			throw new InvalidSubspaceDimensionException();
//		this.init(pTermByDocumentMatrix, pReducedSpaceDimension, pWorkingFolder, LDAIndexing.REDUCE_THEN_APPLY_FEEDBACKS, pForceRebuilding,useHellingers, null, 25, .1, .1, 0, .03, 100);
//	}

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * valore di somiglianza tra due documenti  calcolato attravero
	 * pSimilarityPerformer
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pVectorSimilarityPerformer
	 *        La strategia di calcolo del valore di somiglianza tra due
	 *        documenti
	 * @param pWorkingFolder
	 *        La directory in cui verr memorizzata la decomposizione ai valori
	 *        singolari della matrice termini-documenti
	 * @param pReducedSpaceDimension
	 *        La dimensione del sottospazio
	 * @param pFeedbackStrategy
	 *        La trategia da adottare per l'applicazione dei feedback. E'
	 *        possibile scegliera tra
	 *        <code>LatentSemanticIndexing.APPLY_FEEDBACKS_THEN_REDUCE</code>
	 *        e <code>LatentSemanticIndexing.REDUCE_THEN_APPLY_FEEDBACKS</code>
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 * @throws InvalidSubspaceDimensionException
	 *         Generata quando si sceglie una dimensione per il sottospazio
	 *         latente non valida
	 */
//	public LDAIndexing(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, IVectorSimilarityPerformer pVectorSimilarityPerformer, String pWorkingFolder, int pReducedSpaceDimension, int pFeedbackStrategy, boolean pForceRebuilding, boolean useHellingers) throws InvalidSubspaceDimensionException {
//		super(pTermByDocumentMatrix, pVectorSimilarityPerformer);
//		if (pReducedSpaceDimension < 2)
//			throw new InvalidSubspaceDimensionException();
//		this.init(pTermByDocumentMatrix, pReducedSpaceDimension, pWorkingFolder, pFeedbackStrategy, pForceRebuilding, useHellingers, null, 25, .1, .1, 0, .03, 100);
//	}

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * coseno dell'angolo compreso tra i due vettori che rappresentano due
	 * documenti  usato come valore di somiglianza tra due documenti
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pWorkingFolder
	 *        La directory in cui verr memorizzata la decomposizione ai valori
	 *        singolari della matrice termini-documenti
	 * @param pReducedSpaceDimension
	 *        La dimensione del sottospazio
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 * @param numberTopics
	 * @throws InvalidSubspaceDimensionException
	 *         Generata quando si sceglie una dimensione per il sottospazio
	 *         latente non valida
	 */
//	public LDAIndexing(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, String pWorkingFolder, int pReducedSpaceDimension, boolean pForceRebuilding, boolean useHellingers, int numberTopics) throws InvalidSubspaceDimensionException {
//		super(pTermByDocumentMatrix);
//		if (pReducedSpaceDimension < 2)
//			throw new InvalidSubspaceDimensionException();
//		this.init(pTermByDocumentMatrix, pReducedSpaceDimension, pWorkingFolder, LDAIndexing.REDUCE_THEN_APPLY_FEEDBACKS, pForceRebuilding, useHellingers, null, numberTopics, .1, .1, 0, .03, 100);
//	}

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * coseno dell'angolo compreso tra i due vettori che rappresentano due
	 * documenti  usato come valore di somiglianza tra due documenti
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pWorkingFolder
	 *        La directory in cui verr memorizzata la decomposizione ai valori
	 *        singolari della matrice termini-documenti
	 * @param pReducedSpaceDimension
	 *        La dimensione del sottospazio
	 * @param pFeedbackStrategy
	 *        La trategia da adottare per l'applicazione dei feedback. E'
	 *        possibile scegliera tra
	 *        <code>LatentSemanticIndexing.APPLY_FEEDBACKS_THEN_REDUCE</code>
	 *        e <code>LatentSemanticIndexing.REDUCE_THEN_APPLY_FEEDBACKS</code>
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 * @throws InvalidSubspaceDimensionException
	 *         Generata quando si sceglie una dimensione per il sottospazio
	 *         latente non valida
	 */
//	public LDAIndexing(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, String pWorkingFolder, int pReducedSpaceDimension, int pFeedbackStrategy, boolean pForceRebuilding) throws InvalidSubspaceDimensionException {
//		super(pTermByDocumentMatrix);
//		if (pReducedSpaceDimension < 2)
//			throw new InvalidSubspaceDimensionException();
//		this.init(pTermByDocumentMatrix, pReducedSpaceDimension, pWorkingFolder, pFeedbackStrategy, pForceRebuilding, true, null, 25, .1, .1, 0, .03, 100);
//	}

	public LDAIndexing(HashMap<String,String> doclines, HashMap<String,List<String>> links, String pWorkingFolder,
			int pReducedSpaceDimension, boolean pForceRebuilding, boolean useHellingers, int numberTopics, double alpha, double eta, double beta, double topicThreshold2, int iterations) {
		super(null);
//		if (pReducedSpaceDimension < 2)
//			throw new InvalidSubspaceDimensionException();
		this.init(null, pReducedSpaceDimension, pWorkingFolder, LDAIndexing.REDUCE_THEN_APPLY_FEEDBACKS, pForceRebuilding, useHellingers, doclines, numberTopics, alpha, eta, beta, topicThreshold2, iterations, links);

	}
	
	public HashMap<String, HashMap<String, Double>> getLastComputedTopicWords(double threshold) {
		// Something went wrong; no model and/or R was not initialized.
		if (model == null || re == null)
			return null;
		
		
		HashMap<String, HashMap<String, Double>> map = new HashMap<String, HashMap<String,Double>>();
		int count = 1;
		
		String[] word = re.eval("colnames(model$topics)").asStringArray();
		
		// Assuming lastComputedTopics is insertion-ordered
		for (String topic : lastComputedTopics.keySet()) {
			HashMap<String,Double> wordMap = new HashMap<String,Double>();

			// Get probabilties for each word wrt the topic, put them in wordMap
			// TODO: double check this is getting the right rows...
			
			double[] targetCol = re.eval("model$topics["+count+",]").asDoubleArray();
			targetCol = normalizeColumn(targetCol);
			
			for (int i = 0; i < word.length; i++) {
				if (targetCol[i] > threshold)
					wordMap.put(word[i], targetCol[i]);
			}
				
			map.put(topic, wordMap);
			count++;
		}
		
		//TODO: this should work, right?
		return map;		
	}

	/**
	 * @see org.sesalab.IR.model.VSM.VectorSpaceModel#getRankedList(java.util.Collection,
	 *      java.util.Collection, java.lang.Double)
	 */
	public IRankedList getRankedList(Collection<DocumentType> pQueries, Collection<DocumentType> pTargets, Double pThreshold) throws InvalidQueryException {
		// TODO: reimplement multiple document queries?
		if (!(this.doclines.containsKey(pQueries.iterator().next())))
			throw new InvalidQueryException();
	
		// Something went wrong; no model and/or R was not initialized.
		if (model == null || re == null)
			return null;
	
		 // TODO, TS: what is difference between 2 rankedlist methods??

		// TODO: switch strings over to documenttypes?
		// Given the model and the query, find the similarities between the query and all other docs in the model
	        try {
	        	REXP x=null;
	        	IRankedList rankedList = new BasicRankedList(this.rankedListCutter);
	        	
	        	re.assign("model",model);
			
				DocumentType document = pQueries.iterator().next();
				int documentColumn = getColumn(document); // TOOD: handle multiple queries
				
	        	double[] queryCol = normalizeColumn(re.eval("model$document_sums[,"+documentColumn+"]").asDoubleArray());
				
				for (int d = 1; d <= doclines.size(); d++) {
					if (d == documentColumn)
						continue;
					
					Double similarityValue = -1.0;
					DocumentType targetDocument = (DocumentType) getColumnName(d);
					
					if (useHellingers) {
						double[] targetCol = re.eval("model$document_sums[,"+d+"]").asDoubleArray();
						targetCol = normalizeColumn(targetCol);
						similarityValue = hellingerDistance(queryCol, targetCol);
						// Hellinger ranges from 0 to the sqrt(2), and it is a distance; lower distance implies higher similarity
						// So change hellinger into something that can be interpreted as a similarity %
						similarityValue = (Math.sqrt(2)-similarityValue)/Math.sqrt(2);						
					} else {
						
		        	//	System.out.println(re.eval("predictive.link.probability(matrix(c("+documentColumn+","+d+"),1,2), model$document_sums, alpha=0.1, beta=1)"));
		        		
		        		similarityValue = re.eval("predictive.link.probability(matrix(c("+documentColumn+","+d+"),1,2), model$document_sums, alpha=0.1, beta=1)").asDouble();

					
					}
	        			RankedListElement rankedListElement = null;
					
	        		if (document.compareTo(targetDocument) > 0)
						rankedListElement = this.createRankedListElement(targetDocument, document, similarityValue);
					else
						rankedListElement = this.createRankedListElement(document, targetDocument, similarityValue);
					if (!rankedList.contains(rankedListElement))
						rankedList.addElement(rankedListElement);
					
	        	}
				
				if (!useHellingers) {
	        		// Predictive link probability should be interpreted in relation to itself, so
	        		// normalize values such that the min = 0 and the max = 100
					double minSimilarity = rankedList.getMinSimilarity();
					double maxSimilarity = rankedList.getMaxSimilarity() - minSimilarity;
					
					for (RankedListElement elem : rankedList)
						elem.setSimilarityValue((elem.getSimilarityValue()-minSimilarity)/maxSimilarity);
				}
	        	
				// Get the topic information we want to visualize
				// probably a hashmap of hashmaps;
				// each topic (set of words) has an entry, each entry has an entry for each document in corpus with %.
						// can dispose %s below the threshold tho
				// So, go through each topic. topic_sums is a KxV matrix
				// Faster/better to do this in R? Are hashmaps of hashmaps excessive?

				re.eval("words = top.topic.words(model$topics,num.words=5,by.score=FALSE)");
				int numberTopics = re.eval("ncol(words)").asInt();
				//TODO: change string to documentdescriptor
				lastComputedTopics = new LinkedHashMap<String,HashMap<String,Double>>();
				lastComputedTopicWords = new HashMap<String, HashMap<String,Double>>();
				for (int t = 1; t <= numberTopics; t++ ){
					HashMap<String,Double> topicContents = new HashMap<String,Double>();
					String topicName = "";
					
					// Find the top X words
					String words[] = re.eval("words[,"+t+"]").asStringArray();
					
					for (String word : words)
						topicName += word + ", ";
					
					//TODO, FINAL: put an option to calculate probabilities without recomputing the model here, and
					//             use that when documents change?? should be pretty quick, eliminates need to precalculate 
					//             EVERY possible probability
					
					// For each document
		        	int numberDocs = doclines.size();
					for (int d = 1; d <= numberDocs; d++) {
						// If document is an artifact or the query, save it's probability w.r.t. this topic
						boolean docIsQuery = getColumnName(d).equals(document);
						if (getColumnName(d).toString().charAt(0) != '/' ||
							docIsQuery) {
							// TODO: '/' test is crude, but works for now
			        		//System.out.println("doc:"+termByDocumentMatrix.getDocument(d));
							double[] targetCol = re.eval("model$document_sums[,"+d+"]").asDoubleArray();
							//if (docIsQuery)
							//	System.out.println();
							targetCol = normalizeColumn(targetCol);
						//	System.out.println(targetCol[t-1]);
							
							if (targetCol[t-1] > topicThreshold || docIsQuery) 
								topicContents.put(getColumnName(d).toString(), targetCol[t-1]);
						}
		        		
		        	}
						
					
					lastComputedTopics.put(topicName, topicContents);
				}
				
				
				
	        	// Turn R off for now
	        	// TODO: turn it off when we're done? do we really want to keep R running?
	        	// Doesn't seem to want to come back on once it's been turned off though
	        	//re.stop();
				
				
				return rankedList;
			} catch (Exception e) {
				System.out.println("EX:"+e);
				e.printStackTrace();
			}
			
		return null;
	}
	
	private String getColumnName(int d) {
		int number = 1;
		for (String key : doclines.keySet()) {
			if (number == d)
				return key;
			number++;
		}
		return "";
	}

	private int getColumn(DocumentType document) {
		// find the column # of the document; starting at 1, going up
		int number = 1;
		for (String key : doclines.keySet()) {
			if (key.equals(document.toString()))
				return number;
			number++;
		}
		return -1;
	}

	private double[] normalizeColumn(double[] array) {
		double total = 0;
		// Take an array, which is the number of times a document is assigned to a particular topic, and turn it into probabilities that document belongs to a topic
		for (double d : array)
			total += d;
		
		for(int i = 0; i < array.length; i++)
			array[i] /= total;
		
		return array;
	}

	public double hellingerDistance(double probability1[], double probability2[]) {
		Double sum = 0.0;

		// does this handle the non-probability values we pass it?
		for (int i = 0; i < probability1.length; i++)
			sum += Math.pow(Math.sqrt(probability1[i]) - Math.sqrt(probability2[i]),2);
		
		return Math.sqrt(sum);
	}
	
	public HashMap<String,HashMap<String,Double>> getLastComputedTopics() {
		return lastComputedTopics;
	}
	
	/**
	 * @see org.sesalab.IR.model.VSM.VectorSpaceModel#getRankedList(java.util.Collection,
	 *      java.lang.Double)
	 */
	public IRankedList getRankedList(Collection<DocumentType> pQueries, Double pThreshold) throws InvalidQueryException {
		return this.getRankedList(pQueries, null, pThreshold);
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#getRankedList(java.util.Map,
	 *      java.util.Collection, java.lang.Double)
	 */
	public IRankedList getRankedList(Map<DocumentType, Feedbacks<DocumentType>> pUserFeedbacks, Collection<DocumentType> pTargets, Double pThreshold) throws InvalidQueryException {
		if (!(this.termByDocumentMatrix.containsAll(pUserFeedbacks.keySet())))
			throw new InvalidQueryException();
		
		
		//		
//		IRankedList rankedList = new BasicRankedList(this.rankedListCutter);
//		try {
//			// Si decompone la matrice dei termini-documenti
//			ISparseMatrix matrix = this.termByDocumentMatrix.getWeightedMatrix();
//			
//			// Si determinano le colonne relative ai feedback forniti
//			Map<Integer, Feedbacks<Integer>> columnUserFeedbacks = new HashMap<Integer, Feedbacks<Integer>>();
//			Iterator<DocumentType> queryIterator = pUserFeedbacks.keySet().iterator();
//			while (queryIterator.hasNext()) {
//				DocumentType query = queryIterator.next();
//				Feedbacks<DocumentType> feedbacks = pUserFeedbacks.get(query);
//
//				// Definiamo i feedback positivi
//				List<Integer> correctDocumentsColumnIndex = new LinkedList<Integer>();
//				Iterator<DocumentType> correctDocumentsIterator = feedbacks.getCorrectDocuments().iterator();
//				while (correctDocumentsIterator.hasNext()) {
//					DocumentType correctDocument = correctDocumentsIterator.next();
//					correctDocumentsColumnIndex.add(this.termByDocumentMatrix.getColumn(correctDocument));
//				}
//
//				// Definiamo i feedback negative
//				List<Integer> falsePositivesColumnIndex = new LinkedList<Integer>();
//				Iterator<DocumentType> falsePositivesIterator = feedbacks.getFalsePositiveDocuments().iterator();
//				while (falsePositivesIterator.hasNext()) {
//					DocumentType falsePositive = falsePositivesIterator.next();
//					falsePositivesColumnIndex.add(this.termByDocumentMatrix.getColumn(falsePositive));
//				}
//
//				Integer queryColumnIndex = new Integer(this.termByDocumentMatrix.getColumn(query));
//				Feedbacks<Integer> columnFeedbacks = new Feedbacks<Integer>();
//				columnFeedbacks.setCorrectDocuments(correctDocumentsColumnIndex);
//				columnFeedbacks.setFalsePositiveDocuments(falsePositivesColumnIndex);
//
//				columnUserFeedbacks.put(queryColumnIndex, columnFeedbacks);
//			}
//
//			// Si applicano i feedback
//			IMatrix reducedMatrix = null;
//			if (this.feedbackStrategy == LDAIndexing.APPLY_FEEDBACKS_THEN_REDUCE) {
//				this.vectorFeedbackPerformer.apply(matrix, columnUserFeedbacks);
//				reducedMatrix = this.getReducedDocumentMatrix(matrix);
//			} else {
//				reducedMatrix = this.getReducedDocumentMatrix(matrix); 
//				this.vectorFeedbackPerformer.apply(reducedMatrix, columnUserFeedbacks);
//			}
//
//			// Si ottiene la ranked list
//			rankedList = this.buildRankedList(reducedMatrix, pUserFeedbacks.keySet(), pTargets, pThreshold);
//		} catch (IncorrectVectorSpaceException e) {
//			// Eccezione che non si verifica mai
//			e.printStackTrace();
//		} catch (ObjectNotFoundException e) {
//			// Eccezione che non si verifica mai
//			e.printStackTrace();
//		}

		return null;
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#getRankedList(java.util.Map,
	 *      java.lang.Double)
	 */
	public IRankedList getRankedList(Map<DocumentType, Feedbacks<DocumentType>> pUserFeedbacks, Double pThreshold) throws InvalidQueryException {
		return this.getRankedList(pUserFeedbacks, null, pThreshold);
	}

	/*
	 * Determina la matrice ridotta che fornisce la relazione tra documenti
	 * 
	 * @param pMatrix
	 *        La matrice termini documenti da decomporre
	 * @return La matrice dei documenti
	 */
	private REXP getReducedDocumentMatrix() {
		// TDOO: so, here compute the model based on the design artifacts
		
		// Get the unweighted term-document matrix to create the model from
//		termByDocumentMatrix.setWeigher(new BasicWeigher());
//		IMatrix matrix = termByDocumentMatrix.getWeightedMatrix();
	
		// TODO: User must have R_HOME set; if this isn't done automatically, that could be a problem?
		
		// Initialize JRI, the Java-to-R binding, and then call the LDA library.
		// just making sure we have the right version of everything
		if ( re == null) {
			if (!Rengine.versionCheck()) {
			    System.err.println("COCONUT LDA: JRI ERROR: ** Version mismatch - Java files don't match library version.");
			    return null;
			}
			// Preprep; automate any of it?
			   // install R
			   // install jRI(these need to be installed into the R directory, not user, unless there's a way to load user libs)
			   // install LDA (we could do this)
			   // add C:\Program Files\R\R-2.11.1\bin to path
			   // in windows, R_HOME should automatically be set up apparently
			
			//install.packages("rJava")
			//install.packages("lda")

				// Start R
				String[] args = {"--vanilla"};
				re=new Rengine(args, false, new TextConsole());
				re.eval("library(lda)");	
				
		
				// the engine creates R is a new thread, so we should wait until it's ready
		        if (!re.waitForR()) {
		            System.err.println("Cannot load R");
		            return null;
		        }
		}

        try {
        	REXP x=null;
        	
        	// TODO: cache model, unless forceRebuilding is set?
        	
        	// Turn documents into a format lda will understand.
        	// We could do the lexicalize step in java; could be more efficient since we've already processed the docs some
        	String linksStr = "";
        	
        	// make the empty links, if we don't have any links
        	if (links == null || links.isEmpty()) {
	        	for (int d = 1; d <= doclines.size(); d++) {
	        				
	    			// Construct empty links
	    			if(linksStr != "")
	    				linksStr+= ",";
	    			linksStr+= "integer(0)";
	        	}
        	} else {
        		//Turn the hashmap into the links format the R LDA wants
        		//TODO: any better way to do this?
        		for (String source : doclines.keySet()) {
        			List<String> dests = links.get(source);
        			// Add commas
	    			if(linksStr != "")
	    				linksStr+= ",";
	    			
	    			// No links for this document
	    			if (dests == null || dests.isEmpty())
	    				linksStr+= "integer(0)";
	    			else {
	    				// Make a list of the links
	    				linksStr +="as.vector(list(";
	    				boolean first = true;
	    				for (String dest : dests){
	    					int column = getColumn((DocumentType) dest)-1;
	    					if (column != -1) {
	    						if (!first)
	    							linksStr +=",";
	    						else
	    							first = false;
	    						linksStr += column;
	    					}
	    				}
	    				linksStr += "),mode=\"integer\")";
	    			}
        		}
        	}
        	
        	String[] strArr = new String[doclines.size()];
        	System.arraycopy(doclines.values().toArray(), 0, strArr, 0, doclines.size());

        	// Load LDA, process the corpus, generate the model
        	re.assign("doclines",strArr);
        	// Fix the seed so that we generate the same topics each time
        	// Although, should this even be necessary? overall similarity shouldn't fluctuate, only the individual topics
        	// But it seems to fluctuate +-10%
        	re.eval("set.seed(86753088)");
        	//System.out.println(x=re.eval("doclines"));
			re.eval("corpus = lexicalize(doclines)");
			//			re.eval("model = lda.collapsed.gibbs.sampler(documents=corpus$documents, K=25, vocab=corpus$vocab, num.iterations=100, alpha=.1, eta=.1, compute.log.likelihood=TRUE)");
			re.eval("model = rtm.collapsed.gibbs.sampler(documents=corpus$documents,links=list("+linksStr+"), K="+numberTopics+", vocab=corpus$vocab, num.iterations="+iterations+", alpha="+alpha+", eta="+eta+", beta="+beta+")");
			//System.out.println(re.eval("model"));
			x=re.eval("model");
		//	System.out.println(re.eval("model$document_expects"));

			//System.out.println(re.eval("corpus$vocab"));
			//System.out.println(re.eval("model$document_sums"));
			// Now we can just return the model to use later
			return x;
			
		} catch (Exception e) {
			System.err.println("EX:"+e);
			e.printStackTrace();
		}
	

		// Something went wrong, return nothing
		return null;
	}

	/*
	 * Crea un elemento della ranked list
	 * @param pSource
	 *        Documento source
	 * @param pTarget
	 *        Documento target
	 * @param pSimilarityValue
	 *        Valore di somiglianza
	 * @return Un nuovo elemento della ranked list
	 */
	private RankedListElement<DocumentType> createRankedListElement(DocumentType pSource, DocumentType pTarget, double pSimilarityValue) {
		RankedListElement<DocumentType> element = new RankedListElement<DocumentType>();
		element.setSource(pSource);
		element.setTarget(pTarget);
		element.setSimilarityValue(pSimilarityValue);
		return element;
	}
	
	/*
	 * Inizializza le variabili di istanza della classe
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice termini-documenti
	 * @param pSubspaceDimension
	 *        La dimesnione del sottospazio latente
	 * @param pWorkingFolder
	 *        La directory di lavoro
	 * @param pFeedbackStrategy
	 *        La strategia di feedback da adottare
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 */
	private void init(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, 
			int pSubspaceDimension, String pWorkingFolder, int pFeedbackStrategy, 
			boolean pForceRebuilding, boolean useHellingers, HashMap<String,String> doclines, 
			int numberTopics, double alpha, double eta, double beta, double topicThreshold2, int iterations, HashMap<String,List<String>> links) {
		this.subspaceDimension = pSubspaceDimension;
		this.workingFolder = pWorkingFolder;
		this.feedbackStrategy = pFeedbackStrategy;
		//this.svdManager = new SSVDManager(pTermByDocumentMatrix.getWeightedMatrix(), pWorkingFolder);
		this.forceRebuilding = pForceRebuilding;
		//this.termByDocumentMatrix = pTermByDocumentMatrix;
		this.useHellingers = useHellingers;
		this.doclines = doclines;
		this.numberTopics = numberTopics;

		this.links = links;
		this.alpha = alpha;
		this.eta = eta;
		this.beta = beta;
		this.topicThreshold = topicThreshold2;
		this.iterations = iterations;
		
		// Actually generate the model!
		model = getReducedDocumentMatrix();
	}

	// La strategia da adottare per l'applicazione dei feedback
	private int feedbackStrategy;

	// Determina se effettuare il re-building della decomposizione
	private boolean forceRebuilding;

	// Dimensione del sottospazio latente
	private int subspaceDimension;

	// Il manager per la decomposizione ai valori singolari
	private SSVDManager svdManager;

	// Directory di lavoro
	private String workingFolder;


}