/*
 *  LatentSemanticIndexing.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.model.LSI;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.sesalab.IR.documentSpace.ITermByDocumentMatrix;
import org.sesalab.IR.documentSpace.exception.ObjectNotFoundException;
import org.sesalab.IR.matrix.BasicMatrix;
import org.sesalab.IR.matrix.IMatrix;
import org.sesalab.IR.matrix.ISparseMatrix;
import org.sesalab.IR.matrix.exception.IncorrectDimensionException;
import org.sesalab.IR.matrix.exception.InvalidIndexException;
import org.sesalab.IR.model.Feedbacks;
import org.sesalab.IR.model.LSI.SSVD.SSVDManager;
import org.sesalab.IR.model.LSI.exception.InvalidSubspaceDimensionException;
import org.sesalab.IR.model.VSM.VectorSpaceModel;
import org.sesalab.IR.model.exception.IncorrectVectorSpaceException;
import org.sesalab.IR.model.exception.InvalidQueryException;
import org.sesalab.IR.model.vectorSimilarityPerformer.IVectorSimilarityPerformer;
import org.sesalab.IR.rankedList.BasicRankedList;
import org.sesalab.IR.rankedList.IRankedList;

import Jama.SingularValueDecomposition;

/**
 * Classe che implementa il modello di Information Retrieval denominato Latent
 * Semantic Indexing
 * 
 * @author Paola Sgueglia
 * @param <DocumentType>
 *        Tipo dei documenti indicizzati
 * @param <TermType>
 *        Tipo dei termini indicizzati
 */
public class LatentSemanticIndexing<DocumentType extends Comparable, TermType> extends VectorSpaceModel<DocumentType, TermType> {

	/**
	 * Variabile utilizzata per indicare che bisogna applicare i feedback sulla
	 * matrice termini-documenti
	 */
	public static final int APPLY_FEEDBACKS_THEN_REDUCE = 2;

	/**
	 * Nome del file che contiene la decomposizione ai valori singolari
	 */
	public static final String DECOMPOSITION_FILE_NAME = "decomposition.svd";

	/**
	 * Variabile utilizzata per indicare che bisogna applicare i feedback sulla
	 * matrice dei concetti
	 */
	public static final int REDUCE_THEN_APPLY_FEEDBACKS = 1;

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * valore di somiglianza tra due documenti  calcolato attravero
	 * pSimilarityPerformer
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pVectorSimilarityPerformer
	 *        La strategia di calcolo del valore di somiglianza tra due
	 *        documenti
	 * @param pWorkingFolder
	 *        La directory in cui verr memorizzata la decomposizione ai valori
	 *        singolari della matrice termini-documenti
	 * @param pReducedSpaceDimension
	 *        La dimensione del sottospazio
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 * @throws InvalidSubspaceDimensionException
	 *         Generata quando si sceglie una dimensione per il sottospazio
	 *         latente non valida
	 */
	public LatentSemanticIndexing(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, IVectorSimilarityPerformer pVectorSimilarityPerformer, String pWorkingFolder, int pReducedSpaceDimension, boolean pForceRebuilding) throws InvalidSubspaceDimensionException {
		super(pTermByDocumentMatrix, pVectorSimilarityPerformer);
		if (pReducedSpaceDimension < 2)
			throw new InvalidSubspaceDimensionException();
		this.init(pTermByDocumentMatrix, pReducedSpaceDimension, pWorkingFolder, LatentSemanticIndexing.REDUCE_THEN_APPLY_FEEDBACKS, pForceRebuilding);
	}

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * valore di somiglianza tra due documenti  calcolato attravero
	 * pSimilarityPerformer
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pVectorSimilarityPerformer
	 *        La strategia di calcolo del valore di somiglianza tra due
	 *        documenti
	 * @param pWorkingFolder
	 *        La directory in cui verr memorizzata la decomposizione ai valori
	 *        singolari della matrice termini-documenti
	 * @param pReducedSpaceDimension
	 *        La dimensione del sottospazio
	 * @param pFeedbackStrategy
	 *        La trategia da adottare per l'applicazione dei feedback. E'
	 *        possibile scegliera tra
	 *        <code>LatentSemanticIndexing.APPLY_FEEDBACKS_THEN_REDUCE</code>
	 *        e <code>LatentSemanticIndexing.REDUCE_THEN_APPLY_FEEDBACKS</code>
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 * @throws InvalidSubspaceDimensionException
	 *         Generata quando si sceglie una dimensione per il sottospazio
	 *         latente non valida
	 */
	public LatentSemanticIndexing(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, IVectorSimilarityPerformer pVectorSimilarityPerformer, String pWorkingFolder, int pReducedSpaceDimension, int pFeedbackStrategy, boolean pForceRebuilding) throws InvalidSubspaceDimensionException {
		super(pTermByDocumentMatrix, pVectorSimilarityPerformer);
		if (pReducedSpaceDimension < 2)
			throw new InvalidSubspaceDimensionException();
		this.init(pTermByDocumentMatrix, pReducedSpaceDimension, pWorkingFolder, pFeedbackStrategy, pForceRebuilding);
	}

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * coseno dell'angolo compreso tra i due vettori che rappresentano due
	 * documenti  usato come valore di somiglianza tra due documenti
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pWorkingFolder
	 *        La directory in cui verr memorizzata la decomposizione ai valori
	 *        singolari della matrice termini-documenti
	 * @param pReducedSpaceDimension
	 *        La dimensione del sottospazio
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 * @throws InvalidSubspaceDimensionException
	 *         Generata quando si sceglie una dimensione per il sottospazio
	 *         latente non valida
	 */
	public LatentSemanticIndexing(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, String pWorkingFolder, int pReducedSpaceDimension, boolean pForceRebuilding) throws InvalidSubspaceDimensionException {
		super(pTermByDocumentMatrix);
		if (pReducedSpaceDimension < 2)
			throw new InvalidSubspaceDimensionException();
		this.init(pTermByDocumentMatrix, pReducedSpaceDimension, pWorkingFolder, LatentSemanticIndexing.REDUCE_THEN_APPLY_FEEDBACKS, pForceRebuilding);
	}

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * coseno dell'angolo compreso tra i due vettori che rappresentano due
	 * documenti  usato come valore di somiglianza tra due documenti
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pWorkingFolder
	 *        La directory in cui verr memorizzata la decomposizione ai valori
	 *        singolari della matrice termini-documenti
	 * @param pReducedSpaceDimension
	 *        La dimensione del sottospazio
	 * @param pFeedbackStrategy
	 *        La trategia da adottare per l'applicazione dei feedback. E'
	 *        possibile scegliera tra
	 *        <code>LatentSemanticIndexing.APPLY_FEEDBACKS_THEN_REDUCE</code>
	 *        e <code>LatentSemanticIndexing.REDUCE_THEN_APPLY_FEEDBACKS</code>
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 * @throws InvalidSubspaceDimensionException
	 *         Generata quando si sceglie una dimensione per il sottospazio
	 *         latente non valida
	 */
	public LatentSemanticIndexing(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, String pWorkingFolder, int pReducedSpaceDimension, int pFeedbackStrategy, boolean pForceRebuilding) throws InvalidSubspaceDimensionException {
		super(pTermByDocumentMatrix);
		if (pReducedSpaceDimension < 2)
			throw new InvalidSubspaceDimensionException();
		this.init(pTermByDocumentMatrix, pReducedSpaceDimension, pWorkingFolder, pFeedbackStrategy, pForceRebuilding);
	}

	/**
	 * @see org.sesalab.IR.model.VSM.VectorSpaceModel#getRankedList(java.util.Collection,
	 *      java.util.Collection, java.lang.Double)
	 */
	public IRankedList getRankedList(Collection<DocumentType> pQueries, Collection<DocumentType> pTargets, Double pThreshold) throws InvalidQueryException {
		if (!(this.termByDocumentMatrix.containsAll(pQueries)))
			throw new InvalidQueryException();

		IRankedList rankedList = new BasicRankedList(this.rankedListCutter);

		// Si decompone la matrice dei termini-documenti
		ISparseMatrix matrix = this.termByDocumentMatrix.getWeightedMatrix();
		IMatrix reducedMatrix = this.getReducedDocumentMatrix(matrix);

		// Si costruisce la ranked list
		rankedList = this.buildRankedList(reducedMatrix, pQueries, pTargets, pThreshold);

		return rankedList;
	}

	/**
	 * @see org.sesalab.IR.model.VSM.VectorSpaceModel#getRankedList(java.util.Collection,
	 *      java.lang.Double)
	 */
	public IRankedList getRankedList(Collection<DocumentType> pQueries, Double pThreshold) throws InvalidQueryException {
		return this.getRankedList(pQueries, null, pThreshold);
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#getRankedList(java.util.Map,
	 *      java.util.Collection, java.lang.Double)
	 */
	public IRankedList getRankedList(Map<DocumentType, Feedbacks<DocumentType>> pUserFeedbacks, Collection<DocumentType> pTargets, Double pThreshold) throws InvalidQueryException {
		if (!(this.termByDocumentMatrix.containsAll(pUserFeedbacks.keySet())))
			throw new InvalidQueryException();

		IRankedList rankedList = new BasicRankedList(this.rankedListCutter);
		try {
			// Si decompone la matrice dei termini-documenti
			ISparseMatrix matrix = this.termByDocumentMatrix.getWeightedMatrix();
			
			// Si determinano le colonne relative ai feedback forniti
			Map<Integer, Feedbacks<Integer>> columnUserFeedbacks = new HashMap<Integer, Feedbacks<Integer>>();
			Iterator<DocumentType> queryIterator = pUserFeedbacks.keySet().iterator();
			while (queryIterator.hasNext()) {
				DocumentType query = queryIterator.next();
				Feedbacks<DocumentType> feedbacks = pUserFeedbacks.get(query);

				// Definiamo i feedback positivi
				List<Integer> correctDocumentsColumnIndex = new LinkedList<Integer>();
				Iterator<DocumentType> correctDocumentsIterator = feedbacks.getCorrectDocuments().iterator();
				while (correctDocumentsIterator.hasNext()) {
					DocumentType correctDocument = correctDocumentsIterator.next();
					correctDocumentsColumnIndex.add(this.termByDocumentMatrix.getColumn(correctDocument));
				}

				// Definiamo i feedback negative
				List<Integer> falsePositivesColumnIndex = new LinkedList<Integer>();
				Iterator<DocumentType> falsePositivesIterator = feedbacks.getFalsePositiveDocuments().iterator();
				while (falsePositivesIterator.hasNext()) {
					DocumentType falsePositive = falsePositivesIterator.next();
					falsePositivesColumnIndex.add(this.termByDocumentMatrix.getColumn(falsePositive));
				}

				Integer queryColumnIndex = new Integer(this.termByDocumentMatrix.getColumn(query));
				Feedbacks<Integer> columnFeedbacks = new Feedbacks<Integer>();
				columnFeedbacks.setCorrectDocuments(correctDocumentsColumnIndex);
				columnFeedbacks.setFalsePositiveDocuments(falsePositivesColumnIndex);

				columnUserFeedbacks.put(queryColumnIndex, columnFeedbacks);
			}

			// Si applicano i feedback
			IMatrix reducedMatrix = null;
			if (this.feedbackStrategy == LatentSemanticIndexing.APPLY_FEEDBACKS_THEN_REDUCE) {
				this.vectorFeedbackPerformer.apply(matrix, columnUserFeedbacks);
				reducedMatrix = this.getReducedDocumentMatrix(matrix);
			} else {
				reducedMatrix = this.getReducedDocumentMatrix(matrix); 
				this.vectorFeedbackPerformer.apply(reducedMatrix, columnUserFeedbacks);
			}

			// Si ottiene la ranked list
			rankedList = this.buildRankedList(reducedMatrix, pUserFeedbacks.keySet(), pTargets, pThreshold);
		} catch (IncorrectVectorSpaceException e) {
			// Eccezione che non si verifica mai
			e.printStackTrace();
		} catch (ObjectNotFoundException e) {
			// Eccezione che non si verifica mai
			e.printStackTrace();
		}

		return rankedList;
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#getRankedList(java.util.Map,
	 *      java.lang.Double)
	 */
	public IRankedList getRankedList(Map<DocumentType, Feedbacks<DocumentType>> pUserFeedbacks, Double pThreshold) throws InvalidQueryException {
		return this.getRankedList(pUserFeedbacks, null, pThreshold);
	}

	/*
	 * Determina la matrice ridotta che fornisce la relazione tra documenti
	 * 
	 * @param pMatrix
	 *        La matrice termini documenti da decomporre
	 * @return La matrice dei documenti
	 */
	private IMatrix getReducedDocumentMatrix(ISparseMatrix pMatrix) {
		if (this.forceRebuilding)
			this.svdManager.setMatrix(pMatrix);
		SingularValueDecomposition singularValueDecomposition = this.svdManager.getSingularValueDecomposition(this.forceRebuilding);
		int reducedSpaceDimension = singularValueDecomposition.rank();
		if (this.subspaceDimension < reducedSpaceDimension)
			reducedSpaceDimension = this.subspaceDimension;

		IMatrix reducedMatrix = null;

		try {
			// Matrice dei concetti
			IMatrix conceptMatrix = new BasicMatrix(singularValueDecomposition.getS().getRowDimension(), singularValueDecomposition.getS().getColumnDimension(), singularValueDecomposition.getS().getArrayCopy());
			IMatrix reducedConceptMatrix = conceptMatrix.subMatrix(reducedSpaceDimension, reducedSpaceDimension);

			// Matrice dei documenti
			IMatrix documentMatrix = new BasicMatrix(singularValueDecomposition.getV().getRowDimension(), singularValueDecomposition.getV().getColumnDimension(), singularValueDecomposition.getV().getArrayCopy());
			IMatrix reducedDocumentMatrix = documentMatrix.subMatrix(documentMatrix.getRowCount(), reducedSpaceDimension);

			// Si ottiene la relazione tra documenti
			reducedMatrix = reducedDocumentMatrix.mul(reducedConceptMatrix).transpose();
		} catch (InvalidIndexException e) {
			// Eccezione che non si verifica mai
			e.printStackTrace();
		} catch (IncorrectDimensionException e) {
			// Eccezione che non si verifica mai
			e.printStackTrace();
		}

		return reducedMatrix;
	}

	/*
	 * Inizializza le variabili di istanza della classe
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice termini-documenti
	 * @param pSubspaceDimension
	 *        La dimesnione del sottospazio latente
	 * @param pWorkingFolder
	 *        La directory di lavoro
	 * @param pFeedbackStrategy
	 *        La strategia di feedback da adottare
	 * @param pForceRebuilding
	 *        Indica se forzare il re-build della decomposizione ai valori
	 *        singolari ogni volta che si esegue una query
	 */
	private void init(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, int pSubspaceDimension, String pWorkingFolder, int pFeedbackStrategy, boolean pForceRebuilding) {
		this.subspaceDimension = pSubspaceDimension;
		this.workingFolder = pWorkingFolder;
		this.feedbackStrategy = pFeedbackStrategy;
		this.svdManager = new SSVDManager(pTermByDocumentMatrix.getWeightedMatrix(), pWorkingFolder);
		this.forceRebuilding = pForceRebuilding;
	}

	// La strategia da adottare per l'applicazione dei feedback
	private int feedbackStrategy;

	// Determina se effettuare il re-building della decomposizione
	private boolean forceRebuilding;

	// Dimensione del sottospazio latente
	private int subspaceDimension;

	// Il manager per la decomposizione ai valori singolari
	private SSVDManager svdManager;

	// Directory di lavoro
	private String workingFolder;
}