/*
 *  VectorSpaceModel.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.model.VSM;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import org.sesalab.IR.documentSpace.ITermByDocumentMatrix;
import org.sesalab.IR.documentSpace.exception.ObjectNotFoundException;
import org.sesalab.IR.matrix.IMatrix;
import org.sesalab.IR.matrix.ISparseMatrix;
import org.sesalab.IR.matrix.exception.InvalidIndexException;
import org.sesalab.IR.model.Feedbacks;
import org.sesalab.IR.model.IIRModel;
import org.sesalab.IR.model.exception.IncorrectVectorSpaceException;
import org.sesalab.IR.model.exception.InvalidQueryException;
import org.sesalab.IR.model.vectorFeedbackPerformer.IVectorFeedbackPerformer;
import org.sesalab.IR.model.vectorSimilarityPerformer.CosineSimilarity;
import org.sesalab.IR.model.vectorSimilarityPerformer.IVectorSimilarityPerformer;
import org.sesalab.IR.rankedList.BasicRankedList;
import org.sesalab.IR.rankedList.IRankedList;
import org.sesalab.IR.rankedList.RankedListElement;
import org.sesalab.IR.rankedList.cutter.ConstantCutter;
import org.sesalab.IR.rankedList.cutter.ICutter;

/**
 * Classe che implementa il modello di Information Retrieval denominato Vector
 * Space Model
 * 
 * @author Paola Sgueglia
 * @param <DocumentType>
 *        Tipo dei documenti indicizzati
 * @param <TermType>
 *        Tipo dei termini indicizzati
 */
public class VectorSpaceModel<DocumentType extends Comparable, TermType> implements IIRModel<DocumentType, TermType> {

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * coseno dell'angolo compreso tra i due vettori che rappresentano due
	 * documenti  usato come valore di somiglianza tra due documenti
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 */
	public VectorSpaceModel(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix) {
		this.init(pTermByDocumentMatrix, new CosineSimilarity(), null);
	}

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * coseno dell'angolo compreso tra i due vettori che rappresentano due
	 * documenti  usato come valore di somiglianza tra due documenti. Il metodo
	 * prevede l'applicazione di feedback forniti dall'utente
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pVectorFeedbackPerformer
	 *        La strategia di applicazione dei feedbakc
	 */
	public VectorSpaceModel(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, IVectorFeedbackPerformer<DocumentType, TermType> pVectorFeedbackPerformer) {
		this.init(pTermByDocumentMatrix, new CosineSimilarity(), pVectorFeedbackPerformer);
	}

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * valore di somiglianza tra due documenti  calcolato attravero
	 * pSimilarityPerformer
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pVectorSimilarityPerformer
	 *        La strategia di calcolo del valore di somiglianza tra due
	 *        documenti
	 */
	public VectorSpaceModel(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, IVectorSimilarityPerformer pVectorSimilarityPerformer) {
		this.init(pTermByDocumentMatrix, pVectorSimilarityPerformer, null);
	}

	/**
	 * Crea una nuova istanza a partire dalla matrice dei termini-documenti. Il
	 * valore di somiglianza tra due documenti  calcolato attravero
	 * pSimilarityPerformer. Il metodo prevede l'applicazione di feedback forniti 
	 * dall'utente
	 * 
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pVectorSimilarityPerformer
	 *        La strategia di calcolo del valore di somiglianza tra due
	 *        documenti
	 * @param pVectorFeedbackPerformer
	 *        La strategia di applicazione dei feedback
	 */
	public VectorSpaceModel(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, IVectorSimilarityPerformer pVectorSimilarityPerformer, IVectorFeedbackPerformer<DocumentType, TermType> pVectorFeedbackPerformer) {
		this.init(pTermByDocumentMatrix, pVectorSimilarityPerformer, pVectorFeedbackPerformer);
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#getFeedbacksPerformer()
	 */
	public IVectorFeedbackPerformer<DocumentType, TermType> getFeedbacksPerformer() {
		return this.vectorFeedbackPerformer;
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#getRankedList(java.util.Collection, java.util.Collection, java.lang.Double)
	 */
	public IRankedList getRankedList(Collection<DocumentType> pQueries, Collection<DocumentType> pTargets, Double pThreshold) throws InvalidQueryException {
		if (!(this.termByDocumentMatrix.containsAll(pQueries)))
			throw new InvalidQueryException();

		// Si costruisce la ranked list
		ISparseMatrix matrix = this.termByDocumentMatrix.getWeightedMatrix();
		IRankedList rankedList = this.buildRankedList(matrix, pQueries, pTargets, pThreshold.doubleValue());
		return rankedList;
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#getRankedList(java.util.Collection, java.lang.Double)
	 */
	public IRankedList getRankedList(Collection<DocumentType> pQueries, Double pThreshold) throws InvalidQueryException {
		return this.getRankedList(pQueries, null, pThreshold);
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#getRankedList(java.util.Map, java.util.Collection, java.lang.Double)
	 */
	public IRankedList getRankedList(Map<DocumentType, Feedbacks<DocumentType>> pUserFeedbacks, Collection<DocumentType> pTargets, Double pThreshold) throws InvalidQueryException {
		if (!(this.termByDocumentMatrix.containsAll(pUserFeedbacks.keySet())))
			throw new InvalidQueryException();

		IRankedList rankedList = new BasicRankedList(this.rankedListCutter);
		try {
			// Si applicano i feedback
			ISparseMatrix reweightedMatrix = this.vectorFeedbackPerformer.apply(this.termByDocumentMatrix, pUserFeedbacks);
			rankedList = this.buildRankedList(reweightedMatrix, pUserFeedbacks.keySet(), pTargets, pThreshold);
		} catch (IncorrectVectorSpaceException e) {
			// Eccezione che non si verifica mai
			e.printStackTrace();
		}
		
		return rankedList;
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#getRankedList(java.util.Map, java.lang.Double)
	 */
	public IRankedList getRankedList(Map<DocumentType, Feedbacks<DocumentType>> pUserFeedbacks, Double pThreshold) throws InvalidQueryException {
		return this.getRankedList(pUserFeedbacks, null, pThreshold);
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#setFeedbacksPerformer(org.sesalab.IR.model.vectorFeedbackPerformer.IVectorFeedbackPerformer)
	 */
	public void setFeedbacksPerformer(IVectorFeedbackPerformer<DocumentType, TermType> pVectorFeedbackPerformer) {
		this.vectorFeedbackPerformer = pVectorFeedbackPerformer;
	}

	/**
	 * @see org.sesalab.IR.model.IIRModel#setRankedListCutter(org.sesalab.IR.rankedList.cutter.ICutter)
	 */
	public void setRankedListCutter(ICutter pCutter) {
		this.rankedListCutter = pCutter;
	}

	/*
	 * Crea un elemento della ranked list
	 * @param pSource
	 *        Documento source
	 * @param pTarget
	 *        Documento target
	 * @param pSimilarityValue
	 *        Valore di somiglianza
	 * @return Un nuovo elemento della ranked list
	 */
	private RankedListElement<DocumentType> createRankedListElement(DocumentType pSource, DocumentType pTarget, double pSimilarityValue) {
		RankedListElement<DocumentType> element = new RankedListElement<DocumentType>();
		element.setSource(pSource);
		element.setTarget(pTarget);
		element.setSimilarityValue(pSimilarityValue);
		return element;
	}

	/*
	 * Inizializza gli attributi della classe
	 * @param pTermByDocumentMatrix
	 *        La matrice dei termini documenti
	 * @param pVectorSimilarityPerformer
	 *        La strategia di calcolo del valore di somiglianza tra due
	 *        documenti
	 * @param pVectorFeedbackPerformer
	 *        La strategia di applicazione dei feedback
	 */
	private void init(ITermByDocumentMatrix<DocumentType, TermType> pTermByDocumentMatrix, IVectorSimilarityPerformer pVectorSimilarityPerformer, IVectorFeedbackPerformer<DocumentType, TermType> pVectorFeedbackPerformer) {
		this.termByDocumentMatrix = pTermByDocumentMatrix;
		this.vectorSimilarityPerformer = pVectorSimilarityPerformer;
		this.vectorFeedbackPerformer = pVectorFeedbackPerformer;
		this.rankedListCutter = new ConstantCutter();
	}
	
	/*
	 * Costrisce la ranked list a partire dalla matrice termini-documenti
	 * @param pMatrix
	 *        La matrice termini-documenti
	 * @param pDocument
	 *        Il documento usato come query
	 * @param pTargets
	 *        I documenti da usare come target
	 * @param pThreshold
	 *        La soglia di taglio
	 * @return La ranked list
	 */
	@SuppressWarnings("unchecked")
	protected IRankedList buildRankedList(IMatrix pMatrix, Collection<DocumentType> pQueries, Collection<DocumentType> pTargets, double pThreshold) {
		// Si costruisce la ranked list
		IRankedList rankedList = new BasicRankedList(this.rankedListCutter);
		try {
			Iterator<DocumentType> queriesIterator = pQueries.iterator();
			while (queriesIterator.hasNext()) {
				DocumentType document = queriesIterator.next();
				int documentColumn = this.termByDocumentMatrix.getColumn(document);
				double[] query = pMatrix.getColumn(documentColumn);
				for (int i = 1; i < documentColumn; i++) {
					DocumentType targetDocument = this.termByDocumentMatrix.getDocument(i);
					if ((pTargets == null) || (pTargets.contains(targetDocument))) {
						double similarityValue = this.vectorSimilarityPerformer.calculateSimilarity(query, pMatrix.getColumn(i));
						RankedListElement rankedListElement = null;
						if (document.compareTo(targetDocument) > 0)
							rankedListElement = this.createRankedListElement(targetDocument, document, similarityValue);
						else
							rankedListElement = this.createRankedListElement(document, targetDocument, similarityValue);
						if (!rankedList.contains(rankedListElement))
							rankedList.addElement(rankedListElement);
					}
				}
				for (int i = documentColumn + 1; i <= this.termByDocumentMatrix.getDocumentCount(); i++) {
					DocumentType targetDocument = this.termByDocumentMatrix.getDocument(i);
					if ((pTargets == null) || (pTargets.contains(targetDocument))) {
						double similarityValue = this.vectorSimilarityPerformer.calculateSimilarity(query, pMatrix.getColumn(i));
						RankedListElement rankedListElement = null;
						if (document.compareTo(targetDocument) > 0)
							rankedListElement = this.createRankedListElement(targetDocument, document, similarityValue);
						else
							rankedListElement = this.createRankedListElement(document, targetDocument, similarityValue);
						if (!rankedList.contains(rankedListElement))
							rankedList.addElement(rankedListElement);
					}
				}
			}
		} catch (InvalidIndexException e) {
			// Eccezione che non si pu verificare
		} catch (ObjectNotFoundException e) {
			// Eccezione che non si pu verificare
		} catch (IncorrectVectorSpaceException e) {
			// Eccezione che non si pu verificare
		}
		return rankedList.cut(pThreshold);
	}

	// Strategia di applicazione dei feedback
	protected IVectorFeedbackPerformer<DocumentType, TermType> vectorFeedbackPerformer;

	// Strategia di calcolo del valore di somiglianza tra due documenti
	private IVectorSimilarityPerformer vectorSimilarityPerformer;

	// L'algoritmo per il taglio della ranked list
	protected ICutter rankedListCutter;

	// Matrice dei termini documenti
	protected ITermByDocumentMatrix<DocumentType, TermType> termByDocumentMatrix;
}