/*
 *  AdvanceMath.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.model.util;

import java.text.NumberFormat;

import org.sesalab.IR.model.exception.IncorrectVectorSpaceException;

/**
 * Classe che fa da contenitore ad una serie di routine matematiche avanzante
 * necessarie per l'implementazione del metodo di Lanczos.
 */

public class AdvanceMath {

	/**
	 * Copia due vettori o una parte di due vettori
	 * 
	 * @param length
	 *        Numero di elementi da copiare
	 * @param source
	 *        Vettore origine
	 * @param indSource
	 *        Indice di inizio vettore origine
	 * @param incSource
	 *        Incremento dell'indice di scorrimento del vettore origine
	 * @param target
	 *        Vettore destinazione
	 * @param indTarget
	 *        Indice di inizio vettore destinazione
	 * @param incTarget
	 *        Incremento dell'indice di scorrimento del vettore destinazione
	 */
	public static void advanceCopy(long length, double[] source, long indSource, long incSource, double[] target, long indTarget, long incTarget) {

		long i, indexSource = indSource, indexTarget = indTarget;

		// Controllo parametri
		if ((length <= 0) || (incSource == 0) || (incTarget == 0))
			return;

		// Si determina l'indice di inizio nel caso l'incremento sia negativo
		if (incSource < 0)
			indexSource += (-length + 1) * incSource;
		if (incTarget < 0)
			indexTarget += (-length + 1) * incTarget;

		// Si procede alla copia
		for (i = 0; i < length; i++) {
			target[(int) indexTarget] = source[(int) indexSource];
			indexSource += incSource;
			indexTarget += incTarget;
		}
	}

	/**
	 * Approssima un double ad un intero.
	 * 
	 * @param x
	 *        Valore da approssimare
	 * @return Il valore approssimato
	 */

	public static int approx(double x) {
		double dec;
		int res;

		res = (int) x;
		dec = x - res;

		if (dec >= 0.5)
			res++;

		return (res);

	}
	
	/**
	 * Approssima un double ad una data cifra decimale 
	 * @param pValue
	 * 		  Il valore da approssimare
	 * @param pMaximumFractionDigits
	 * 		  Il numero massimo di cifre decimali
	 * @return il valore approssimato
	 */
	public static double approx(double pValue, int pMaximumFractionDigits) {
		NumberFormat numberFormat = NumberFormat.getNumberInstance();
		numberFormat.setMaximumFractionDigits(pMaximumFractionDigits);
		String formattedApproxedValue = numberFormat.format(pValue).replace(",", ".");
		return Double.valueOf(formattedApproxedValue);
	}

	/**
	 * Calcola il coseno dell'angolo compreso tra due vettori.
	 * 
	 * @param x
	 *        Primo vettore
	 * @param y
	 *        Secondo vettore
	 * @return Il coseno dell'angolo compreso tra i due vettori
	 * @throws IncorrectVectorSpaceException
	 *         Generata quando i vettori non appartengono allo stesso spazio
	 */

	public static double calculateCosine(double[] x, double[] y) throws IncorrectVectorSpaceException {
		double sumX2 = 0, sumY2 = 0, sumXY = 0, den;

		// Condizione di errore: i due vettori non appartengono allo stesso
		// spazio
		if (x.length != y.length)
			throw new IncorrectVectorSpaceException("Vector dimensions mast agree!");

		for (int i = 0; i < x.length; i++) {
			sumX2 += Math.pow(x[i], 2);
			sumY2 += Math.pow(y[i], 2);
			sumXY += x[i] * y[i];
		}

		den = Math.sqrt(sumX2 * sumY2);
		if (den > 0)
			return (sumXY / den);
		return (0);
	}

	/**
	 * Calcola il prodotto scalare tra due vettori o tra due parti di vettori *
	 * 
	 * @param length
	 *        Numero di elementi da copiare
	 * @param x
	 *        Primo vettore
	 * @param startx
	 *        Indice di inizio del primo vettore
	 * @param incx
	 *        Incremento dell'indice di scorrimento del primo vettore
	 * @param y
	 *        Secondo vettore
	 * @param starty
	 *        Indice di inizio del secondo vettore
	 * @param incy
	 *        Incremento dell'indice di scorrimento del secondo vettore
	 * @return Il prodotto scalare dei due vettori
	 */

	public static double dotProduct(long length, double[] x, long startx, long incx, double[] y, long starty, long incy) {

		long i, indx = startx, indy = starty;
		double res = 0.0;

		// Controllo parametri
		if ((length <= 0) || (incx == 0) || (incy == 0))
			return (res);

		// Si determina l'indice di inizio nel caso l'incremento sia negativo
		if (incx < 0)
			indx += (-length + 1) * incx;
		if (incy < 0)
			indy += (-length + 1) * incy;

		// Si procede al calcolo del prodotto scalare
		for (i = 0; i < length; i++) {
			res += x[(int) indx] * y[(int) indy];
			indx += incx;
			indy += incy;
		}

		return (res);
	}

	/**
	 * Ordina simultaneamente due vettori x, y. L'ordinamento seguito  quello
	 * decrescente dato in x
	 * 
	 * @param indexGap
	 *        Indice inizio ordinamento
	 * @param length
	 *        Numero elementi da ordinare
	 * @param x
	 *        Primo vettore (double)
	 * @param y
	 *        Secondo vettore (double)
	 */

	public static void dualSort(long indexGap, long length, double[] x, double[] y) {
		double tmp;
		long i, j, index;

		if (indexGap == 0)
			return;
		for (i = indexGap; i < length; i++) {
			j = i - indexGap;
			index = i;
			while ((j >= 0) && (x[(int) j] > x[(int) index])) {
				tmp = x[(int) j];
				x[(int) j] = x[(int) index];
				x[(int) index] = tmp;
				tmp = y[(int) j];
				y[(int) j] = y[(int) index];
				y[(int) index] = tmp;
				j -= indexGap;
				index = j + indexGap;
			}
		}

		// Chiamata ricorsiva a dualSort
		dualSort(indexGap / 2, length, x, y);
	}

	/**
	 * Ordina simultaneamente due vettori x, y. L'ordinamento seguito  quello
	 * decrescente dato in x
	 * 
	 * @param indexGap
	 *        Indice inizio ordinamento
	 * @param length
	 *        Numero elementi da ordinare
	 * @param x
	 *        Primo vettore (double)
	 * @param y
	 *        Secondo vettore (String)
	 */

	public static void dualSort(long indexGap, long length, double[] x, String[] y) {
		double tmp;
		String app;
		long i, j, index;

		if (indexGap == 0)
			return;
		for (i = indexGap; i < length; i++) {
			j = i - indexGap;
			index = i;
			while ((j >= 0) && (x[(int) j] > x[(int) index])) {
				tmp = x[(int) j];
				x[(int) j] = x[(int) index];
				x[(int) index] = tmp;
				app = y[(int) j];
				y[(int) j] = y[(int) index];
				y[(int) index] = app;
				j -= indexGap;
				index = j + indexGap;
			}
		}

		// Chiamata ricorsiva a dualSort
		dualSort(indexGap / 2, length, x, y);
	}

	/**
	 * Restituisce a se a e b sono concordi, -a altrimenti.
	 * 
	 * @param a
	 *        Primo termine
	 * @param b
	 *        Secondo termine
	 * @return <code>a</code> se <code>a</code> e <code>b</code> sono
	 *         concordi, <code>-a</code> altrimenti
	 */

	public static double fsign(double a, double b) {

		if (((a >= 0.0) && (b >= 0.0)) || ((a < 0.0) && (b < 0.0)))
			return (a);
		return (-a);

	}

	/**
	 * Determina la precisione di macchina.
	 * 
	 * @return La precisione di macchina
	 */

	public static double getMachinePrecision() {
		double eeps;

		eeps = 1;
		// eeps dovr essere il pi piccolo valore che sommato ad 1 da ancora 1
		while ((1 + eeps) != 1)
			eeps = eeps / 2;

		return (eeps);
	}

	/**
	 * Ottiene l'indice del massimo valore in valore assoluto.
	 * 
	 * @param length
	 *        Numero di elementi su cui lavorare
	 * @param x
	 *        Vettore
	 * @param ind
	 *        Indice di inizio
	 * @param inc
	 *        Incremento indice di scorrimento del vettore
	 * @return L'indice del massimo valore in valore assoluto
	 */

	public static long indexMaxAbs(long length, double[] x, long ind, long inc) {
		long ix, i, indMax, index = ind;
		double tmp, max;

		// Casi banali
		if (length < 1)
			return (-1);
		if (length == 1)
			return (0);
		if (inc == 0)
			return (-1);

		// Si determina l'indice di inizio nel caso l'incremento sia negativo
		if (inc < 0)
			ix = (-length + 1) * inc;
		else
			ix = 0;

		// Si procede alla ricerca del massimo
		indMax = ix;
		index += ix;
		max = Math.abs(x[(int) index]);
		for (i = 1; i < length; i++) {
			ix += inc;
			index += inc;
			tmp = Math.abs(x[(int) index]);
			if (tmp > max) {
				max = tmp;
				indMax = ix;
			}
		}
		return (indMax);
	}

	/**
	 * Calcola il logaritmo in base 10 di x.
	 * 
	 * @param x
	 *        Argomento del logaritmo
	 * @return Il logaritmo di <code>x</code>
	 */

	public static double log10(double x) {
		double val1, val2;

		val1 = Math.log(x);
		val2 = Math.log(10);
		return (val1 / val2);
	}

	/**
	 * Calcola il logaritmo in base 2 di x.
	 * 
	 * @param x
	 *        Argomento del logaritmo
	 * @return Il logaritmo di <code>x</code>
	 */

	public static double log2(double x) {
		double val1, val2;

		val1 = Math.log(x);
		val2 = Math.log(2);
		return (val1 / val2);
	}

	/**
	 * Calcola la radice quadrata di a^2 + b^2 senza incorrerre in problemi di
	 * overflow
	 * 
	 * @param a
	 *        Primo termine
	 * @param b
	 *        Secondo termine
	 * @return La radice quadrata di <code>a^2 + b^2</code>
	 */

	public static double pythag(double a, double b) {
		double p, r, s, t, u, temp;

		p = Math.max(Math.abs(a), Math.abs(b));
		if (p != 0.0) {
			temp = Math.min(Math.abs(a), Math.abs(b)) / p;
			r = temp * temp;
			t = 4.0 + r;
			while (t != 4.0) {
				s = r / t;
				u = 1.0 + 2.0 * s;
				p *= u;
				temp = s / u;
				r *= temp * temp;
				t = 4.0 + r;
			}
		}

		return (p);
	}

	/**
	 * Scala un vettore x di un parametro k.
	 * 
	 * @param length
	 *        Numero elementi da scalare
	 * @param k
	 *        Parametro di scalatura
	 * @param x
	 *        Vettore da scalare
	 * @param inc
	 *        Incremento dell'indice di scorrimento del vettore
	 */

	public static void scales(long length, double k, double[] x, long inc) {
		long i, index = 0;

		// Caso banale
		if ((length <= 0) || (inc == 0))
			return;

		// Si determina l'indice di inizio nel caso l'incremento sia negativo
		if (inc < 0)
			index += (-length + 1) * inc;

		// Si procede alla scalatura
		for (i = 0; i < length; i++) {
			x[(int) index] *= k;
			index += inc;
		}
	}

	/**
	 * Scala un vettore x di un parametro k e mette il suo contenuto nel vettore
	 * y
	 * 
	 * @param length
	 *        Numero elementi da scalare
	 * @param k
	 *        Parametro di scalatura
	 * @param x
	 *        Vettore da scalare
	 * @param incx
	 *        Incremento dell'indice di scorrimento del vettore da scalare
	 * @param y
	 *        Vettore output
	 * @param incy
	 *        Incremento dell'indice di scorrimento del vettore output
	 */

	public static void scales(long length, double k, double[] x, long incx, double[] y, long incy) {
		long i, indx = 0, indy = 0;

		// Caso banale
		if ((length <= 0) || (incx == 0) || (incy == 0) || (k == 0.0))
			return;

		// Si determina l'indice di inizio nel caso l'incremento sia negativo
		if (incx < 0)
			indx += (-length + 1) * incx;
		if (incy < 0)
			indy += (-length + 1) * incy;

		// Si procede alla scalatura
		for (i = 0; i < length; i++) {
			y[(int) indy] = k * x[(int) indx];
			indx += incx;
			indy += incy;
		}
	}

	/**
	 * Ottiene un vettore y sommandogli il vettore x scalato di un parametro k
	 * 
	 * @param length
	 *        Numero elementi da scalare
	 * @param k
	 *        Parametro di scalatura
	 * @param x
	 *        Vettore da scalare
	 * @param indx
	 *        Indice di inizio del vettore da scalare
	 * @param incx
	 *        Incremento dell'indice di scorrimento del vettore da scalare
	 * @param y
	 *        Vettore output
	 * @param indy
	 *        Indice di inizio del vettore output
	 * @param incy
	 *        Incremento dell'indice di scorrimento del vettore output
	 */

	public static void scalesAndSum(long length, double k, double[] x, long indx, long incx, double[] y, long indy, long incy) {
		long i, ix = indx, iy = indy;

		// Caso banale
		if ((length <= 0) || (incx == 0) || (incy == 0) || (k == 0.0))
			return;

		// Si determina l'indice di inizio nel caso l'incremento sia negativo
		if (incx < 0)
			ix += (-length + 1) * incx;
		if (incy < 0)
			iy += (-length + 1) * incy;

		// Si procede alla scalatura
		for (i = 0; i < length; i++) {
			y[(int) iy] += k * x[(int) ix];
			ix += incx;
			iy += incy;
		}
	}

	/**
	 * Scambia gli elementi di due vettori o di una parte di due vettori
	 * 
	 * @param length
	 *        Numero elementi da scambiare
	 * @param x
	 *        Primo vettore
	 * @param incx
	 *        Incremento dell'indice di scorrimento del primo vettore
	 * @param y
	 *        Secondo vettore
	 * @param incy
	 *        Incremento dell'indice di scorrimento del secondo vettore
	 */

	public static void swap(long length, double[] x, long incx, double[] y, long incy) {
		long i, indx = 0, indy = 0;
		double tmp;

		// Caso banale
		if ((length <= 0) || (incx == 0) || (incy == 0))
			return;

		// Si determina l'indice di inizio nel caso l'incremento sia negativo
		if (incx < 0)
			indx += (-length + 1) * incx;
		if (incy < 0)
			indy += (-length + 1) * incy;

		// Si procede allo swapping
		for (i = 0; i < length; i++) {
			tmp = y[(int) indy];
			y[(int) indy] = x[(int) indx];
			x[(int) indx] = tmp;
			indy += incy;
			indx += incx;
		}
	}
	
	/**
	 * Consente di sommare due vettori
	 * @param pFirst Il primo vettore
	 * @param pSecond Il secondo vettore
	 * @return Il vettore somma
	 */
	public static double[] sum(double[] pFirst, double[] pSecond) {
		double[] sum = new double[pFirst.length];
		
		for (int i = 0; i < sum.length; i++)
			sum[i] = pFirst[i] + pSecond[i];
		
		return sum;
	}

	/**
	 * Consente di sottrarre due vettori
	 * @param pFirst Il primo vettore
	 * @param pSecond Il secondo vettore
	 * @return Il vettore differenza
	 */
	public static double[] sub(double[] pFirst, double[] pSecond) {
		double[] sub = new double[pFirst.length];
		
		for (int i = 0; i < sub.length; i++)
			sub[i] = pFirst[i] - pSecond[i];
		
		return sub;
	}
	
	/**
	 * Consente di moltiplicare uno scalare per un vettore
	 * @param pScalar Lo scalare
	 * @param pVector Il vettore
	 * @return Il risultato della moltiplicazione
	 */
	public static double[] mul(double pScalar, double[] pVector) {
		double[] mul = new double[pVector.length];
		
		for (int i = 0; i < mul.length; i++)
			mul[i] = pScalar * pVector[i];
		
		return mul;
	}
}