/*
 *  StandardRochioVectorFeedbackPerformer.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.model.vectorFeedbackPerformer;

import java.util.Iterator;
import java.util.Map;

import org.sesalab.IR.documentSpace.ITermByDocumentMatrix;
import org.sesalab.IR.documentSpace.exception.ObjectNotFoundException;
import org.sesalab.IR.matrix.IMatrix;
import org.sesalab.IR.matrix.ISparseMatrix;
import org.sesalab.IR.matrix.exception.InvalidIndexException;
import org.sesalab.IR.model.Feedbacks;
import org.sesalab.IR.model.exception.IncorrectVectorSpaceException;
import org.sesalab.IR.model.util.AdvanceMath;

/**
 * @author Rocco Oliveto
 * @param <DocumentType> 
 * @param <TermType> 
 *
 */
public class StandardRochioVectorFeedbackPerformer<DocumentType, TermType> implements IVectorFeedbackPerformer<DocumentType, TermType> {
	
	/**
	 * 
	 */
	public StandardRochioVectorFeedbackPerformer() {
		this.init(1, 1, 1);
	}

	/**
	 * @param pScaleQuery
	 * @param pScalePositiveFeedbacks
	 * @param pScaleNegativeFeedbacks
	 */
	public StandardRochioVectorFeedbackPerformer(double pScaleQuery, double pScalePositiveFeedbacks, double pScaleNegativeFeedbacks) {
		this.init(pScaleQuery, pScalePositiveFeedbacks, pScaleNegativeFeedbacks);
	}
	/**
	 * @see org.sesalab.IR.model.vectorFeedbackPerformer.IVectorFeedbackPerformer#apply(org.sesalab.IR.documentSpace.ITermByDocumentMatrix, java.util.Map)
	 */
	public ISparseMatrix apply(ITermByDocumentMatrix<DocumentType, TermType> pMatrix, Map<DocumentType, Feedbacks<DocumentType>> pUserFeedbacks) throws IncorrectVectorSpaceException {
		ISparseMatrix sparseMatrix = pMatrix.getWeightedMatrix();
		try {
			Iterator<DocumentType> queryIterator = pUserFeedbacks.keySet().iterator();
			
			while (queryIterator.hasNext()) {
				DocumentType query = queryIterator.next();
				Feedbacks<DocumentType> feedbacks = pUserFeedbacks.get(query);
				int queryColumnIndex = pMatrix.getColumn(query);
				double[] queryColumn = sparseMatrix.getColumn(queryColumnIndex);
				
				// Definiamo i feedback positivi
				Iterator<DocumentType> correctDocumentsIterator = feedbacks.getCorrectDocuments().iterator(); 
				double[] positiveFeedbacks = new double[queryColumn.length];
				while (correctDocumentsIterator.hasNext()) {
					DocumentType correctDocument = correctDocumentsIterator.next();
					double[] correctDocumentColumn = sparseMatrix.getColumn(pMatrix.getColumn(correctDocument));
					positiveFeedbacks = AdvanceMath.sum(positiveFeedbacks, correctDocumentColumn);
				}
				
				// Definiamo i feedback negativi
				Iterator<DocumentType> falsePositivesDocumentsIterator = feedbacks.getFalsePositiveDocuments().iterator(); 
				double[] negativeFeedbacks = new double[queryColumn.length];
				while (falsePositivesDocumentsIterator.hasNext()) {
					DocumentType falsePositiveDocument = falsePositivesDocumentsIterator.next();
					double[] falsePositiveDocumentColumn = sparseMatrix.getColumn(pMatrix.getColumn(falsePositiveDocument));
					negativeFeedbacks = AdvanceMath.sum(negativeFeedbacks, falsePositiveDocumentColumn);
				}
				
				queryColumn = AdvanceMath.mul(this.scaleQuery, queryColumn);
				double scalePositive = (feedbacks.getCorrectDocuments().size() != 0) ? this.scalePositiveFeedbacks / feedbacks.getCorrectDocuments().size() : 0;
				double scaleNegative = (feedbacks.getFalsePositiveDocuments().size() != 0) ? this.scaleNegativeFeedbacks / feedbacks.getFalsePositiveDocuments().size() : 0;
				positiveFeedbacks = AdvanceMath.mul(scalePositive, positiveFeedbacks);
				negativeFeedbacks = AdvanceMath.mul(scaleNegative, negativeFeedbacks);
				
				queryColumn = AdvanceMath.sum(queryColumn, positiveFeedbacks);
				queryColumn = AdvanceMath.sub(queryColumn, negativeFeedbacks);
				
				for (int i = 0; i < queryColumn.length; i++)
					sparseMatrix.set(i + 1, queryColumnIndex, queryColumn[i]);
			}
		} catch (InvalidIndexException e) {
			// Eccezione che non si verifa mai
			e.printStackTrace();
		} catch (ObjectNotFoundException e) {
			// Eccezione che non si verifa mai
			e.printStackTrace();
		}
		return sparseMatrix;
	}
	
	/**
	 * @see org.sesalab.IR.model.vectorFeedbackPerformer.IVectorFeedbackPerformer#apply(org.sesalab.IR.matrix.IMatrix, java.util.Map)
	 */
	public void apply(IMatrix pMatrix, Map<Integer, Feedbacks<Integer>> pUserFeedbacks) throws IncorrectVectorSpaceException {
		try {
			Iterator<Integer> queryIterator = pUserFeedbacks.keySet().iterator();
			
			while (queryIterator.hasNext()) {
				Integer queryColumnIndex = queryIterator.next();
				Feedbacks<Integer> feedbacks = pUserFeedbacks.get(queryColumnIndex);
				double[] queryColumn = pMatrix.getColumn(queryColumnIndex.intValue());
				
				// Definiamo i feedback positivi
				Iterator<Integer> correctDocumentsIterator = feedbacks.getCorrectDocuments().iterator(); 
				double[] positiveFeedbacks = new double[queryColumn.length];
				while (correctDocumentsIterator.hasNext()) {
					Integer correctDocumentColumnIndex = correctDocumentsIterator.next();
					double[] correctDocumentColumn = pMatrix.getColumn(correctDocumentColumnIndex);
					positiveFeedbacks = AdvanceMath.sum(positiveFeedbacks, correctDocumentColumn);
				}
				
				// Definiamo i feedback negativi
				Iterator<Integer> falsePositivesDocumentsIterator = feedbacks.getFalsePositiveDocuments().iterator(); 
				double[] negativeFeedbacks = new double[queryColumn.length];
				while (falsePositivesDocumentsIterator.hasNext()) {
					Integer falsePositiveDocumentColumnIndex = falsePositivesDocumentsIterator.next();
					double[] falsePositiveDocumentColumn = pMatrix.getColumn(falsePositiveDocumentColumnIndex);
					negativeFeedbacks = AdvanceMath.sum(negativeFeedbacks, falsePositiveDocumentColumn);
				}
				
				queryColumn = AdvanceMath.mul(this.scaleQuery, queryColumn);
				double scalePositive = (feedbacks.getCorrectDocuments().size() != 0) ? this.scalePositiveFeedbacks / feedbacks.getCorrectDocuments().size() : 0;
				double scaleNegative = (feedbacks.getFalsePositiveDocuments().size() != 0) ? this.scaleNegativeFeedbacks / feedbacks.getFalsePositiveDocuments().size() : 0;
				positiveFeedbacks = AdvanceMath.mul(scalePositive, positiveFeedbacks);
				negativeFeedbacks = AdvanceMath.mul(scaleNegative, negativeFeedbacks);
				
				queryColumn = AdvanceMath.sum(queryColumn, positiveFeedbacks);
				queryColumn = AdvanceMath.sub(queryColumn, negativeFeedbacks);
				
				for (int i = 0; i < queryColumn.length; i++)
					pMatrix.set(i + 1, queryColumnIndex, queryColumn[i]);
			}
		} catch (InvalidIndexException e) {
			// Eccezione che non si verifa mai
			e.printStackTrace();
		}
	}

	private void init(double pScaleQuery, double pScalePositiveFeedbacks, double pScaleNegativeFeedbacks) {
		this.scaleQuery = pScaleQuery;
		this.scalePositiveFeedbacks = pScalePositiveFeedbacks;
		this.scaleNegativeFeedbacks = pScaleNegativeFeedbacks;
	}
	
	private double scaleQuery;
	private double scalePositiveFeedbacks;
	private double scaleNegativeFeedbacks;
}
