/*
 *  GenericLexProcessing.java - SE@SA Lab 2006
 *  
 *  This file is part of COCONUT.
 *
 *  COCONUT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  COCONUT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with COCONUT; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
package org.sesalab.IR.parser;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

/**
 * Classe che implementa un analizzatore lessicale
 * 
 * @author Paola Sgueglia
 * @since Febbraio 2006
 * @author Francesco Zurolo
 * @since Settembre 2005
 * @author Rocco Oliveto
 * @since Dicembre 2003
 */
public class GenericLexProcessing implements ILexProcessing {

	// Dimensione del buffer: circa 4 Mb
	private static final int BUFFER_SIZE = 4194304;

	// Lunghezza minima delle parole
	private static final int MIN_LENGTH = 2;

	/**
	 * Costruisce un nuovo analizzatore lessicale. Non prevede l'utilizzo di una
	 * stop-word list.
	 */
	public GenericLexProcessing() {
		this.stopWordList = new HashSet<String>();
	}

	/**
	 * Costruisce un nuovo analizzatore lessicale. Prevede l'utilizzo di una
	 * stop-word list.
	 * 
	 * @param pStopWordListFileName
	 *        Nome del file contenente la stop list.
	 * @throws IOException
	 *         Causata da un errore durante la lettura del file contenente la
	 *         stop word list
	 */
	public GenericLexProcessing(String pStopWordListFileName) throws IOException {
		BufferedReader stopWordListFile = null;
		try {
			// Dichiarazione buffer di lettura per il file contenente la stop
			// word list
			stopWordListFile = new BufferedReader(new FileReader(pStopWordListFileName));
			String word;

			this.stopWordList = new HashSet<String>();

			// Caricamento della stop word list
			while ((word = stopWordListFile.readLine()) != null)
				this.stopWordList.add(word);

		} finally {
			if (stopWordListFile != null)
				stopWordListFile.close();
		}
	}

	/**
	 * @see org.sesalab.IR.parser.ILexProcessing#analize(java.lang.String)
	 */
	public Map<String, Integer> analize(String pDocumentFileName) throws IOException {
		// Dichiarazione canale e buffer di lettura
		FileChannel fileChannel = new FileInputStream(pDocumentFileName).getChannel();
		ByteBuffer byteBuffer = ByteBuffer.allocate(GenericLexProcessing.BUFFER_SIZE);
		try {
			// Variabili di appoggio
			String word;
			String[] words;
			int occ;
			char c;

			this.wordList = new HashMap<String, Integer>();

			// Lettura del documento di input
			fileChannel.read(byteBuffer);
			byteBuffer.flip();
			while (true) {
				word = "";
				// Lettura di una parola
				while (byteBuffer.hasRemaining()) {
					c = (char) byteBuffer.get();
					if (newWord(c))
						break;
					if (isLetter(c))
						word += c;
				}
				// Se la parola  valida
				if ((word.length() > GenericLexProcessing.MIN_LENGTH) && (!(isNumber(word)))) {
					words = this.splitCamelCaseIdentifier(word);
					for (int i = 0; i < words.length; i++) {
						word = words[i];
						// Si ricerca la parola all'interno della STOP LIST
						if ((word.length() > GenericLexProcessing.MIN_LENGTH) && (!(this.stopWordList.contains(word)))) {
							// Si ricerca la parola all'interno della mappa
							if (this.wordList.containsKey(word))
								// Incremento delle occorrenze
								occ = this.wordList.get(word) + 1;// String.valueOf(Integer.parseInt(occ)
							// + 1);
							else
								// Parola non presente nella mappa
								occ = 1;
							// La parola viene inserita nella mappa delle parole
							this.wordList.put(word, occ);
						}
					}
				}

				// Fine del file
				if (!byteBuffer.hasRemaining())
					break;
			}
		} finally {
			if (fileChannel != null)
				fileChannel.close();
		}

		// Si carica il glossario dei termini
		return (this.wordList);
	}

	/*
	 * Determina se il carattere letto  una lettera 
	 * @param pCharacter Carattere letto
	 * @return <code>true</code> se pCharacter  una lettera, <code>false</code>
	 * 		   altrimenti.
	 */
	private boolean isLetter(char pCharacter) {
		boolean punctuation, bracket, otherSimbol, tab;

		// Punteggiatura
		punctuation = (pCharacter == '.') || (pCharacter == ':') || (pCharacter == ',') || (pCharacter == ';') || (pCharacter == '"') || (pCharacter == '!') || (pCharacter == '?');
		// Parentesi
		bracket = (pCharacter == '<') || (pCharacter == '>') || (pCharacter == '(') || (pCharacter == ')') || (pCharacter == '[') || (pCharacter == ']');
		// Altri simboli
		otherSimbol = (pCharacter == '/') || (pCharacter == '\\') || (pCharacter == '=') || (pCharacter == '^') || (pCharacter == '-') || (pCharacter == '|');
		// Tabulazioni
		tab = (pCharacter == 9);
		return (!(punctuation || bracket || otherSimbol || tab));
	}

	/*
	 * Determina se una stringa letta rappresenta un numero
	 * @param pString 
	 * 		  Stringa letta
	 * @return <code>true</code> se pString  una lettera, <code>false</code> 
	 * 		   altrimenti
	 */
	private boolean isNumber(String pString) {
		boolean digit;
		// La funzione considera un numero una parola che inizia con una cifra
		char firstChar = pString.charAt(0);

		digit = (firstChar >= '0') && (firstChar <= '9');
		return (digit);
	}

	/*
	 * Provvede all'individuazione di nuove parole
	 * @param pCharacter Carattere letto
	 * @return <code>true</code> se pCharacter rappresenta un carattere di divisione,
	 * 		   <code>false</code> altrimenti.
	 */
	private boolean newWord(char pCharacter) {
		boolean newLine, space, lowDash, apostrophe;

		// New line
		newLine = (pCharacter == (char) 10) || (pCharacter == (char) 13);
		// Spazio
		space = (pCharacter == ' ');
		// Apostrofo
		apostrophe = (pCharacter == '\'');
		// Trattino basso
		lowDash = (pCharacter == '_');
		return (newLine || space || apostrophe || lowDash);
	}

	/*
	 * Provvede ad effettuare lo splitting dei nomi delle variabili
	 * @param pIdentifier 
	 * 		  Nome della variabile
	 * @return Un vettore di stringhe contenente le diverse parole che compongono 
	 * 		   pIdentifier.
	 */
	private String[] splitCamelCaseIdentifier(String pIdentifier) {
		String[] words, aux;
		int i = 1, index = 0;
		char character;
		words = new String[1];
		words[index] = String.valueOf(pIdentifier.charAt(0));
		words[index] = words[index].toLowerCase();
		while (i < pIdentifier.length()) {
			character = pIdentifier.charAt(i);
			// Se il carattere letto  UPPER CASE
			if ((character >= 65) && (character <= 90)) {
				// Individuazione di una nuova componente
				aux = words;
				words = new String[words.length + 1];
				System.arraycopy(aux, 0, words, 0, words.length - 1);
				index++;
				words[index] = String.valueOf(character).toLowerCase();
			} else
				// Composizione di una componente
				words[index] += String.valueOf(character).toLowerCase();
			i++;
		}

		return (words);
	}

	// Lista utilizzata per contenere la STOP LIST
	private HashSet<String> stopWordList;

	// Mappa utilizzata per memorizzare le parole del testo (chiave) e le loro
	// occorrenze
	private HashMap<String, Integer> wordList;
}
