package traceability.recovery.coconut;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.NoSuchElementException;
import java.util.Scanner;
import java.util.Set;

import org.eclipse.core.resources.IProject;
import org.sesalab.IR.documentSpace.BasicTermByDocumentMatrix;
import org.sesalab.IR.documentSpace.storage.FileDocumentSpaceStorage;
import org.sesalab.IR.documentSpace.weigher.NormalisedTfIdfWeigher;
import org.sesalab.IR.model.IIRModel;
import org.sesalab.IR.model.exception.InvalidQueryException;
import org.sesalab.IR.rankedList.IRankedList;
import org.sesalab.IR.rankedList.RankedListElement;

import traceability.links.Granularity;
import traceability.links.TraceabilityLink;
import traceability.management.Manager;
import traceability.recovery.Preprocessor;
import traceability.recovery.RecoveryTechnique;
import traceability.recovery.RecoveryTechniqueParams;

/**
 * Superclass for links recovery techniques that rely on COCONUT's IR API.
 *  
 * @author Sam Klock
 *
 */
@SuppressWarnings("restriction")
public abstract class COCONUTRecoveryTechnique implements RecoveryTechnique {
	
	/**
	 * Constructs a term-document matrix for interested techniques.
	 * @param project the project from which to construct the matrix
	 * @param directory the subdirectory in which to store transient information for the matrix
	 * @return the matrix, or null if something went wrong
	 */
	protected final BasicTermByDocumentMatrix<String, String> constructMatrix(IProject project, String directory) {
		Manager manager = Manager.getInstance();
		BasicTermByDocumentMatrix<String, String> matrix = null;
		try {
			matrix = new BasicTermByDocumentMatrix<String, String>(new NormalisedTfIdfWeigher(), 
					 new FileDocumentSpaceStorage<String, String>(manager.getDirPath(project) + File.separator + directory));
		} catch (FileNotFoundException e) {
			//TODO improve
			e.printStackTrace();
			return null;
		} catch (IOException e) {
			// TODO improve
			e.printStackTrace();
			return null;
		}
		
		return matrix;
	}
	
	/**
	 * Helper method used to populate the term-document matrix.
	 * @param matrix the matrix to populate
	 * @param manager traceability manager for current Eclipse session
	 * @param project the project for which we are recovering links
	 * @param targetDir subdirectory containing the documents to insert in the matrix
	 * @param prefix filename prefixes for the target documents
	 * @return names of documents added to the matrix
	 */
	protected final Collection<String> populateMatrixWithArtifacts(BasicTermByDocumentMatrix<String, String> matrix, 
			Manager manager, IProject project, String targetDir, String prefix) {
		File dir = new File(manager.getDirPath(project) + File.separator + targetDir);
		if(!dir.isDirectory() || !dir.canRead()) {
			//MUST be able to read the source directory
			return null;
		}
		
		File mapper = new File(dir.getAbsolutePath() + File.separator + Preprocessor.MAPPER_NAME);
		if(!mapper.canRead()) {
			//MUST be able to read the mapper
			return null;
		}
		
		//Add source documents to matrix
		HashMap<String, Integer> termMap = null;
		Collection<String> docs = new HashSet<String>();
		Scanner scanner = null;
		for(String s : dir.list()) {
			if(s.startsWith(prefix) && !s.equals(Preprocessor.MAPPER_NAME)) {
				try {
					scanner = new Scanner(new File(dir.getAbsolutePath() + File.separator + s));
				} catch (FileNotFoundException e) {
					//MUST be able to read each source file
					e.printStackTrace();
					return null;
				}
				
				termMap = new HashMap<String, Integer>();
				while(scanner.hasNext()) {
					String next = scanner.next();
					
					if(termMap.containsKey(next)) {
						termMap.put(next, termMap.get(next) + 1);
					} else {
						termMap.put(next, 1);
					}
				}
				scanner.close();
				
				String artifactID = getArtifactFor(s, mapper, prefix);
				if(artifactID == null) {
					//No artifact could be found in the mapper for the given file
					return null;
				}
				matrix.addDocument(artifactID, termMap);
				docs.add(artifactID);
			}
		}
		
		return docs;
	}
	
	/**
	 * Retrieves the name of the artifact whose terms are stored in the file named by s.  Called 
	 * by populateMatrixWithArtifacts().
	 * @param s filename storing artifact's terms
	 * @param mapper file mapping filenames to artifact names
	 * @param prefix filename prefix for files being inserted in the matirx
	 * @return name of the artifact described by the file named by s
	 */
	private final String getArtifactFor(String s, File mapper, String prefix) {
		Scanner scanner = null;
		String id = null;
		try {
			scanner = new Scanner(mapper);
		} catch (FileNotFoundException e) {
			//Mapper unreadable
			return null;
		}
		
		s = s.substring(prefix.length(), s.indexOf(Preprocessor.SUFFIX));
		int index = -1;
		try {
			index = Integer.parseInt(s);
		} catch(NumberFormatException e) {
			//Bad filename
			return null;
		}
		
		for(int i = 0; i < index; i++) {
			try {
				id = scanner.nextLine();
			} catch(NoSuchElementException e) {
				//Indexing error
				return null;
			}
		}
		
		scanner.close();
		return id;
	}
	
	/**
	 * Runs the the technique, filtering the results returned according to the threshold/minimum 
	 * similarity specified by the techniques parameters.
	 * @param model the model specifying the technique to run
	 * @param queries queries for the model
	 * @param targets targets for the queries
	 * @return the set of recovered traceability links, or null if something went wrong
	 */
	protected final Set<TraceabilityLink> runTechnique(IIRModel<String, String> model, 
			Collection<String> queries, Collection<String> targets) {
		IRankedList results = null;
		Set<TraceabilityLink> links = null;
		RecoveryTechniqueParams theParams = getParameters();
		if(!(theParams instanceof COCONUTRecoveryTechniqueParams)) {
			//Need the parameters to be COCONUTRecoveryTechniqueParams
			return null;
		}
		COCONUTRecoveryTechniqueParams params = (COCONUTRecoveryTechniqueParams)theParams;
		try {
			if(params.getWhetherThreshold()) {
				results = model.getRankedList(queries, targets, 0.00);
			} else {
				results = model.getRankedList(queries, targets, params.getMinimumSimilarity());
			}
		} catch(InvalidQueryException e) {
			//THERE'S A BUG IN THE CODE
			e.printStackTrace();
			return null;
		}
		
		links = new HashSet<TraceabilityLink>();
		int i = 0;
		Granularity gran = params.getGranularity();
		for(RankedListElement<String> e : results) {
			String source = e.getSource();
			String doc = e.getTarget();
			
			if(!targets.contains(doc)) {
				//Follows a quirk in how COCONUT builds its ranked lists
				String temp = source;
				source = doc;
				doc = temp;
			}
			double score = e.getSimilarityValue();
			TraceabilityLink link = new TraceabilityLink(source, doc, score, gran); 
			links.add(link);
			
			i++;
			if(params.getWhetherThreshold() && params.getThreshold() == i) {
				break;
			}
		}
		
		return links;
	}
}
