package traceability.recovery.lsi;

import java.io.File;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.core.resources.IProject;

import org.sesalab.IR.documentSpace.BasicTermByDocumentMatrix;
import org.sesalab.IR.model.LSI.LatentSemanticIndexing;
import org.sesalab.IR.model.LSI.exception.InvalidSubspaceDimensionException;
import org.sesalab.IR.model.vectorSimilarityPerformer.CosineSimilarity;

import traceability.links.TraceabilityLink;
import traceability.management.Manager;
import traceability.recovery.Preprocessor;
import traceability.recovery.RecoveryTechniqueParams;
import traceability.recovery.coconut.COCONUTRecoveryTechnique;

/**
 * 
 * Implements latent semantic indexing for Traceclipse.  This recovery technique depends heavily on 
 * the implementation of LSI in COCONUT by Sgueglia et al. and SE@SA Lab, 2006. 
 * 
 * @see org.sesalab.IR.model.LSI.LatentSemanticIndexing
 * @author Sam Klock
 *
 */
@SuppressWarnings("restriction")
public final class LSIRecoveryTechnique extends COCONUTRecoveryTechnique {

	/**
	 * Name for this technique in the recovery wizard.
	 */
	private static final String NAME = "Latent Semantic Indexing";
	
	/**
	 * Directory in which to story the SVD of the term-document matrix.
	 */
	private static final String DIRECTORY = "lsi_index";
	/**
	 * Parameterization of the COCONUT LSI implementation forcing it to rebuild 
	 * the SVD for each query.
	 */
	private static final boolean FORCE_REBUILD = true;
	
	/**
	 * Set of links recovered via LSI.
	 */
	private Set<TraceabilityLink> links;
	/**
	 * Parameters for LSI.
	 */
	private LSIRecoveryTechniqueParams params;
	
	/**
	 * Constructor.
	 */
	public LSIRecoveryTechnique() {
		params = new LSIRecoveryTechniqueParams();
		links = new HashSet<TraceabilityLink>();
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#getName()
	 */
	@Override
	public String getName() {
		return NAME;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#getParameters()
	 */
	@Override
	public RecoveryTechniqueParams getParameters() {
		return params;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#getRecoveredLinks()
	 */
	@Override
	public Set<TraceabilityLink> getRecoveredLinks() {
		return links;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#run(org.eclipse.core.resources.IProject)
	 */
	@Override
	public boolean run(IProject project) {
		Manager manager = Manager.getInstance();
		
		BasicTermByDocumentMatrix<String, String> matrix = constructMatrix(project, DIRECTORY);
		if(matrix == null) {
			//Couldn't construct matrix for some reason...
			return false;
		}
		
		Collection<String> queries = populateMatrixWithArtifacts(matrix, manager, project, 
				Preprocessor.SOURCE_DIR, Preprocessor.SOURCE_PREFIX);
		Collection<String> targets = populateMatrixWithArtifacts(matrix, manager, project, 
				Preprocessor.DOC_DIR, Preprocessor.DOC_PREFIX);
		
		if(queries == null || targets == null) {
			//Couldn't populate the matrix for some reason...
			return false;
		}
		
		LatentSemanticIndexing<String, String> lsi = null;
		try {
			lsi = new LatentSemanticIndexing<String, String>(matrix, new CosineSimilarity(), 
					manager.getDirPath(project) + File.separator + DIRECTORY, 
					(Integer)params.getParameterValue(LSIRecoveryTechniqueParams.PARAMETER_ID), FORCE_REBUILD);
		} catch (InvalidSubspaceDimensionException e) {
			//Dimensionality value for params is probably less than or equal to zero
			e.printStackTrace();
			return false;
		}
		
		links = runTechnique(lsi, queries, targets);
		return (links != null);
	}
}
