package traceability.recovery.vsm;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.core.resources.IProject;
import org.sesalab.IR.documentSpace.BasicTermByDocumentMatrix;
import org.sesalab.IR.model.VSM.VectorSpaceModel;
import org.sesalab.IR.model.vectorSimilarityPerformer.CosineSimilarity;

import traceability.links.TraceabilityLink;
import traceability.management.Manager;
import traceability.recovery.Preprocessor;
import traceability.recovery.RecoveryTechniqueParams;
import traceability.recovery.coconut.COCONUTRecoveryTechnique;

/**
 * 
 * Implements the vector space model as a traceability links recovery technique.  This technique relies 
 * heavily on COCONUT's implementation of VSM by Sgueglia et al. and SE@SA Lab, 2006. 
 * 
 * @see org.sesalab.IR.model.VSM.VectorSpaceModel
 * @author Sam Klock
 *
 */
@SuppressWarnings("restriction")
public class VSMRecoveryTechnique extends COCONUTRecoveryTechnique {

	private static final String NAME = "Vector Space Model";
	private static final String DIRECTORY = "vsm_matrix";
	private VSMRecoveryTechniqueParams params;
	private Set<TraceabilityLink> links;
	
	/**
	 * Constructor.
	 */
	public VSMRecoveryTechnique() {
		links = new HashSet<TraceabilityLink>();
		params = new VSMRecoveryTechniqueParams();
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#getName()
	 */
	@Override
	public String getName() {
		return NAME;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#getParameters()
	 */
	@Override
	public RecoveryTechniqueParams getParameters() {
		return params;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#getRecoveredLinks()
	 */
	@Override
	public Set<TraceabilityLink> getRecoveredLinks() {
		return links;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#run(org.eclipse.core.resources.IProject)
	 */
	@Override
	public boolean run(IProject project) {
		Manager manager = Manager.getInstance();
		BasicTermByDocumentMatrix<String, String> matrix = this.constructMatrix(project, DIRECTORY);
		
		if(matrix == null) {
			//Couldn't construct the matrix for some reason...
			return false;
		}
		
		Collection<String> queries = this.populateMatrixWithArtifacts(matrix, manager, project, 
				Preprocessor.SOURCE_DIR, Preprocessor.SOURCE_PREFIX);
		Collection<String> targets = this.populateMatrixWithArtifacts(matrix, manager, project, 
				Preprocessor.DOC_DIR, Preprocessor.DOC_PREFIX);
		
		if(queries == null || targets == null) {
			//Couldn't populate the matrix for some reason...
			return false;
		}
		
		VectorSpaceModel<String, String> vsm = 
			new VectorSpaceModel<String, String>(matrix, new CosineSimilarity());
		
		links = this.runTechnique(vsm, queries, targets);
		return (links != null);
	}

}
