package traceability.links;

import java.io.Serializable;

/**
 * 
 * Simple representation of a traceability link.
 * 
 * @author Sam Klock
 *
 */
public class TraceabilityLink implements Serializable {

	//TODO Implement facility for incorporating versioning information into stored links
	
	//For java.io.Serializable
	private static final long serialVersionUID = -10132010011715L;
	
	/**
	 * Similarity value specifying that a particular link is user-specified.
	 */
	public static final double USER_SPECIFIED = 1.5;
	
	/**
	 * Similarity value between source element and documentation element.
	 */
	private double sim;
	/**
	 * Path to the source element.
	 */
	private String source;
	/**
	 * Path to the documentation element.
	 */
	private String doc;
	/**
	 * Granularity of the source element.
	 */
	private Granularity gran;
	/**
	 * Optional textual description of the traceability link.
	 */
	private String description;
	
	/**
	 * Builds a new user-specified traceability link.
	 * @param source the source artifact for the link
	 * @param doc the documentation artifact for the link
	 * @param gran the granularity level with which to view the source element
	 */
	public TraceabilityLink(String source, String doc, Granularity gran) {
		if(source == null || doc == null) {
			throw new IllegalArgumentException("source/doc cannot be null");
		}
		this.source = source;
		this.doc = doc;
		this.sim = USER_SPECIFIED;
		this.gran = gran;
		this.description = "";
	}
	
	/**
	 * Builds a new recovered traceability link with the specified similarity value.
	 * @param source the source artifact for the link
	 * @param doc the documentation artifact for the link
	 * @param sim the similarity value for the link
	 * @param gran the granularity level with which to view the source element
	 */
	public TraceabilityLink(String source, String doc, double sim, Granularity gran) {
		if(source == null || doc == null) {
			throw new IllegalArgumentException("source/doc cannot be null");
		}
		this.source = source;
		this.doc = doc;
		this.sim = sim;
		this.gran = gran;
	}
	
	/**
	 * Retrieves the path to the source element for this link.
	 * @return path to source as string
	 */
	public String getSource() {
		return source;
	}
	
	/**
	 * Retrieves the path to the documentation element for this link.
	 * @return path to document as string
	 */
	public String getDocument() {
		return doc;
	}
	
	/**
	 * Retrieves the similarity value for this link.
	 * @return the similarity value as double
	 * @see USER_SPECIFIED
	 */
	public double getSimilarity() {
		return sim;
	}
	
	/**
	 * Retrieves the granularity level of the source element.
	 * @return the granularity as Granularity
	 * @see Granularity
	 */
	public Granularity getSourceGranularity() {
		return gran;
	}

	/**
	 * Returns a textual description (if present) of the traceability link.
	 * @return the textual description
	 */
	public String getDescription() {
		return description;
	}
	
	/**
	 * Sets the description of this traceability link to the given string.
	 * 
	 * If the string is null, an empty string is used in its place.
	 * @param desc textual description of this traceability link
	 */
	public void setDescription(String desc) {
		if(desc == null) {
			this.description = "";
		} else {
			this.description = desc;
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object other) {
		if(other == this) {
			return true;
		}
		if(other == null || other.getClass() != this.getClass()) {
			return false;
		}
		
		//Equality based on source and documentation (other features not essential)
		TraceabilityLink oth = (TraceabilityLink)other;
		return (this.source.equals(oth.source) &&
				this.doc.equals(oth.doc));
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		//For equals(): can only consider source and doc for hashcode
		return (int)(source.hashCode() + doc.hashCode());
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return source + "(" + gran.toString() + ") -> " + doc + ": " + 
			(sim == USER_SPECIFIED ? "user-specified" : sim);
	}
}
