package traceability.links;

import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Represents a set of traceability links.
 * 
 * This class is serializable.  Via the Manager, for a given project, a TraceabilityLinksSet is serialized 
 * into XML (via XStream) when a set is saved; similarly, when a set is restored, a TraceabilityLinksSet is 
 * deserialized from XML.  Accordingly, this class is the foundational representation of a set of traceability 
 * links for any Eclipse project.
 * 
 * @author Sam Klock
 * @see traceability.management.Manager
 * @see traceability.links.TraceabilityLink
 */
public class TraceabilityLinkSet implements Serializable {
	
	private static final long serialVersionUID = 20101110224503L;
	
	/**
	 * Mapping of source artifacts to links for those source artifacts.
	 */
	private Map<String, List<TraceabilityLink>> sourceToDoc;
	/**
	 * Mapping of documentation artifacts to links for those documentation artifacts.
	 */
	private Map<String, List<TraceabilityLink>> docToSource;
	
	/**
	 * Represents the entire set of traceability links.  Computed once when the entire set is requested, 
	 * and modified accordingly when links are added or removed.
	 */
	private transient Set<TraceabilityLink> set;
	
	/**
	 * The number of traceability links in this set.
	 */
	private int numLinks;
	
	/**
	 * Constructor.  Creates a new, empty set.
	 */
	public TraceabilityLinkSet() {
		sourceToDoc = new HashMap<String, List<TraceabilityLink>>();
		docToSource = new HashMap<String, List<TraceabilityLink>>();
		set = new HashSet<TraceabilityLink>();
		numLinks = 0;
	}
	
	/**
	 * Adds a new traceability link to the set.
	 * 
	 * The link is only added if a link for the given source and destination are not already present 
	 * in the set.
	 * 
	 * @param link traceability link to add
	 * @return true if the link was added; false otherwise
	 */
	public boolean addLinkToSet(TraceabilityLink link) {
		//Do not add null links; throw exception instead
		if(link == null) {
			throw new NullPointerException("link == null");
		}
		
		//First handle the source-to-document mapping
		//If we already have links for this source, add to the associated list
		if(sourceToDoc.containsKey(link.getSource())) {
			List<TraceabilityLink> linkList = sourceToDoc.get(link.getSource());
			
			//Link already present?  Do nothing and return false
			if(linkList.contains(link)) {
				return false;
			}
			
			//Otherwise simply add the link
			linkList.add(link);
		} else {//Else, create the list for this source entity
			List<TraceabilityLink> linkList = new LinkedList<TraceabilityLink>();
			linkList.add(link);
			
			sourceToDoc.put(link.getSource(), linkList);
		}
		
		//Second, handle the document-to-source mapping
		if(docToSource.containsKey(link.getDocument())) {
			List<TraceabilityLink> linkList = docToSource.get(link.getDocument());
			linkList.add(link);
		} else {
			List<TraceabilityLink> linkList = new LinkedList<TraceabilityLink>();
			linkList.add(link);
			
			docToSource.put(link.getDocument(), linkList);
		}
		
		//Finally, increment size and return
		numLinks++;
		if(set != null) {
			set.add(link);
		}
		return true;
	}
	
	/**
	 * Removes the given traceability link from the set.  If there is a link with the same source and 
	 * documentation strings in the set as the given link, it will be removed.
	 * @param link link to remove
	 * @return the corresponding link removed from the set, or null if no link was removed
	 */
	public TraceabilityLink removeLink(TraceabilityLink link) {
		//No link provided or link source is not in the source-to-document map?
		//Return null
		if(link == null || !sourceToDoc.containsKey(link.getSource())) {
			return null;
		}
		
		//Else, remove it first from the source-to-document map
		//If the link itself is not in that map, return null
		List<TraceabilityLink> linkList = sourceToDoc.get(link.getSource());
		if(linkList.contains(link)) {
			linkList.remove(link);
			if(linkList.isEmpty()) {
				sourceToDoc.remove(link.getSource());
			}
		} else {
			return null;
		}
		
		//Then remove from document-to-source map.  If the link is in 
		//one map, it must be in the other.
		linkList = docToSource.get(link.getDocument());
		linkList.remove(link);
		if(linkList.isEmpty()) {
			docToSource.remove(link.getDocument());
		}
		
		//Then simply return the link
		numLinks--;
		if(set != null) {
			set.remove(link);
		}
		return link;
	}
	
	/**
	 * Removes all links for the given source artifact.
	 * @param source source artifact for which to remove links
	 * @return a (possibly zero-length) array of links that were removed
	 */
	public TraceabilityLink[] removeLinksForSource(String source) {
		return removeLinksFor(source, sourceToDoc);
	}
	
	/**
	 * Removes all links for the given documentation artifact.
	 * @param doc documentation artifact for which to remove links
	 * @return a (possibly zero-length) array of links that were removed
	 */
	public TraceabilityLink[] removeLinksForDocument(String doc) {
		return removeLinksFor(doc, docToSource);
	}
	
	/**
	 * Retrieves the entire set of traceability links.
	 * 
	 * Note that the set returned is not the way in which the links are internally represented.  Modifying 
	 * the returned set will not modify the set of links stored in this object, although it will affect 
	 * the set that other entities calling this method see.
	 * @return entire set of traceability links
	 */
	public Set<TraceabilityLink> getLinks() {
		if(set != null && set.size() == numLinks) {
			return set;
		}
		
		set = new HashSet<TraceabilityLink>();
		for(String s : sourceToDoc.keySet()) {
			List<TraceabilityLink> linkList = sourceToDoc.get(s);
			for(TraceabilityLink l : linkList) {
				set.add(l);
			}
		}
		
		return set;
	}
	
	/**
	 * Retrieves a set of traceability links for the given source artifact.
	 * @param source source artifact for which to retrieve links
	 * @return traceability links for the source artifact in this set
	 */
	public Set<TraceabilityLink> getLinksForSource(String source) {
		return getLinksFor(source, sourceToDoc);
	}
	
	/**
	 * Retrieves a set of traceability links for the given documentation artifact.
	 * @param doc documentation artifact for which to retrieve links
	 * @return traceability links for the documentation artifact in this set
	 */
	public Set<TraceabilityLink> getLinksForDocument(String doc) {
		return getLinksFor(doc, docToSource);
	}
	
	/**
	 * Retrieves the size of this set (i.e., the number of traceability links stored in this set).
	 * @return the size
	 */
	public int size() {
		return numLinks;
	}
	
	/**
	 * Private helper method for getLinksFor{Source,Document}.
	 * @param key artifact for which to retrieve links
	 * @param map sourceToDoc or docToSource for source or documentation
	 * @return a (possibly empty) set
	 */
	private Set<TraceabilityLink> getLinksFor(String key, Map<String, List<TraceabilityLink>> map) {
		Set<TraceabilityLink> set = new HashSet<TraceabilityLink>();
		
		if(key == null || !map.containsKey(key)) {
			return set;
		}
		
		List<TraceabilityLink> linkList = (List<TraceabilityLink>)map.get(key);
		for(TraceabilityLink l : linkList) {
			set.add(l);
		}
		
		return set;
	}
	
	/**
	 * Private helper method for removeLinksFor{Source,Document}
	 * @param key artifact for which to remove links
	 * @param map sourceToDoc or DocToSource for source or documentation
	 * @return a (possibly zero-length) array of links removed
	 */
	private TraceabilityLink[] removeLinksFor(String key, Map<String, List<TraceabilityLink>> map) {
		if(map.containsKey(key)) {
			List<TraceabilityLink> toRemove = (List<TraceabilityLink>)map.get(key);
			TraceabilityLink [] toReturn = toRemove.toArray(new TraceabilityLink[0]);
			
			for(TraceabilityLink l : toReturn) {
				removeLink(l);
			}
			
			return toReturn;
		} else {
			return new TraceabilityLink[0];
		}
	}
}
