package traceability.management;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.HashSet;
import java.util.Scanner;
import java.util.Set;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;

import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.io.xml.DomDriver;

import traceability.links.Granularity;
import traceability.links.TraceabilityLink;
import traceability.links.TraceabilityLinkSet;

/**
 * Provides traceability links management functionality.
 * 
 * This is a singleton class, i.e., only one instance exists.
 * 
 * @author Sam Klock
 *
 */
public class Manager {

	/**
	 * Subdirectory within a project where links are to be stored.
	 */
	private static final String SUBDIR = ".trace";	//TODO project preference?
	/**
	 * File in which links are stored.
	 */
	private static final String FILE   = "links.xml";	//TODO project preference?
	/**
	 * File in which the location of documentation is stored.
	 */
	private static final String DOC_FILE = "docloc.cfg";	//TODO project preference?
	/**
	 * Single instance.  Created on demand.
	 */
	private static Manager theManager = null;
	
	/**
	 * Current set of links with which to work.
	 */
	private TraceabilityLinkSet currentSet;
	
	/**
	 * Project for which the current set exists.
	 */
	private IProject currentProject;
	
	/**
	 * Location of documentation for the current project.
	 */
	private String docLocation;
	
	/**
	 * The results of the last instance of traceability links generation.
	 */
	private Set<TraceabilityLink> lastGenerated;
	
	/**
	 * Retrieves the single, global instance of the Manager.
	 * @return the Manager
	 */
	public static Manager getInstance() {
		if(theManager == null) {
			theManager = new Manager();
		}
		return theManager;
	}
	
	/**
	 * Private constructor.
	 */
	private Manager() {
		currentSet = null;
	}
	
	/**
	 * Sets the current project for which to manage traceability links.
	 * 
	 * Specifically, this method loads the traceability link set for the given project from disk.  
	 * If no set exists, it is created.
	 * 
	 * @param project the project for which to manage links
	 * @throws ManagerException if a problem occurs loading the project's traceability link set
	 */
	public void setProject(IProject project) throws ManagerException {
		//Don't need to do anything if the given project is the current project.
		if(currentProject != null && currentProject.equals(project)) {
			return;
		}
		String dirPath = getDirPath(project);
		
		//First: ensure that the links directory actually exists
		File dirFile = new File(dirPath);
		if(!dirFile.isDirectory()) {
			if(dirFile.exists()) {
				throw new ManagerException("Expected traceability links directory " + 
						dirPath + " is not a directory");
			} else if(!dirFile.mkdir()) {
				throw new ManagerException("Cannot create directory " + dirPath + 
						" for traceability links storage.");
			}
		}
		
		//Second: load the traceability link set (if it exists)
		//Otherwise, create it.
		//Either way, save the current set.
		saveSet();
		currentProject = project;
		restoreSet();
		
		File docFile = new File(dirPath + File.separator + DOC_FILE);
		if(!docFile.exists()) {
			try {
				docFile.createNewFile();
			} catch (IOException e) {
				throw new ManagerException("Could not create documentation file");
			}
			docLocation = null;
		} else {
			try {
				Scanner scanner = new Scanner(docFile);
				if(scanner.hasNextLine()) {
					docLocation = scanner.nextLine();
					if(docLocation.equals("")) {
						docLocation = null;
					}
				} else {
					docLocation = null;
				}
				scanner.close();
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Retrieves the current project for which this manager is handling traceability links.
	 * @return the project
	 */
	public IProject getProject() {
		return currentProject;
	}
	
	/**
	 * Retrieves the user-specified location of textual documentation for the current project.
	 * @return documentation location
	 */
	public String getDocumentationLocation() {
		if(docLocation == null) {
			return "";
		}
		return docLocation;
	}
	
	/**
	 * Sets the location of textual documentation for the current project.
	 * @param docLoc documentation location
	 * @throws ManagerException if the specified location is not a directory
	 */
	public void setDocumentationLocation(String docLoc) throws ManagerException {
		if(docLoc == null) {
			throw new ManagerException("Attempt to set documentation location to null");
		}
		File docLocFile = new File(docLoc);
		if(docLocFile.exists() && docLocFile.isDirectory()) {
			docLocation = docLoc;
		} else {
			throw new ManagerException("Attempt to set documentation location to non-directory: " + docLoc);
		}
	}
	
	/**
	 * Saves the current set of traceability links to disk.  Does nothing if there is no current set.
	 * @throws ManagerException if an issue occurs while writing to disk
	 */
	public void saveSet() throws ManagerException {
		//Nothing to do in this case.
		if(currentSet == null) {
			return;
		}
		
		String xmlPath = getDirPath(currentProject) + File.separator + FILE;
		
		XStream str = new XStream();
		str.setMode(XStream.NO_REFERENCES);
		
		FileOutputStream out = null;
		try {
			out = new FileOutputStream(xmlPath);
		} catch (FileNotFoundException e) {
			throw new ManagerException("Cannot open links file " + xmlPath + " for writing.");
		}
		str.toXML(currentSet, out);
		
		try {
			out.close();
		} catch (IOException e) {
			//Probably shouldn't be an issue
			e.printStackTrace();
		}
		
		String docPath = getDirPath(currentProject) + File.separator + DOC_FILE;
		
		PrintStream pOut = null;
		try {
			pOut = new PrintStream(new File(docPath));
		} catch (FileNotFoundException e) {
			throw new ManagerException("Cannot open links file " + xmlPath + " for writing.");
		}
		
		pOut.println(docLocation);
		
		try {
			out.close();
		} catch (IOException e) {
			//Probably shouldn't be an issue
			e.printStackTrace();
		}
	}
	
	/**
	 * Restores set of traceability links for the current project from disk.
	 * @throws ManagerException
	 */
	public void restoreSet() throws ManagerException {
		//Nothing to do in this case.
		if(currentProject == null) {
			return;
		}
		
		String xmlPath = getDirPath(currentProject) + File.separator + FILE;

		File xmlFile = new File(xmlPath);
		XStream stream = new XStream(new DomDriver());

		try {
			FileInputStream in = new FileInputStream(xmlFile);
			currentSet = (TraceabilityLinkSet)stream.fromXML(in);
			in.close();
		} catch(FileNotFoundException e) {
			currentSet = new TraceabilityLinkSet();
		} catch(IOException e) {
			throw new ManagerException("I/O exception occurred while loading link set: " + e.getMessage());
		} catch(Exception e) {
			e.printStackTrace();
		}
		
		String docPath = getDirPath(currentProject) + File.separator + DOC_FILE;
		try {
			Scanner scanner = new Scanner(new File(docPath));
			if(scanner.hasNextLine()) {
				setDocumentationLocation(scanner.nextLine());
			} else {
				docLocation = null;
			}
			scanner.close();
		} catch (Exception e) {
			docLocation = null;
		}
	}
	
	/**
	 * Adds a traceability link for the current project with the given source, destination 
	 * (documentation), and granularity.  The link is presumed to be user-specified.
	 * @param source link source
	 * @param dest link documentation
	 * @param gran granularity of source
	 * @return true if added; false otherwise
	 */
	public boolean addLink(String source, String dest, Granularity gran) {
		return addLink(source, dest, gran, TraceabilityLink.USER_SPECIFIED, "");
	}
	
	/**
	 * Adds a traceability link for the current project with the given source, destination 
	 * (documentation), granularity, and similarity.
	 * @param source link source
	 * @param dest link documentation
	 * @param gran granularity of source
	 * @param sim similarity value for source and documentation
	 * @return true if added; false otherwise
	 */
	public boolean addLink(String source, String dest, Granularity gran, double sim) {
		return addLink(source, dest, gran, sim, "");
	}
	
	/**
	 * Adds a traceability link for the current project with the given source, destination 
	 * (documentation), granularity, similarity, and description.  The link is presumed to 
	 * be user-specified.
	 * @param source link source
	 * @param dest link documentation
	 * @param gran granularity of source
	 * @param desc description of the added link
	 * @return true if added; false otherwise
	 */
	public boolean addLink(String source, String dest, Granularity gran, String desc) {
		return addLink(source, dest, gran, TraceabilityLink.USER_SPECIFIED, desc);
	}
	
	/**
	 * Adds a traceability link for the current project with the given source, destination 
	 * (documentation), granularity, similarity, and description.
	 * @param source link source
	 * @param dest link documentation
	 * @param gran granularity of source
	 * @param sim similarity value for source and documentation
	 * @param desc description of the added link
	 * @return true if added; false otherwise
	 */
	public boolean addLink(String source, String dest, Granularity gran, double sim, String desc) {
		if(currentSet == null) {
			return false;
		}
		TraceabilityLink newLink = new TraceabilityLink(source, dest, sim, gran);
		newLink.setDescription(desc);
		
		return currentSet.addLinkToSet(newLink);
	}
	
	/**
	 * Adds the given traceability link for the current project to the current set.
	 * @param link the link to add
	 * @return true if added; false otherwise
	 */
	public boolean addLink(TraceabilityLink link) {
		if(currentSet == null) {
			return false;
		}
		return currentSet.addLinkToSet(link);
	}
	
	/**
	 * Removes the given traceability link from the current set.
	 * @param link the link to remove
	 * @return true if the link was removed; false otherwise
	 */
	public boolean removeLink(TraceabilityLink link) {
		if(currentSet == null) {
			return false;
		}
		if(lastGenerated != null) {
			lastGenerated.remove(link);
		}
		return (currentSet.removeLink(link) != null);
	}
	
	/**
	 * Retrieves the path (in OS format) to the directory where traceability-related files are stored 
	 * for the given project.
	 * @param currentProject the project for which to retrieve the directory 
	 * @return a string representing the path to the traceability directory
	 */
	public String getDirPath(IProject currentProject) {
		String rootPath = ResourcesPlugin.getWorkspace().getRoot().getLocation().toOSString();
		return rootPath + currentProject.getFullPath().toOSString() + 
				File.separator + SUBDIR; 
	}
	
	/**
	 * Retrieves all traceability links in the current set for the given source element.
	 * @param source the source element
	 * @return links for the source element
	 */
	public Set<TraceabilityLink> getLinksForSource(String source) {
		if(currentProject == null) {
			return new HashSet<TraceabilityLink>();
		}
		return currentSet.getLinksForSource(source);
	}
	
	/**
	 * Retrieves all traceability links in the current set for the given documentation element.
	 * @param doc the documentation element
	 * @return links for the documentation element
	 */
	public Set<TraceabilityLink> getLinksForDocument(String doc) {
		if(currentProject == null) {
			return new HashSet<TraceabilityLink>();
		}
		return currentSet.getLinksForDocument(doc);
	}
	
	/**
	 * Retrieves all traceability links in the current set.
	 * @return the traceability links
	 */
	public Set<TraceabilityLink> getAllLinks() {
		if(currentProject == null) {
			return new HashSet<TraceabilityLink>();
		}
		return currentSet.getLinks();
	}
	
	/**
	 * Retrieves the results of the most recent link generation.
	 * @return most recent results, or null if none within the lifetime of this Manager
	 */
	public Set<TraceabilityLink> getLastGenerationResults() {
		return lastGenerated;
	}
	
	/**
	 * Sets the set of traceability links corresponding to the results of the most 
	 * recent link generation.
	 * @param lastGen most recent results
	 */
	public void setGenerationResults(Set<TraceabilityLink> lastGen) {
		lastGenerated = lastGen;
	}
}
