package traceability.popup.actions;

import java.util.Set;

import org.eclipse.core.resources.IProject;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;

import traceability.links.TraceabilityLink;
import traceability.management.Manager;
import traceability.management.ManagerException;
import traceability.views.TraceabilityContentProvider;
import traceability.views.TraceabilityView;
import traceability.wizard.LinkRecoveryWizard;

/**
 * 
 * ObjectActionDelegate for generating traceability links for a project.
 * 
 * @author Sam Klock
 * @see traceability.wizard
 */
public class GenerateLinks implements IObjectActionDelegate {

	/**
	 * Instance of this class.
	 */
	private static GenerateLinks instance;
	
	/**
	 * Shell for child dialogs.
	 */
	private Shell shell;
	/**
	 * Project for which to generate links.
	 */
	private IProject target;
	/**
	 * Most recent results of link generation.
	 */
	private Set<TraceabilityLink> mostRecentResults;
	
	/**
	 * Constructor.
	 */
	public GenerateLinks() {
		super();
		instance = this;
	}
	
	/**
	 * Retrieves most recently constructed instance of this class.
	 * @return the instance
	 */
	public static GenerateLinks getInstance() {
		return instance;
	}
	
	/**
	 * Retrieves the most recent results of link generation.
	 * @return the set, or null if no links were generated recently
	 */
	public Set<TraceabilityLink> getMostRecentResults() {
		return mostRecentResults;
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.IObjectActionDelegate#setActivePart(org.eclipse.jface.action.IAction, org.eclipse.ui.IWorkbenchPart)
	 */
	@Override
	public void setActivePart(IAction action, IWorkbenchPart targetPart) {
		shell = targetPart.getSite().getShell();
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.IActionDelegate#run(org.eclipse.jface.action.IAction)
	 */
	@Override
	public void run(IAction action) {
		Manager manager = Manager.getInstance();

		try {
			manager.setProject(target);
		} catch (ManagerException e) {
			MessageDialog.openError(shell, "Project Error", "Could not generate links for the selected project.");
			e.printStackTrace();
			return;
		}
		
		LinkRecoveryWizard wizard = new LinkRecoveryWizard();
		try {
			WizardDialog dialog = new WizardDialog(shell, wizard);
			dialog.create();
			dialog.open();
		} catch (Exception e) {
			//TODO improve
			e.printStackTrace();
		}
		
		mostRecentResults = wizard.getResults();
		TraceabilityView.getInstance().setInput(TraceabilityContentProvider.FOR_RESULTS);
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.IActionDelegate#selectionChanged(org.eclipse.jface.action.IAction, org.eclipse.jface.viewers.ISelection)
	 */
	@Override
	public void selectionChanged(IAction action, ISelection selection) {
		//Only handle structured selections
		if(selection.isEmpty() || 
				!(selection instanceof IStructuredSelection)) {
			return;
		}
		
		//Only accept first selection
		target = (IProject)((IStructuredSelection)selection).getFirstElement();
	}

}
