package traceability.recovery;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;

import org.eclipse.core.resources.IProject;

import traceability.links.Granularity;

/**
 * Special preprocessor for preprocessing textual documentation.  Certain Preprocessor methods 
 * (e.g., get/setGranularity()) have no effect.
 * 
 * @author Sam Klock
 *
 */
public class DocumentationPreprocessor extends Preprocessor {

	/**
	 * Textual identifier for this preprocessor.
	 */
	private static final String ID = "Documentation Preprocessor";
	
	/* User-specified settings */
	private IProject project;
	private StemmerSetting stemming = StemmerSetting.StemAll;
	private boolean nonLiterals = true;
	private boolean splitIDs = true;
	private String stopWordsFile = null;
	
	/**
	 * Stemmer used by the preprocessor
	 */
	private Stemmer stemmer = null;
	/**
	 * Stop words to remove
	 */
	private List<String> stopWords = new ArrayList<String>();
	/**
	 * Target directory containing documentation to preprocess
	 */
	private String target = null;
	/**
	 * Errors encountered during preprocessing
	 */
	private List<String> errors;
	
	/**
	 * Current text being preprocessed
	 */
	private String codeText;
	/**
	 * Scanner that reads documentation
	 */
	private Scanner codeReader;
	
	/**
	 * Constructor.
	 */
	public DocumentationPreprocessor() {
		stemmer = new Stemmer();
		errors = new ArrayList<String>();
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getRemoveNonLiterals()
	 */
	@Override
	public boolean getRemoveNonLiterals() {
		return nonLiterals;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getSplitIdentifiers()
	 */
	@Override
	public boolean getSplitIdentifiers() {
		return splitIDs;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getStemming()
	 */
	@Override
	public StemmerSetting getStemming() {
		return stemming;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getStopWordsFile()
	 */
	@Override
	public String getStopWordsFile() {
		if(stopWordsFile == null) {
			return "";
		}
		return stopWordsFile;
	}
	
	/**
	 * Sets the documentation location (the absolute filename of a directory) for this preproprocessor.
	 * 
	 * In the future, this method may be specified by an interface.
	 * 
	 * @param loc location of documentation to preprocess
	 */
	public void setDocumentationLocation(String loc) {
		target = loc;
	}
	
	/**
	 * Retrieves the documentation location (the absolute filename of a directory) for this preprocessor.
	 * 
	 * In the future, this method may be specified by an interface.
	 * 
	 * @return location of documentation to preprocess
	 */
	public String getDocumentationLocation() {
		if(target == null) {
			return "";
		}
		return target;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setRemoveNonLiterals(boolean)
	 */
	@Override
	public void setRemoveNonLiterals(boolean whether) {
		nonLiterals = whether;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setSplitIdentifiers(boolean)
	 */
	@Override
	public void setSplitIdentifiers(boolean whether) {
		splitIDs = whether;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setStemming(traceability.recovery.StemmerSetting)
	 */
	@Override
	public void setStemming(StemmerSetting setting) {
		stemming = setting;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setStopWordsFile(java.lang.String)
	 */
	@Override
	public void setStopWordsFile(String file) {
		stopWordsFile = file;
		stopWords = new ArrayList<String>();
		if(stopWordsFile == null) {
			return;
		}
		File theFile = new File(stopWordsFile);
		Scanner scanner = null;;
		try {
			scanner = new Scanner(theFile);
			while(scanner.hasNextLine()) {
				stopWords.add(scanner.nextLine().toLowerCase());
			}
			scanner.close();
		} catch (Exception e) {
			stopWords.clear();
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getProject()
	 */
	@Override
	public IProject getProject() {
		return project;
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setProject(org.eclipse.core.resources.IProject)
	 */
	@Override
	public void setProject(IProject project) {
		this.project = project;
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getErrors()
	 */
	@Override
	public List<String> getErrors() {
		return errors;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#preprocess()
	 */
	@Override
	public boolean preprocess() {
		if(!beginDocumentationPreprocessing()) {
			return false;
		}
		
		File docLocationDir = new File(target);
		
		//If we can't read the documentation directory, we can't proceed
		if(!docLocationDir.isDirectory() || !docLocationDir.canRead()) {
			addError("Documentation directory does not exist or cannot be read.");
			return false;
		}
		
		//Process each file in the directory
		for(String filename : docLocationDir.list()) {
			File docFile = new File(docLocationDir + File.separator + filename);
			if(docFile.canRead() && !docFile.isDirectory()) {
				//Read the file
				codeText = "";
				try {
					codeReader = new Scanner(docFile);
				} catch (FileNotFoundException e) {
					errors.add("Could not open \"" + docFile + "\" for reading.");
					return false;
				}
				while(codeReader.hasNext()) {
					String next = codeReader.next();
					if(!next.equals("\n")) {
						codeText += next + " ";
					}
				}
				codeReader.close();
				
				//We don't parse documentation, just process it
				processString(codeText.toCharArray());
				if(!flushBuffer(docLocationDir + File.separator + filename)) {
					return false;
				}
			}
		}
		
		endPreprocessing();
		return true;
	}
	
	/**
	 * Processes a string.  Output is produced via this method.
	 * @param source the sequence of characters comprising the string
	 */
	private void processString(char [] source) {
		if(nonLiterals) {
			for(int i = 0; i < source.length; i++) {
				if(!Character.isJavaIdentifierPart(source[i])) {
					source[i] = ' ';
				}
			}
		}
		
		String [] sourceWords = (new String(source)).split(" ");
		for(String s : sourceWords) {
			String [] sParts = {s};
			if(splitIDs) {
				sParts = splitIdentifier(s);
			}
			
			for(String sPart : sParts) {
				sPart = sPart.toLowerCase();
				if(stemming != StemmerSetting.NoStemming) {	
					sPart = stem(sPart);
				}
				if(isNotStopWord(sPart)) {
					addToBuffer(sPart.toLowerCase());
				}
			}
		}
	}
	
	/**
	 * Splits the supplied string using camel case and/or the underscore conventions.
	 * @param s string to split
	 * @return the string split into components
	 */
	private String[] splitIdentifier(String s) {
		//Code lifted from org.sesalab.coconut.parsing.JavaDocumentParser
		
		// "under_score" splitting -> "under score"
		s = s.replaceAll("([_])", " ").trim();

		// "camelCase" splitting -> "camel Case"
		s = s.replaceAll("([A-Z])", " $1").trim();
		return s.split(" ");
	}
	
	/**
	 * Stems the given word.
	 * @param word word to stem
	 * @return the stemmed version of the word
	 */
	private String stem(String word) {
		if(stemming == StemmerSetting.StemAll) {
			stemmer.add(word.toCharArray(), word.length());
			stemmer.stem();
			return new String(stemmer.toString());
		} else {
			//TODO implement check whether word is a noun
			stemmer.add(word.toCharArray(), word.length());
			stemmer.stem();
			return new String(stemmer.toString());
		}
	}
	
	/**
	 * Answers whether the given word is a stop word.
	 * @param word the word to check
	 * @return false if the word is a stop word
	 */
	private boolean isNotStopWord(String word) {
		if(stopWordsFile != null) {	//We are removing stop words
			if(stopWords.contains(word)) {
				return false;
			}
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#addError(java.lang.String)
	 */
	@Override
	protected void addError(String message) {
		errors.add(message);
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#clearErrors()
	 */
	@Override
	protected void clearErrors() {
		errors.clear();
	}

	/**
	 * Always returns null.  (Granularity has no meaning in documentation preprocessing.)
	 */
	@Override
	public Granularity getGranularity() {
		return null;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#id()
	 */
	@Override
	public String id() {
		return ID;
	}

	/**
	 * Doesn't do anything.  (Granularity has no meaning in documentation preprocessing.)
	 */
	@Override
	public void setGranularity(Granularity gran) {
		
	}
}
