package traceability.recovery;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Scanner;
import java.util.Set;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.ToolFactory;
import org.eclipse.jdt.core.compiler.IScanner;
import org.eclipse.jdt.core.compiler.ITerminalSymbols;
import org.eclipse.jdt.core.compiler.InvalidInputException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;

import traceability.links.Granularity;
import traceability.util.NameUtilities;

/**
 * Preprocessor for Java documents.
 * 
 * @author Sam Klock
 *
 */
public class JavaDocumentPreprocessor extends Preprocessor {

	/**
	 * Textual identifier for this preprocessor.
	 */
	private static final String ID = "Java Document Preprocessor";
	
	/* User-specified settings */
	private boolean nonLiterals = true;
	private boolean splitIDs = true;
	private boolean stemming = true;
	private boolean stemAllWords = true;
	private String  stopWordsFile = null;
	private Granularity granularity = Granularity.Package;
	
	/**
	 * Set of stop words to remove from artifacts.
	 */
	private Set<String> stopWords = new HashSet<String>();
	/**
	 * List of errors encountered during preprocessing.
	 */
	private List<String> errors = new ArrayList<String>();
	/**
	 * Project to preprocess.
	 */
	private IJavaProject project = null;
	/**
	 * Stemmer used during preprocessing.
	 */
	private Stemmer stemmer = null;
	
	/**
	 * Scanner responsible for reading code from source artifacts.
	 */
	private Scanner codeReader;
	/**
	 * Current text from source artifacts being processed.
	 */
	private String  codeText;
	/**
	 * Eclipse JDT used to help parse code.
	 */
	private IScanner codeScanner;
	
	/* State variables used to produce output. */
	private String currentPackageName;
	private String currentPath;
	private String currentType;
	private String currentMethod;
	private String currentMethodReturn;
	private String currentMethodArguments;
	
	/**
	 * Abstract syntax tree parser from the Eclipse JDT.
	 */
	private ASTParser parser;
	
	/**
	 * Constructor.
	 * @param nonLiterals whether to remove non-literals
	 * @param splitIDs whether to split identifiers
	 * @param stemming whether to stem words
	 * @param stemAllWords whether to stem all words
	 * @param stopWordsFile if given, file from which to draw stop words
	 * @param granularity granularity with which to view source
	 * @param docLocation location of documentation
	 * @param project Java project to preprocess
	 */
	public JavaDocumentPreprocessor() {
		codeScanner = ToolFactory.createScanner(true, false, true, false);
		parser = ASTParser.newParser(AST.JLS3);
		parser.setKind(ASTParser.K_COMPILATION_UNIT);
		stemmer = new Stemmer();
	}
	
	/**
	 * Retrieves stop words from the selected file.  If an exception occurs or the selected file is null, 
	 * the stop word list is cleared.
	 */
	private void getStopWords() {
		stopWords = new HashSet<String>();
		if(stopWordsFile == null) {
			return;
		}
		File theFile = new File(stopWordsFile);
		Scanner scanner = null;;
		try {
			scanner = new Scanner(theFile);
			while(scanner.hasNextLine()) {
				stopWords.add(scanner.nextLine().toLowerCase());
			}
			scanner.close();
		} catch (Exception e) {
			stopWords.clear();
		}
	}
	
	/**
	 * Preprocesses the artifacts in the given project, along with the selected documentation.
	 * @return true if successful; false otherwise
	 */
	public boolean preprocess() {
		errors = new ArrayList<String>();
		if(!beginSourcePreprocessing()) {
			return false;
		}
		
		//First: preprocess source
		try {
			for(IPackageFragment p : project.getPackageFragments()) {
				//Ignore packages from external libaries
				if(p.getResource() == null) {
					continue;
				}
				
				//Need fresh writers for this if granularity is package-level
				if(granularity == Granularity.Package) {
					return false;
				}

				//Handle the package
				if(!handlePackage(p)) {
					return false;
				}
			}
		} catch (JavaModelException e) {
			errors.add("JavaModelException: " + e.getMessage() + " (Is this a Java project?)");
			return false;
		}
			
		endPreprocessing();
		return true;
	}
	
	/**
	 * Iteratively handles the contents of the given package.
	 * @param thePackage the package
	 * @return true if successful
	 */
	private boolean handlePackage(IPackageFragment thePackage) {
		try {
			if(thePackage.isDefaultPackage()) {
				currentPackageName = null;
			} else {
				currentPackageName = thePackage.getElementName();
			}
			for(IJavaElement e : thePackage.getChildren()) {
				//We're only interested in compilation units
				if(e instanceof ICompilationUnit) {
					if(!handleCompilationUnit((ICompilationUnit)e)) {
						return false;
					}
				}
				
				if(granularity == Granularity.Package) {
					//Just handled a package: print its name
					if(!flushBuffer(currentPackageName == null ?
							NameUtilities.DEFAULT_PACKAGE_NAME : 
							currentPackageName)) {
						return false;
					}
				}
			}
			
		} catch (JavaModelException e) {
			errors.add("JavaModelException: " + e.getMessage() + " (Is this a Java project?)");
			return false;
		}
		return true;
	}
	
	/**
	 * Handles the contents of a compilation unit (file) according to the level of granularity at 
	 * which we are preprocessing.
	 * @param unit the compilation unit
	 * @return true if successful
	 */
	@SuppressWarnings("unchecked")
	private boolean handleCompilationUnit(ICompilationUnit unit) {
		//First: scan in the file
		currentPath = unit.getResource().getRawLocation().makeAbsolute().toOSString();
		try {
			codeReader = new Scanner(new File(currentPath));
		} catch(IOException e) {
			errors.add("Could not open \"" + currentPath + "\" for reading.");
			return false;
		}
		codeText = "";
		while(codeReader.hasNext()) {
			String next = codeReader.next();
			if(!next.equals("\n")) {
				codeText += next + " ";
			}
		}
		codeReader.close();
		
		//Then: handle according to granularity
		switch(granularity) {
		case Package:
			//Just parse
			parseString(codeText);
			break;
		case File:
			parseString(codeText);
			
			//Handled the file: map it
			if(!flushBuffer(currentPath)) {
				return false;
			}
			break;
		case Class:
		case Method:
			//Distinction between class and method handled in handleTypeDeclarations()
			//First, build an AST for the compilation unit.
			parser.setSource(codeText.toCharArray());
			CompilationUnit ast = (CompilationUnit)parser.createAST(null);
			
			//Get the declared types
			List children = ast.types();
			if(children == null || children.isEmpty()) {	//Nothing if there aren't any
				break;
			}
			
			//Pass the children down to handleTypeDeclarations (if appropriate)
			for(Object c : children) {
				ASTNode child = (ASTNode)c;
				if(child.getNodeType() == ASTNode.TYPE_DECLARATION) {
					handleTypeDeclarations((TypeDeclaration)child);
				}
			}
			break;
		}
		return true;
	}
	
	/**
	 * Handles a type declaration depending on the level of granularity at which we're preprocessing.
	 * @param node the type declaration
	 * @return true if successful
	 */
	private boolean handleTypeDeclarations(TypeDeclaration node) {
		currentType = node.getName().getFullyQualifiedName();
		if(granularity == Granularity.Class) {			
			//Parse it
			int start = node.getStartPosition();
			int length = node.getLength();
			String classContents = codeText.substring(start, start + length);
			parseString(classContents);
			
			//Handled a class: map it
			if(currentPackageName == null) {
				if(!flushBuffer(currentType)) {
					return false;
				}
			} else {
				if(!flushBuffer(currentPackageName + "." + currentType)) {
					return false;
				}
			}
		} else {	//granularity == Granularity.Method
			for(MethodDeclaration m : node.getMethods()) {
				//Parse the method and get its FQN
				currentMethod = m.getName().getFullyQualifiedName();
				currentMethodArguments = "";
				for(Object p : m.parameters()) {
					SingleVariableDeclaration d = (SingleVariableDeclaration)p;
					int typeStart = d.getType().getStartPosition();
					int typeEnd = d.getType().getLength() + typeStart;
					String type = codeText.substring(typeStart, typeEnd);
					currentMethodArguments += type + ", ";
				}
				if(currentMethodArguments.length() > 0) {
					currentMethodArguments = currentMethodArguments.substring(0, 
							currentMethodArguments.length() - 2);
				}
				Type ret = m.getReturnType2();
				if(ret == null) {
					currentMethodReturn = "";
				} else {
					int typeStart = ret.getStartPosition();
					int typeEnd = ret.getLength() + typeStart;
					currentMethodReturn = codeText.substring(typeStart, typeEnd);
				}
				int start = m.getStartPosition();
				int length = m.getLength();
				String methodContents = codeText.substring(start, start + length);
				parseString(methodContents);
				
				//Handled a method: map it
				if(currentPackageName == null) {
					if(!flushBuffer(currentMethodReturn + " " + currentType + "." + 
							currentMethod + "(" + currentMethodArguments + ")")) {
						return false;
					}
				} else {
					if(!flushBuffer(currentMethodReturn + " " + currentPackageName + "." + 
							currentType + "." + currentMethod + "(" + currentMethodArguments + ")")) {
						return false;
					}
				}
			}
		}
		return true;
	}
	
	/**
	 * Parses a given string using a Java scanner
	 * @param contents the string to parse
	 */
	private void parseString(String contents) {
		codeScanner.setSource(contents.toCharArray());
		try {
			int token = codeScanner.getNextToken();
			char [] source;
			while(token != ITerminalSymbols.TokenNameEOF) {
				source = codeScanner.getRawTokenSource();
				//Big switch statement to avoid ugly if-statement
				switch(token) {
				case ITerminalSymbols.TokenNameLongLiteral:
				case ITerminalSymbols.TokenNameFloatingPointLiteral:
				case ITerminalSymbols.TokenNameDoubleLiteral:
				case ITerminalSymbols.TokenNameCharacterLiteral:
				case ITerminalSymbols.TokenNameIntegerLiteral:
				case ITerminalSymbols.TokenNameIdentifier:
				case ITerminalSymbols.TokenNameCOMMENT_BLOCK:
				case ITerminalSymbols.TokenNameCOMMENT_JAVADOC:
				case ITerminalSymbols.TokenNameCOMMENT_LINE:
				case ITerminalSymbols.TokenNameStringLiteral:
					processString(source);
					break;
				}
				
				token = codeScanner.getNextToken();
			}
		} catch (InvalidInputException e) {
			//Lexical error: fall back to a scanner to handle the rest of the input
			errors.add("Warning: lexical error in " + currentPath + "; preprocessing output may be incorrect.");
			
			String toScan = contents.substring(codeScanner.getCurrentTokenEndPosition() + 1);
			Scanner scanner = new Scanner(toScan);
			while(scanner.hasNext()) {
				processString(scanner.next().toCharArray());
			}
		}
	}
	
	/**
	 * Processes a string.  Output is produced via this method.
	 * @param source the sequence of characters comprising the string
	 */
	private void processString(char [] source) {
		if(nonLiterals) {
			for(int i = 0; i < source.length; i++) {
				if(!Character.isJavaIdentifierPart(source[i])) {
					source[i] = ' ';
				}
			}
		}
		
		String [] sourceWords = (new String(source)).split(" ");
		for(String s : sourceWords) {
			String [] sParts = {s};
			if(splitIDs) {
				sParts = splitIdentifier(s);
			}
			
			for(String sPart : sParts) {
				sPart = sPart.toLowerCase();
				if(stemming) {	
					sPart = stem(sPart);
				}
				if(isNotStopWord(sPart)) {
					addToBuffer(sPart.toLowerCase());
				}
			}
			
		}
	}
	
	/**
	 * Splits the supplied string using camel case and/or the underscore conventions.
	 * @param s string to split
	 * @return the string split into components
	 */
	private String[] splitIdentifier(String s) {
		//Code lifted from org.sesalab.coconut.parsing.JavaDocumentParser
		
		// "under_score" splitting -> "under score"
		s = s.replaceAll("([_])", " ").trim();

		// "camelCase" splitting -> "camel Case"
		s = s.replaceAll("([A-Z])", " $1").trim();
		return s.split(" ");
	}
	
	/**
	 * Stems the given word.
	 * @param word word to stem
	 * @return the stemmed version of the word
	 */
	private String stem(String word) {
		if(stemAllWords) {
			stemmer.add(word.toCharArray(), word.length());
			stemmer.stem();
			return new String(stemmer.toString());
		} else {
			//TODO implement check whether word is a noun
			stemmer.add(word.toCharArray(), word.length());
			stemmer.stem();
			return new String(stemmer.toString());
		}
	}
	
	/**
	 * Answers whether the given word is a stop word.
	 * @param word the word to check
	 * @return false if the word is a stop word
	 */
	private boolean isNotStopWord(String word) {
		if(stopWordsFile != null) {	//We are removing stop words
			if(stopWords.contains(word)) {
				return false;
			}
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getGranularity()
	 */
	@Override
	public Granularity getGranularity() {
		return granularity;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getProject()
	 */
	@Override
	public IProject getProject() {
		return project.getProject();
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getRemoveNonLiterals()
	 */
	@Override
	public boolean getRemoveNonLiterals() {
		return nonLiterals;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getSplitIdentifiers()
	 */
	@Override
	public boolean getSplitIdentifiers() {
		return splitIDs;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getStemming()
	 */
	@Override
	public StemmerSetting getStemming() {
		if(stemming) {
			if(stemAllWords) {
				return StemmerSetting.StemAll;
			} else {
				return StemmerSetting.StemNouns;
			}
		}
		return StemmerSetting.NoStemming;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getStopWordsFile()
	 */
	@Override
	public String getStopWordsFile() {
		if(stopWordsFile != null) {
			return stopWordsFile;
		} else {
			return "";
		}
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#id()
	 */
	@Override
	public String id() {
		return ID;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setGranularity(traceability.links.Granularity)
	 */
	@Override
	public void setGranularity(Granularity gran) {
		granularity = gran;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setProject(org.eclipse.core.resources.IProject)
	 */
	@Override
	public void setProject(IProject project) {
		if(project == null) {
			throw new NullPointerException();
		}
		try {
			this.project = (IJavaProject)project.getNature(JavaCore.NATURE_ID);
		} catch (CoreException e) {
			throw new IllegalArgumentException("Project must be a Java project");
		}
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setRemoveNonLiterals(boolean)
	 */
	@Override
	public void setRemoveNonLiterals(boolean whether) {
		nonLiterals = whether;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setSplitIdentifiers(boolean)
	 */
	@Override
	public void setSplitIdentifiers(boolean whether) {
		splitIDs = whether;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setStemming(traceability.recovery.StemmerSetting)
	 */
	@Override
	public void setStemming(StemmerSetting setting) {
		switch(setting) {
		case StemAll:
			stemming = true;
			stemAllWords = true;
			break;
		case StemNouns:
			stemming = true;
			stemAllWords = false;
			break;
		case NoStemming:
			stemming = false;
			stemAllWords = false;
			break;
		}
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#setStopWordsFile(java.lang.String)
	 */
	@Override
	public void setStopWordsFile(String file) {
		stopWordsFile = file;
		getStopWords();
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#getErrors()
	 */
	@Override
	public List<String> getErrors() {
		return errors;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#addError(java.lang.String)
	 */
	@Override
	protected void addError(String message) {
		errors.add(message);
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.Preprocessor#clearErrors()
	 */
	@Override
	protected void clearErrors() {
		errors.clear();
	}
}
