package traceability.recovery;

import traceability.links.Granularity;

/**
 * Parameters for Lucene.
 * 
 * There's just one parameter: maximum field length.
 * 
 * @author Sam Klock
 *
 */
public class LuceneRecoveryTechniqueParams implements RecoveryTechniqueParams {

	/* Constants pertaining to max field length. */
	private static final String MAX_FIELD_LENGTH_ID = "MFL";
	private static final String MAX_FIELD_LENGTH_NAME = "Maximum field length";
	private static final ParameterType MAX_FIELD_LENGTH_TYPE = ParameterType.Integer;
	private static final int MAX_FIELD_LENGTH_DEFAULT = 300;
	
	private int threshold = 100;
	private double minSimilarity = 0.80;
	private boolean useThreshold = true;
	
	/**
	 * Maximum field length value.
	 */
	private int value;
	/**
	 * Granularity for analysis.
	 */
	private Granularity granularity;
	
	/**
	 * Constructor.
	 */
	public LuceneRecoveryTechniqueParams() {
		value = MAX_FIELD_LENGTH_DEFAULT;
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechniqueParams#getLabelForParameter(java.lang.String)
	 */
	@Override
	public String getLabelForParameter(String id) {
		if(!MAX_FIELD_LENGTH_ID.equals(id)) {
			return null;
		}
		return MAX_FIELD_LENGTH_NAME;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechniqueParams#getParameterIDs()
	 */
	@Override
	public String[] getParameterIDs() {
		String [] ret = {MAX_FIELD_LENGTH_ID};
		return ret;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechniqueParams#getParameterValue(java.lang.String)
	 */
	@Override
	public Object getParameterValue(String id) {
		if(!MAX_FIELD_LENGTH_ID.equals(id)) {
			return null;
		}
		return value;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechniqueParams#getTypeForParameter(java.lang.String)
	 */
	@Override
	public ParameterType getTypeForParameter(String id) {
		if(!MAX_FIELD_LENGTH_ID.equals(id)) {
			return null;
		}
		return MAX_FIELD_LENGTH_TYPE;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechniqueParams#setValueForParameter(java.lang.String, java.lang.Object)
	 */
	@Override
	public void setValueForParameter(String id, Object val) {
		if(MAX_FIELD_LENGTH_ID.equals(id)) {
			value = (Integer)val;
		}
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechniqueParams#setGranularity(traceability.links.Granularity)
	 */
	@Override
	public void setGranularity(Granularity granularity) {
		this.granularity = granularity;
	}
	
	/**
	 * Retrieves the granularity for Lucene.
	 * @return the granularity
	 */
	public Granularity getGranularity() {
		return granularity;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechniqueParams#setMinimumSimilarity(double)
	 */
	@Override
	public void setMinimumSimilarity(double min) {
		minSimilarity = min;
		
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechniqueParams#setThreshold(int)
	 */
	@Override
	public void setThreshold(int threshold) {
		this.threshold = threshold;
	}
	
	/**
	 * Retrieves the minimum number of links for Lucene to retrieve.
	 * @return minimum number of links
	 */
	public int getThreshold() {
		return threshold;
	}
	
	/**
	 * Retrieves the minimum similarity of the links that Lucene should retrieve.
	 * @return minimum similarity
	 */
	public double getMinimumSimilarity() {
		return minSimilarity;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechniqueParams#setWhetherThreshold(boolean)
	 */
	@Override
	public void setWhetherThreshold(boolean whether) {
		useThreshold = whether;
	}
	
	/**
	 * Determines whether Lucene should use the threshold or minimum similarity values 
	 * specified in these parameters.
	 * @return if true, use the threshold values; else, use the minimum similarity values
	 */
	public boolean getWhetherThreshold() {
		return useThreshold;
	}

}
