package traceability.recovery;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.List;

import org.eclipse.core.resources.IProject;

import traceability.links.Granularity;
import traceability.management.Manager;

/**
 * <p>Defines an interface for preprocessors for various source artifacts.  The abstract Preprocessor class 
 * provides built-in functionality to help ensure consistency in the output of all implemented 
 * preprocessors.  In particular, all preprocessors are expected to deposit output in a set of 
 * uniformly named files (with conventions given by <code>SOURCE_DIR</code>, <code>SOURCE_PREFIX</code></p>, 
 * <code>SUFFIX</code>, and <code>MAPPER_NAME</code>), which may then be accessed in a preprocessor-blind 
 * manner by recovery techniques.  This class handles writing output to disk; subclasses are expected 
 * to handle the tasks of accessing source artifacts inside projects and processing them appropriately.</p> 
 * 
 * <p>Classes extending Preprocessor to add support for new languages/project types are expected to rely 
 * on the following methods for producing output:</p>
 * 
 * <ul>
 * <li><code>beginSourcePreprocessing()</code>, which must be called when preprocessing begins.</li>
 * <li><code>addToBuffer()</code>, to add strings to the buffer for the current source artifact.</li>
 * <li><code>flushBuffer()</code>, to empty the buffer for the current source artifact and open a new 
 * buffer for the next.</li>
 * <li><code>endPreprocessing()</code>, to end preprocessing.
 * </ul>
 * 
 * <p>The Preprocessor class supports only one preprocessing session at a time.  While this class is also 
 * used to implement the hard-coded documentation preprocessor for Traceclipse, the plugin currently does 
 * not support external implementations of new documentation preprocessors.</p> 
 * 
 * @author Sam Klock
 * @since 0.2.0
 */
public abstract class Preprocessor {
	
	/**
	 * Error message produced when attempting to begin a new preprocessing session while preprocessing 
	 * is already in progress.
	 */
	private static final String ILLEGAL_STATE = "Already preprocessing";
	
	/**
	 * Global manager instance.
	 */
	private static Manager manager = Manager.getInstance();
	
	/**
	 * Buffer used to store output for the current artifact.
	 */
	private StringBuffer buffer;
	/**
	 * Print stream for the mapper file for the current preprocessing session.
	 */
	private PrintStream mapperOut;
	/**
	 * Integer identifier for the current artifact.
	 */
	private int id;
	/**
	 * Whether we are preprocessing for source or documentation.
	 */
	private boolean source;
	
	/**
	 * Subdirectory under [project_directory]/[traceability_directory] in which the output of 
	 * preprocessing for source artifacts is stored.
	 */
	public static final String SOURCE_DIR = "ppsource";
	/**
	 * Subdirectory under [project_directory]/[traceability_directory] in which the output of 
	 * preprocessing for documentation artifacts is stored.
	 */
	public static final String DOC_DIR = "ppdoc";
	/**
	 * The string prepended to the names of files storing the output of preprocessing for 
	 * a particular source artifact inside <code>SOURCE_DIR</code>.
	 */
	public static final String SOURCE_PREFIX = "source";
	/**
	 * The string prepended to the names of files storing the output of preprocessing for 
	 * a particular documentation artifact inside <code>DOC_DIR</code>.
	 */
	public static final String DOC_PREFIX = "doc";
	/**
	 * The string appended to the names of files storing the output of preprocessing for 
	 * a particular source or documentation artifact inside <code>SOURCE_DIR</code> or 
	 * <code>DOC_DIR</code> respectively.
	 */
	public static final String SUFFIX = ".txt";
	/**
	 * The name of the file inside either <code>SOURCE_DIR</code> or <code>DOC_DIR</code> mapping 
	 * output files to the artifacts from which they were produced.
	 */
	public static final String MAPPER_NAME = "mapper.txt";
	
	/**
	 * Initiates preprocessing as the preprocessor is parameterized.
	 * 
	 * If errors occur, appropriate messages are expected to be added to a list of errors (via 
	 * <code>addError()</code>, which callers may access via <code>getErrors()</code>.  Subclasses 
	 * must implement both of these methods.
	 *  
	 * @return true if successful; false otherwise
	 * @see getErrors()
	 */
	public abstract boolean preprocess();
	/**
	 * Retrieves a list of textual descriptions of errors that occurred during the previous 
	 * preprocessing session.
	 * 
	 * This method should never return null; if no errors have been recorded, it should return 
	 * an empty list.
	 * @return a list of errors that occurred during the previous preprocessing session
	 */
	public abstract List<String> getErrors();
	/**
	 * Adds an error message to the list of textual error messages.
	 * 
	 * @param message the message to add to the list
	 */
	protected abstract void addError(String message);
	/**
	 * Clears the list of error messages.
	 * 
	 * Following a call to this method, it must be possible to cleanly add a message to the error list 
	 * via addError().
	 */
	protected abstract void clearErrors();
	/**
	 * Retrieves a textual identifier for this preprocessor implementation.
	 * @return a textual, human-readable identifier
	 */
	public abstract String id();
	
	/**
	 * Sets the project for which preprocessing is to take place.
	 * @param project the project to preprocess
	 */
	public abstract void setProject(IProject project);
	/**
	 * Sets the granularity with which to view source artifacts.
	 * 
	 * For documentation preprocessors, this method presumably has no meaning.
	 * @param gran the granularity with which to view source artifacts.
	 * @see traceability.links.Granularity
	 */
	public abstract void setGranularity(Granularity gran);
	/**
	 * Sets whether to remove non-literals during preprocessing.  If <code>whether</code> is true, 
	 * the preprocessor should subsequently remove non-literals; otherwise, it should not.
	 * @param whether whether to remove non-literals
	 */
	public abstract void setRemoveNonLiterals(boolean whether);
	/**
	 * Sets whether to split identifiers during preprocessing.  If <code>whether</code> is true, 
	 * the preprocessor should subsequently split identifiers appropriately; otherwise, it should not.
	 * 
	 * The discipline with which preprocessors split identifiers is up to the implementation.  The 
	 * supplied JavaDocumentPreprocessor, for example, follows both the camel-case and underscore identifier 
	 * conventions.
	 * @param whether whether to split identifiers
	 */
	public abstract void setSplitIdentifiers(boolean whether);
	/**
	 * Sets the level of stemming to apply to terms appearing in documents.
	 * @param setting level of stemming
	 * @see traceability.recovery.StemmerSetting
	 */
	public abstract void setStemming(StemmerSetting setting);
	/**
	 * Sets the file from which to draw stop words.
	 * 
	 * If the given filename cannot be accessed, preprocessors are expected <em>not</em> to fail.  
	 * A message may be added to the list of errors.  If <code>file</code> is null, then the preprocessor 
	 * should treat it as an empty file (i.e., no stop words).
	 * @param file name of a file from which to draw stop words, or null if no stop words are desired
	 */
	public abstract void setStopWordsFile(String file);
	
	/**
	 * Retrieves the project the preprocessor is set to process.
	 * @return the target project
	 */
	public abstract IProject getProject();
	/**
	 * Retrieves the granularity with which the preprocessor is set to view source artifacts.
	 * 
	 * For documentation preprocessors, this method presumably has no meaning.  The supplied implementation, 
	 * DocumentationPreprocessor, always returns null.
	 * @return the granularity
	 */
	public abstract Granularity getGranularity();
	/**
	 * Retrieves whether the preprocessor is set to remove non-literals from documents during preprocessing.
	 * @return true if the preprocessor is set to remove non-literals; false otherwise
	 * @see traceability.links.Granularity
	 */
	public abstract boolean getRemoveNonLiterals();
	/**
	 * Retrieves whether the preprocessor is set to split identifiers from documents during preprocessing.
	 * @return true if the preprocessor is set to split identifiers; false otherwise
	 */
	public abstract boolean getSplitIdentifiers();
	/**
	 * Retrieves the level of stemming the preprocessor is set to apply to documents during preprocessing.
	 * @return the level of stemming
	 * @see traceability.recovery.StemmerSetting
	 */
	public abstract StemmerSetting getStemming();
	/**
	 * Retrieves the name of the file from which stop words are drawn.
	 * @return filename of the stop words file
	 */
	public abstract String getStopWordsFile();
	
	/**
	 * Public utility method used to retrieve the absolute path to the directory in which 
	 * preprocessed source artifacts are stored.
	 * 
	 * @param project the project for which the source directory is desired
	 * @return absolute path to the directory in which preprocessed source artifacts are stored
	 */
	public static final String getSourceDirectory(IProject project) {
		return manager.getDirPath(project) + File.separator + SOURCE_DIR;
	}
	
	/**
	 * Public utility method used to retrieve the absolute path to the directory in which 
	 * preprocessed documentation artifacts are stored.
	 * 
	 * @param project the project for which the documentation directory is desired
	 * @return absolute path to the directory in which preprocessed documentation artifacts are stored
	 */
	public static final String getDocumentationDirectory(IProject project) {
		return manager.getDirPath(project) + File.separator + DOC_DIR;
	}
	
	/**
	 * Public utility method used to retrieve the absolute path to the file in which a  
	 * preprocessed source artifact identified by <code>line</code> is stored.  If <code>line</code> 
	 * is zero, then the name of the mapper file is returned.
	 * 
	 * @param project the project for which the source filename is desired
	 * @return absolute path to the file in which the identified preprocessed source artifact is stored, or 
	 * to the mapper file if <code>line</code> is zero
	 */
	public static final String getAbsoluteFilenameForSource(IProject project, int line) {
		if(line > 0) {
			return getSourceDirectory(project) + File.separator + SOURCE_PREFIX + line + SUFFIX;
		} else {
			return getSourceDirectory(project) + File.separator + MAPPER_NAME;
		}
	}
	
	/**
	 * Public utility method used to retrieve the absolute path to the file in which a  
	 * preprocessed documentation artifact identified by <code>line</code> is stored.  If <code>line</code> 
	 * is zero, then the name of the mapper file is returned.
	 * 
	 * @param project the project for which the documentation filename is desired
	 * @return absolute path to the file in which the identified preprocessed documentation artifact is 
	 * stored, or to the mapper file if <code>line</code> is zero
	 */
	public static final String getAbsoluteFilenameForDocumentation(IProject project, int line) {
		if(line > 0) {
			return getDocumentationDirectory(project) + File.separator + DOC_PREFIX + line + SUFFIX;
		} else {
			return getDocumentationDirectory(project) + File.separator + MAPPER_NAME;
		}
	}
	
	/**
	 * Sets the preprocessor up for the first source artifact during preprocessing.  Subclasses 
	 * must call this method when beginning to preprocess source artifacts in a project.
	 * 
	 * Before this method can be called again, subclasses must call endPreprocessing(), presumably 
	 * after handling all source artifacts in a project.
	 * 
	 * If the method does not finish successfully, it will return false and add an error message 
	 * to the list of errors.
	 * 
	 * @return true if preprocessing has started successfully; false otherwise
	 */
	protected boolean beginSourcePreprocessing() {
		clearErrors();
		if(id != 0) {
			addError(ILLEGAL_STATE);
			return false;
		}
		id = 1;
		source = true;
		initBuffer();
		if(!clearTargetDirectory()) {
			return false;
		}
		try {
			mapperOut = new PrintStream(new File(getAbsoluteFilenameForSource(getProject(), 0)));
		} catch (FileNotFoundException e) {
			addError("I/O error: " + e.getMessage());
			return false;
		}
		
		return true;
	}
	
	/**
	 * Sets the preprocessor up for the first documentation artifact during preprocessing.  Subclasses 
	 * must call this method when beginning to preprocess documentation artifacts in a project.
	 * 
	 * Before this method can be called again, subclasses must call endPreprocessing(), presumably 
	 * after handling all documentation artifacts in a project.
	 * 
	 * If the method does not finish successfully, it will return false and add an error message 
	 * to the list of errors.
	 * 
	 * @return true if preprocessing has started successfully; false otherwise
	 */
	protected boolean beginDocumentationPreprocessing() {
		clearErrors();
		if(id != 0) {
			addError(ILLEGAL_STATE);
			return false;
		}
		id = 1;
		source = false;
		initBuffer();
		if(!clearTargetDirectory()) {
			return false;
		}
		try {
			mapperOut = new PrintStream(new File(getAbsoluteFilenameForDocumentation(getProject(), 0)));
		} catch (FileNotFoundException e) {
			addError("I/O error: " + e.getMessage());
			return false;
		}
		
		return true;
	}
	
	/**
	 * Returns true if the preprocessor is currently preprocessing.
	 * @return true if preprocessing; false otherwise
	 */
	protected boolean isPreprocessing() {
		return (id > 0);
	}
	
	/**
	 * Finishes the current preprocessing session.  All subclasses should call this method when finished 
	 * preprocessing.
	 */
	protected void endPreprocessing() {
		id = 0;
		if(mapperOut != null) {
			mapperOut.close();
			mapperOut = null;
		}
	}
	
	/**
	 * Adds the given string of characters to the preprocessing output buffer for the current artifact.
	 * 
	 * This method, along with flushBuffer(), constitutes the means by which subclasses produce output 
	 * usable by recovery techniques.
	 * 
	 * @param characters the characters to add to the output buffer
	 */
	protected void addToBuffer(String characters) {
		if(id == 0) {
			//Fail fast
			return;
		}
		buffer.append(characters);
		buffer.append(' ');
	}
	
	/**
	 * Flushes the output buffer for the current artifact to disk and initializes a new buffer for 
	 * the next artifact.
	 * 
	 * If flushBuffer() is not successful, it will return false and add a message to the textual list 
	 * of error messages.
	 * 
	 * @param artifactID string identifying the current artifact (for which the buffer is being flushed)
	 * @return true if successful; false otherwise
	 */
	protected boolean flushBuffer(String artifactID) {
		if(id == 0) {
			addError("Not preprocessing; no buffer to flush");
			return false;
		}
		PrintStream bufferOut = null;
		
		try {
			if(source) {
				bufferOut = new PrintStream(new File(getAbsoluteFilenameForSource(getProject(), id)));
			} else {
				bufferOut = new PrintStream(new File(getAbsoluteFilenameForDocumentation(getProject(), id)));
			}
		} catch(FileNotFoundException e) {
			addError("I/O error: " + e.getMessage());
			return false;
		}
		
		bufferOut.print(buffer.toString());
		bufferOut.close();
		
		mapperOut.println(artifactID);
		id++;
		initBuffer();
		return true;
	}
	
	/**
	 * Clears the target directory for source or documentation preprocessing, as appropriate.
	 * 
	 * An error is added via addError() if unsuccessful.
	 * @return true if successful; false otherwise.
	 */
	private boolean clearTargetDirectory() {
		File targetDir = null;
		
		if(source) {
			targetDir = new File(getSourceDirectory(getProject()));
		} else {
			targetDir = new File(getDocumentationDirectory(getProject()));
		}
		
		if(!targetDir.isDirectory()) {
			if(!targetDir.mkdir()) {
				if(source) {
					addError("Could not create output directory for source preprocessing.");
				} else {
					addError("Could not create output directory for directory preprocessing.");
				}
				return false;
			}
		} else {
			for(File f : targetDir.listFiles()) {
				if(!f.delete()) {
					if(source) {
						addError("Could not clear output directory for source preprocessing.");
					} else {
						addError("Could not clear output directory for source preprocessing.");
					}
					return false;
				}
			}
		}
		
		return true;
	}
	
	/**
	 * Initializes a new buffer.
	 */
	private void initBuffer() {
		buffer = new StringBuffer();
	}
}
