package traceability.util;

import java.util.LinkedList;
import java.util.List;
import java.util.Scanner;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.Signature;

import traceability.management.Manager;

/**
 * Provides a set of methods for transforming artifact names into Eclipse resources and back.  Each 
 * method in his class assumes that the artifact 
 * 
 * Currently, this class only supports Java.  Support may be added for Eclipse toolkits on other languages 
 * in the future.
 * 
 * Each method presupposes that the artifact being referenced belongs to the current project in the Manager.  
 * Callers should ensure that the current project is set correctly before using these methods.
 *   
 * @author Sam Klock
 * @see traceability.management.Manager
 */
public final class NameUtilities {

	/**
	 * Manager from which to extract projects.
	 */
	private static Manager manager = Manager.getInstance();
	/**
	 * Name to give default packages (in Java).
	 */
	public static final String DEFAULT_PACKAGE_NAME = "(default package)";
	
	/**
	 * Constructor.  This class is not instantiable or extendable; all methods are intended to be static.
	 */
	private NameUtilities() {
		//NOT INSTANTIABLE
	}
	
	/**
	 * Retrieves the Java project IMethod for the given method name.
	 * 
	 * The input string is expected to be of the form 
	 * "<code>returType package.class.method(paramType, paramType, ...)<code>". 
	 * @param methodString method name
	 * @return IMethod corresponding to the method name
	 */
	public static IMethod getJavaMethodFor(String methodString) {
		IProject project = manager.getProject();
		IJavaProject javaProject = JavaCore.create(project);
		if(javaProject == null) {
			return null;
		}
		
		String className = extractClassNameFromMethod(methodString);
		IType theClass = getJavaClassFor(className);
		
		if(theClass == null) {
			return null;
		}
		
		List<String> parameters = getParameterTypesForMethod(methodString);
		String [] parameterArray = parameters.toArray(new String[0]);
		
		IMethod method = theClass.getMethod(extractMethodName(methodString), parameterArray);
		
		return method;
	}
	
	/**
	 * Retrieves the Java class (or interface) IType for the given class name.
	 * 
	 * The input string is expected to be of the form 
	 * "<code>package.className</code>".
	 * 
	 * @param classString fully-qualified class/interface name
	 * @return IType corresponding to the class name
	 */
	public static IType getJavaClassFor(String classString) {
		IProject project = manager.getProject();
		IJavaProject javaProject = JavaCore.create(project);

		if(javaProject == null) {
			return null;
		}
		
		IType theClass = null;
		try {
			theClass = javaProject.findType(classString);
		} catch(JavaModelException e) {
			//FIXME
			e.printStackTrace();
		}
		
		return theClass;
	}
	
	/**
	 * Retrieves the Java package IPackageFragment for the given package name.
	 * 
	 * The input string is expected to be of the form "<code>package1.package2.</code>".
	 * @param packageString package name
	 * @return IPackageFragment for the given package name
	 */
	public static IPackageFragment getJavaPackageFor(String packageString) {
		IProject project = manager.getProject();
		IJavaProject javaProject = JavaCore.create(project);
		if(javaProject == null) {
			return null;
		}
		
		IPath path = Path.fromPortableString(packageToPortablePath(packageString, project));
		try {
			return javaProject.findPackageFragment(path);
		} catch(JavaModelException e) {
			//FIXME
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Computes a fully-qualified method name for the given Java IMethod.
	 * @param method method for which to compute the FQN
	 * @return the FQN of the given method
	 */
	public static String getNameForJavaMethod(IMethod method) {
		IJavaElement parent = method.getParent();
		String retMethod = method.getElementName();
		while(parent != null && !(parent instanceof ICompilationUnit)) {
			retMethod = parent.getElementName() + "." + retMethod;
			parent = parent.getParent();
		}
		if(parent instanceof ICompilationUnit) {
			parent = parent.getParent();
			if(parent instanceof IPackageFragment &&
					!((IPackageFragment)parent).isDefaultPackage()) {
				retMethod = parent.getElementName() + "." + retMethod;
			}
		}
		
		try {
			retMethod = Signature.toString(method.getReturnType()) + " " + retMethod + "(";
			if(method.getNumberOfParameters() > 0) {
				for(String s : method.getParameterTypes()) {
					retMethod += Signature.toString(s) + ", ";
				}
				retMethod = retMethod.substring(0, retMethod.length() - 2) + ")";
			} else {
				retMethod += ")";
			}
		} catch (JavaModelException e) {
			e.printStackTrace(); //TODO improve handling
			retMethod += "EXCEPTION!!!";
		}
		return retMethod;
	}
	
	/**
	 * Retrieves the fully-qualified class/interface name for the given IType.
	 * @param theClass IType (class/interface) for which to compute an FQN
	 * @return the FQN of the type
	 */
	public static String getNameForJavaClass(IType theClass) {
		String retClass = theClass.getElementName();
		IJavaElement theParent = theClass.getParent();
		while(theParent != null && !(theParent instanceof ICompilationUnit)) {
			retClass = theParent.getElementName() + "." + retClass;
			theParent = theParent.getParent();
		}
		if(theParent instanceof ICompilationUnit) {
			theParent = theParent.getParent();
			if(theParent instanceof IPackageFragment &&
					!((IPackageFragment)theParent).isDefaultPackage()) {
				retClass = theParent.getElementName() + "." + retClass;
			}
		}
		return retClass;
	}
	
	/**
	 * Retrieves the fully-qualified package name for the given package IPackageFragment.
	 * @param thePackage package for which to compute an FQN
	 * @return FQN of the package
	 */
	public static String getNameForJavaPackage(IPackageFragment thePackage) {
		String retPackage = thePackage.getElementName();
		if(retPackage.equals("")) { //default package name
			retPackage = DEFAULT_PACKAGE_NAME;
		}
		return retPackage;
	}
	
	/**
	 * Helper method for getJavaMethodFor; pulls the fully-qualified class name from the 
	 * method's name.
	 * @param methodName method's FQN
	 * @return class name
	 */
	private static String extractClassNameFromMethod(String methodName) {
		int start = -1, end = -1;
		for(int i = 0; i < methodName.length(); i++) {
			if(start == -1 && methodName.charAt(i) == ' ') {
				start = ++i;
			} else if(start != -1 && methodName.charAt(i) == '('){
				while(methodName.charAt(i) != '.') {
					i--;
					if(i < 0) {
						//THIS IS BAD
						return "";
					}
				}
				end = i;
				break;
			}
		}
		if(start == -1 || end == -1 || end <= start) {
			//THIS IS BAD
			return "";
		}
		return methodName.substring(start, end);
	}
	
	/**
	 * Helper for getJavaMethodFor; pulls the simple method name from the method's 
	 * fully-qualified name.
	 * @param fqn method's fully-qualified name
	 * @return method's simple name
	 */
	private static String extractMethodName(String fqn) {
		int start = -1, end = -1;
		for(int i = 0; i < fqn.length(); i++) {
			if(start == -1 && fqn.charAt(i) == '(') {
				while(fqn.charAt(i) != '.') {
					i--;
				}
				start = i;
			} else if(start != -1 && fqn.charAt(i) == ')') {
				end = i;
			}
		}
		
		if(start == -1 || end == -1 || end <= start) {
			return "";
		}
		return fqn.substring(start, end);
	}
	
	/**
	 * Helper for getJavaMethodFor; pulls a list of parameters from the method's 
	 * fully-qualified name. 
	 * @param method method's FQN
	 * @return list of parameter types as a string
	 */
	private static List<String> getParameterTypesForMethod(String method) {
		LinkedList<String> list = new LinkedList<String>();
		
		int start;
		int end = -1;
		for(start = 0; start < method.length(); start++) {
			if(end == -1 && method.charAt(start) == '(') {
				start++;
				end = start;
				break;
			} else if(end != -1 && method.charAt(start) != ' ') {
				for( ; end < method.length(); end++) {
					if(method.charAt(end) == ',' || method.charAt(end) == ')') {
						list.add(method.substring(start, end));
						start = end;
					}
				}
			}
		}
		
		return list;		
	}
	
	/**
	 * Helper method for getJavaPackageFor; converts a package's FQN into a portable string, 
	 * used to represent the package as a path.
	 * @param packageName FQN of the package
	 * @param project project for the package
	 * @return package name represented as a portable path
	 */
	private static String packageToPortablePath(String packageName, IProject project) {
		String path = project.getProjectRelativePath().toPortableString();
		if(packageName.equals(DEFAULT_PACKAGE_NAME)) {
			return path;
		}
		Scanner scanner = new Scanner(packageName);
		scanner.useDelimiter(".");
		while(scanner.hasNext()) {
			path += "/" + scanner.next();
		}
		return path;
	}
}
