package traceability.views;

import java.util.Set;

import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.Viewer;

import traceability.links.TraceabilityLink;
import traceability.management.Manager;

/**
 * 
 * Content provider used to add links to the Traceability Links View.
 * 
 * The input elements for this provider take the form of strings identifying the kinds of links to use to 
 * populate the view.  There are currently four classes of input strings: one for projects (i.e., view 
 * all links for the current project), one for documentation (i.e., view all links for a particular 
 * documentation artifact), one for source (i.e., view all links for a particular source artifact), 
 * and one for the results of the most recent traceability links recovery operation.
 * 
 * @author Sam Klock
 *
 */
public class TraceabilityContentProvider implements IStructuredContentProvider {

	/**
	 * Input element prefix for projects.
	 */
	public static final String FOR_PROJECT = "***PROJECT";
	/**
	 * Input element prefix for documentation.
	 */
	public static final String FOR_DOC     = "***DOC:";
	/**
	 * Input element prefix for source artifacts.
	 */
	public static final String FOR_SRC     = "***SRC:";
	/**
	 * Input element prefix for recovery results.
	 */
	public static final String FOR_RESULTS = "***RESULTS";
	
	/**
	 * Manager instance (used to collect links to push into the view).
	 */
	private Manager manager;
	
	/**
	 * Constructor.
	 */
	public TraceabilityContentProvider() {
		manager = Manager.getInstance();
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(java.lang.Object)
	 */
	@Override
	public Object[] getElements(Object inputElement) {
		Set<TraceabilityLink> linkSet = null;
		
		String input = (String)inputElement;
		
		if(inputElement.equals(FOR_PROJECT)) {
			linkSet = manager.getAllLinks();
		} else if(inputElement.equals(FOR_RESULTS)) {
			linkSet = manager.getLastGenerationResults();
	    } else if(input.startsWith(FOR_DOC)) {
			linkSet = manager.getLinksForDocument(input.substring(FOR_DOC.length()));
		} else if(input.startsWith(FOR_SRC)) {
			linkSet = manager.getLinksForSource(input.substring(FOR_SRC.length()));
		}
		
		if(linkSet == null) {
			return new Object[0];
		} else {
			Object [] ret = new Object[linkSet.size()];
			int i = 0;
			for(TraceabilityLink l : linkSet) {
				ret[i++] = l;
			}
			return ret;
		}
	}

	/*
	 * Doesn't do anything.
	 * (non-Javadoc)
	 * @see org.eclipse.jface.viewers.IContentProvider#dispose()
	 */
	@Override
	public void dispose() {
		//No state of which to dispose
	}

	/*
	 * Doesn't do anything.
	 * (non-Javadoc)
	 * @see org.eclipse.jface.viewers.IContentProvider#inputChanged(org.eclipse.jface.viewers.Viewer, java.lang.Object, java.lang.Object)
	 */
	@Override
	public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		//No state to change
	}

}
