package traceability.views;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Arrays;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Scanner;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.TableEditor;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Link;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.part.ViewPart;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerComparator;

import traceability.links.Granularity;
import traceability.links.TraceabilityLink;
import traceability.management.Manager;
import traceability.util.TraceabilitySelectionListener;


/**
 * 
 * Implementation of the Traceability Links View.
 * 
 * @author Sam Klock
 *
 */
public class TraceabilityView extends ViewPart {

	/**
	 * The ID of the view as specified by the extension.
	 */
	public static final String ID = "traceability.views.TraceabilityView";
	
	/* Presentation constants */
	private static final int PREVIEW_LENGTH = 20;
	private static final int DOCUMENTATION_COLUMN = 2;
	private static final int SOURCE_COLUMN = 1;
	private static final int REMOVE_COLUMN = 4;
	private static final int LINK_OFFSET = 10;
	private static final int REMOVE_VERT_OFFSET = 2;

	/* UI elements */
	private TableViewer viewer;
	private static TraceabilityView instance;
	private List<Button> removeButtons;
	private List<Link> links;
	private List<TableEditor> editors;
	private TableColumn sourceColumn;
	private TableColumn docColumn;
	
	private ViewerComparator granComparator = new ViewerComparator() {
		
		@Override
		public void sort(Viewer view, Object [] elements) {
			Arrays.sort(elements, new Comparator<Object>() {

				@Override
				public int compare(Object arg0, Object arg1) {
					TraceabilityLink t0 = (TraceabilityLink)arg0;
					TraceabilityLink t1 = (TraceabilityLink)arg1;
					return t0.getSourceGranularity().compareTo(t1.getSourceGranularity());
				}
				
			});
		}
	};
	
	private ViewerComparator sourceComparator = new ViewerComparator() {
		
		@Override
		public void sort(Viewer view, Object [] elements) {
			Arrays.sort(elements, new Comparator<Object>() {

				@Override
				public int compare(Object arg0, Object arg1) {
					TraceabilityLink t0 = (TraceabilityLink)arg0;
					TraceabilityLink t1 = (TraceabilityLink)arg1;
					return t0.getSource().compareTo(t1.getSource());
				}
				
			});
		}
	};
	
	private ViewerComparator docComparator = new ViewerComparator() {

		@Override
		public void sort(Viewer view, Object [] elements) {
			Arrays.sort(elements, new Comparator<Object>() {

				@Override
				public int compare(Object arg0, Object arg1) {
					TraceabilityLink t0 = (TraceabilityLink)arg0;
					TraceabilityLink t1 = (TraceabilityLink)arg1;
					return t0.getDocument().compareTo(t1.getDocument());
				}
				
			});
		}
	};
	
	private ViewerComparator descComparator = new ViewerComparator() {

		@Override
		public void sort(Viewer view, Object [] elements) {
			Arrays.sort(elements, new Comparator<Object>() {
				
				@Override
				public int compare(Object arg0, Object arg1) {
					TraceabilityLink t1 = (TraceabilityLink)arg0;
					TraceabilityLink t2 = (TraceabilityLink)arg1;
					
					Scanner scanner = null;
					String t1desc = "";
					String t2desc = "";
					try {
						scanner = new Scanner(new File(t1.getDocument()));
					} catch(FileNotFoundException e) {
						t1desc = "(Documentation file not found)";
					}
					for(int i = 0; i < PREVIEW_LENGTH && scanner.hasNext(); i++) {
						t1desc += scanner.next() + " ";
					}
					scanner.close();
					
					try {
						scanner = new Scanner(new File(t2.getDocument()));
					} catch(FileNotFoundException e) {
						t2desc = "(Documentation file not found)";
					}
					for(int i = 0; i < PREVIEW_LENGTH && scanner.hasNext(); i++) {
						t2desc += scanner.next() + " ";
					}
					scanner.close();
					
					return t1desc.compareTo(t2desc);
				}
			});
		}
	};
	
	private ViewerComparator simComparator = new ViewerComparator() {
		@Override
		public void sort(Viewer view, Object [] elements) {
			Arrays.sort(elements, new Comparator<Object>() {

				@Override
				public int compare(Object arg0, Object arg1) {
					TraceabilityLink l0 = (TraceabilityLink)arg0;
					TraceabilityLink l1 = (TraceabilityLink)arg1;
					if(l0.getSimilarity() > l1.getSimilarity()) {
						return -1;
					} else if(l0.getSimilarity() < l1.getSimilarity()) {
						return 1;
					}
					return 0;
				}
				
			});
		}
	};
		
	
	/**
	 * Constructor.  This class is intended to be single-instance, but Eclipse needs access to a constructor.  
	 * Fortunately, it only builds one instance.
	 */
	public TraceabilityView() {
		instance = this;
	}
	
	/**
	 * Retrieves the table viewer for this instance.
	 * @return the table viewer
	 */
	public TableViewer getViewer() {
		return viewer;
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	@Override
	public void createPartControl(Composite parent) {
		viewer = new TableViewer(parent, SWT.V_SCROLL | SWT.H_SCROLL | SWT.BORDER);
		viewer.setContentProvider(new TraceabilityContentProvider());
		Table table = viewer.getTable();
		table.setHeaderVisible(true);
		table.setLinesVisible(true);
		
		generateColumns();
		
		GridData gridData = new GridData();
		gridData.verticalAlignment = GridData.FILL;
		gridData.horizontalSpan = 2;
		gridData.grabExcessHorizontalSpace = true;
		gridData.grabExcessVerticalSpace = true;
		gridData.horizontalAlignment = GridData.FILL;
		viewer.getControl().setLayoutData(gridData);
		removeButtons = new LinkedList<Button>();
		links = new LinkedList<Link>();
		editors = new LinkedList<TableEditor>();
	}
	
	/**
	 * Private helper method that builds the columns for the table viewer.
	 */
	private void generateColumns() {
		/* GRANULARITY COLUMN */
		TableViewerColumn granularity = generateColumn("Granularity");
		granularity.setLabelProvider(new ColumnLabelProvider() {
			
			@Override
			public String getText(Object element) {
				TraceabilityLink link = (TraceabilityLink)element;
				return link.getSourceGranularity().toString();
			}
		});
		granularity.getColumn().addSelectionListener(new TraceabilitySelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				viewer.setComparator(granComparator);
				recreateControls();
			}
			
		});
		
		/* PATH COLUMN */
		TableViewerColumn path = generateColumn("Path");
		path.setLabelProvider(new ColumnLabelProvider() {
			
			@Override
			public String getText(Object element) {
				return "";
			}
		});
		sourceColumn = path.getColumn();
		sourceColumn.setResizable(false);
		sourceColumn.addSelectionListener(new TraceabilitySelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				viewer.setComparator(sourceComparator);
				recreateControls();
			}
			
		});
		
		/* DOCUMENTATION COLUMN */
		TableViewerColumn documentation = generateColumn("Documentation");
		documentation.setLabelProvider(new ColumnLabelProvider() {
			
			@Override
			public String getText(Object element) {
				return "";
			}
		});
		docColumn = documentation.getColumn();
		docColumn.setResizable(false);
		docColumn.addSelectionListener(new TraceabilitySelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				viewer.setComparator(docComparator);
				recreateControls();
			}
			
		});
		
		/* DESCRIPTION COLUMN */
		TableViewerColumn description = generateColumn("Description");
		description.setLabelProvider(new ColumnLabelProvider(){
			
			@Override
			public String getText(Object element) {
				TraceabilityLink link = (TraceabilityLink)element;
				String doc = link.getDocument();
				
				Scanner scanner = null;
				try {
					scanner = new Scanner(new File(doc));
				} catch(FileNotFoundException e) {
					return "(Documentation file not found)";
				}
				
				String ret = "";
				for(int i = 0; i < PREVIEW_LENGTH && scanner.hasNext(); i++) {
					ret += scanner.next() + " ";
				}
				scanner.close();
				return ret;
			}
		});
		description.getColumn().addSelectionListener(new TraceabilitySelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				viewer.setComparator(descComparator);
				recreateControls();
			}
			
		});
		
		/* REMOVE LINK */
		TableViewerColumn remove = generateColumn("Remove Link");
		remove.setLabelProvider(new ColumnLabelProvider() {
			
			@Override
			public String getText(Object element) {
				return "";
			}
		});
		
		/* SIMILARITY COLUMN */
		TableViewerColumn sim = generateColumn("Similarity");
		sim.setLabelProvider(new ColumnLabelProvider() {
			
			@Override
			public String getText(Object element) {
				TraceabilityLink link = (TraceabilityLink)element;
				
				if(link.getSimilarity() == TraceabilityLink.USER_SPECIFIED) {
					return "User-specified";
				} else {
					return String.format("%1.2f", link.getSimilarity());
				}
			}
		});
		sim.getColumn().addSelectionListener(new TraceabilitySelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				viewer.setComparator(simComparator);
				recreateControls();
			}
			
		});
		
		sim.getViewer().setComparator(simComparator);
	}
	
	/**
	 * Private helper method for building individual columns.
	 * @param title column title
	 * @return generated column
	 */
	private TableViewerColumn generateColumn(String title) {
		TableViewerColumn column = new TableViewerColumn(viewer, SWT.NONE);
		TableColumn inColumn = column.getColumn();
		inColumn.setText(title);
		inColumn.setResizable(true);
		inColumn.setMoveable(true);
		inColumn.setWidth(150);
		
		return column;
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.part.WorkbenchPart#setFocus()
	 */
	@Override
	public void setFocus() {
		viewer.getControl().setFocus();
	}
	
	/**
	 * Sets the input for the viewer.  After doing so, this method generates appropriate controls for each 
	 * row in the table (remove button, source/documentation links).
	 * 
	 * @param input input for the viewer
	 * @see traceability.views.TraceabilityContentProvider#getElements(Object)
	 */
	public void setInput(Object input) {
		viewer.setInput(input);
		
		recreateControls();
	}
	
	private void recreateControls() {
		
		/* Dispose of the old widgets */
		for(Button b : removeButtons) {
			b.dispose();
		}
		removeButtons.clear();
		for(Link l : links) {
			l.dispose();
		}
		links.clear();
		for(TableEditor e : editors) {
			e.dispose();
		}
		editors.clear();
		
		/* Build the new ones */
		Table table = viewer.getTable();
		
		int minSourceWidth = 0;
		int minDocWidth = 0;
		for(TableItem t : table.getItems()) {
			/* REMOVE BUTTONS */
			TableEditor removeEditor = new TableEditor(table);
			Button remove = new Button(table, SWT.NONE);
			remove.setText("Remove");
			remove.setData(t.getData());
			remove.addSelectionListener(new TraceabilitySelectionListener() {

				@Override
				public void widgetSelected(SelectionEvent e) {
					try {
						Button source = (Button)e.getSource();
						TraceabilityLink link = (TraceabilityLink)source.getData();
						Manager manager = Manager.getInstance();
						
						//Successful removal means we should update the view
						if(manager.removeLink(link)) {
							setInput(viewer.getInput());
						}
					} catch (Exception e2) {
						e2.printStackTrace();
					} catch (Error e2) {
						e2.printStackTrace();
					}
				}
				
			});
			remove.pack();
			removeEditor.minimumWidth = remove.getSize().x;
			removeEditor.minimumHeight = remove.getSize().y - REMOVE_VERT_OFFSET;
			removeEditor.horizontalAlignment = SWT.CENTER;
			removeEditor.setEditor(remove, t, REMOVE_COLUMN);
			removeButtons.add(remove);
			editors.add(removeEditor);
			
			TraceabilityLink link = (TraceabilityLink)t.getData();
		
			/* SOURCE LINKS */
			TableEditor linkEditor = new TableEditor(table);
			Link sourceLink = new Link(table, SWT.NONE);
			sourceLink.setBackground(table.getBackground());
			if(link.getSourceGranularity() == Granularity.Package) {
				sourceLink.setText(link.getSource());
			} else {
				sourceLink.setText("<A HREF=\"" + link.getSource() + "\">" + 
						link.getSource() + 
						"</A>");
			}
			sourceLink.setData(link);
			sourceLink.addSelectionListener(SourceLinkListener.getInstance());
			
			sourceLink.pack();
			linkEditor.minimumWidth = sourceLink.getSize().x;
			linkEditor.minimumHeight = sourceLink.getSize().y;
			linkEditor.horizontalAlignment = SWT.LEFT;
			linkEditor.setEditor(sourceLink, t, SOURCE_COLUMN);
			links.add(sourceLink);
			editors.add(linkEditor);
			if(minSourceWidth < sourceLink.getSize().x) {
				minSourceWidth = sourceLink.getSize().x + LINK_OFFSET;
			}
			
			/* DOCUMENTATION LINKS */
			TableEditor docEditor = new TableEditor(table);
			Link docLink = new Link(table, SWT.NONE);
			docLink.setBackground(table.getBackground());
			docLink.setText("<A HREF=\"" + link.getDocument() + "\">" + 
					link.getDocument() + 
					"</A>");
			docLink.setData(link);
			docLink.addSelectionListener(new TraceabilitySelectionListener() {

				@SuppressWarnings("deprecation")
				@Override
				public void widgetSelected(SelectionEvent e) {
					//TODO handle documentation not in the workspace
					IFile [] files = null;
					IWorkbenchPage page = getSite().getPage();
					Link source = (Link)e.getSource();
					TraceabilityLink link = (TraceabilityLink)source.getData();
					IPath path = Path.fromPortableString(link.getDocument());
					IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
					files = root.findFilesForLocation(path);	//TODO avoid using deprecated method here
					if(files.length != 0) {
						try {
							IDE.openEditor(page, files[0]);
						} catch (PartInitException exc) {
							MessageDialog.openError(viewer.getTable().getShell(), 
									"Error Opening File", 
									"Could not open file for editing.");
						} catch (Exception exc) {
							exc.printStackTrace();
						} catch (Error err) {
							err.printStackTrace();
						}
					}
					
				}
			});
			
			docLink.pack();
			docEditor.minimumWidth = docLink.getSize().x;
			docEditor.minimumHeight = docLink.getSize().y;
			docEditor.horizontalAlignment = SWT.LEFT;
			docEditor.setEditor(docLink, t, DOCUMENTATION_COLUMN);
			links.add(docLink);
			editors.add(docEditor);
			if(minDocWidth < docLink.getSize().x) {
				minDocWidth = docLink.getSize().x + LINK_OFFSET;
			}
		}
		sourceColumn.setWidth(minSourceWidth);
		docColumn.setWidth(minDocWidth);
	}
	
	/**
	 * Retrieves the instance of this view.
	 * @return single instance, or null if not yet instantiated
	 */
	public static TraceabilityView getInstance() {
		return instance;
	}
}