package traceability.wizard;

import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Set;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.wizard.Wizard;

import traceability.links.TraceabilityLink;
import traceability.management.Manager;
import traceability.management.ManagerException;
import traceability.recovery.DocumentationPreprocessor;
import traceability.recovery.Preprocessor;
import traceability.recovery.RecoveryTechnique;
import traceability.recovery.RecoveryTechniqueParams;
import traceability.recovery.StemmerSetting;

/**
 * Defines a wizard interface through which (retrospective) traceability links recovery techniques 
 * may be applied.
 * 
 * @author Sam Klock
 *
 */
public class LinkRecoveryWizard extends Wizard {
	
	/* Wizard pages */
	private GranularityWizardPage granularityPage;
	private PreprocessingWizardPage preprocessingPage;
	private TechniqueWizardPage techniquePage;
	private DocumentationWizardPage docPage;
	private ResultsWizardPage resultsPage;
	
	/**
	 * Results of link generation.
	 */
	private Set<TraceabilityLink> results;
	
	/**
	 * Constructor.
	 */
	public LinkRecoveryWizard() {
		try{
			setWindowTitle("Traceability Links Recovery");
			granularityPage = new GranularityWizardPage();
			preprocessingPage = new PreprocessingWizardPage();
			techniquePage = new TechniqueWizardPage();
			docPage = new DocumentationWizardPage();
			resultsPage = new ResultsWizardPage();
		} catch (Exception e) {
			e.printStackTrace();
		} catch (Error e) {
			e.printStackTrace();
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.eclipse.jface.wizard.Wizard#addPages()
	 */
	@Override
	public void addPages() {
		addPage(granularityPage);
		addPage(preprocessingPage);
		addPage(techniquePage);
		addPage(docPage);
		addPage(resultsPage);
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.eclipse.jface.wizard.Wizard#performFinish()
	 */
	@Override
	public boolean performFinish() {
		Manager manager = Manager.getInstance();
		Preprocessor preprocessor = null;
		List<String> errors;
		if(manager.getDocumentationLocation().equals("")) {
			try {
				manager.setDocumentationLocation(docPage.getDocumentationLocation());
			} catch (ManagerException e) {
				e.printStackTrace();
			}
		}
		
		//Preprocess the project
		try {
			//Initialize...
			preprocessor = preprocessingPage.getPreprocessor();
			preprocessor.setGranularity(granularityPage.getSelectedGranularity());
			preprocessor.setProject(manager.getProject());
			preprocessor.setRemoveNonLiterals(preprocessingPage.removeNonLiterals());
			preprocessor.setSplitIdentifiers(preprocessingPage.splitIdentifiers());
			if(preprocessingPage.stemWords()) {
				if(preprocessingPage.stemAllWords()) {
					preprocessor.setStemming(StemmerSetting.StemAll);
				} else {
					preprocessor.setStemming(StemmerSetting.StemNouns);
				}
			} else {
				preprocessor.setStemming(StemmerSetting.NoStemming);
			}
			preprocessor.setStopWordsFile(preprocessingPage.getStopWordsFile());
			
			//... then preprocess
			if(!preprocessor.preprocess()) {
				//Preprocessor failure: report to the user and drop out early
				errors = preprocessor.getErrors();
				if(errors.isEmpty()) {
					MessageDialog.openError(getShell(), "Error Source During Preprocessing", 
							"Source preprocessing was not successful.");
				} else {
					String message = "";
					for(String s : errors) {
						message += s + "\n";
					}
					MessageDialog.openError(getShell(), "Error During Source Preprocessing", 
							"The following error(s) occurred during Source preprocessing:\n" + message);
				}
				return true;
			}
		} catch (Exception e) {
			//TODO improve exception handling
			e.printStackTrace();
			MessageDialog.openError(getShell(), "Error During Preprocessing", e.getMessage());
			return true;
		}
		
		//Check for non-critical errors from the project preprocessor.
		//If there are any, notify the user (but continue).
		errors = preprocessor.getErrors();
		if(errors.size() > 0) {
			String message = "";
			for(String s : errors) {
				message += s + "\n";
			}
			MessageDialog.openError(getShell(), "Error During Source Preprocessing", 
					"The source preprocessor completed successfully, but it emitted the following warnings:\n" 
						+ message);
		}
		
		//Next: preprocess documentation.  (The documentation preprocessor is hard-coded.)
		//Initialize...
		DocumentationPreprocessor docPreprocessor = new DocumentationPreprocessor();
		docPreprocessor.setProject(manager.getProject());
		docPreprocessor.setRemoveNonLiterals(preprocessingPage.removeNonLiterals());
		docPreprocessor.setSplitIdentifiers(preprocessingPage.splitIdentifiers());
		if(preprocessingPage.stemWords()) {
			if(preprocessingPage.stemAllWords()) {
				docPreprocessor.setStemming(StemmerSetting.StemAll);
			} else {
				docPreprocessor.setStemming(StemmerSetting.StemNouns);
			}
		} else {
			docPreprocessor.setStemming(StemmerSetting.NoStemming);
		}
		docPreprocessor.setDocumentationLocation(docPage.getDocumentationLocation());
		
		//... then preprocess.
		if(!docPreprocessor.preprocess()) {
			//Upon failure, notify the user (as with the project preprocessor)
			errors = docPreprocessor.getErrors();
			if(errors.isEmpty()) {
				MessageDialog.openError(getShell(), "Error During Documentation Preprocessing", 
						"Documentation preprocessing was not successful.");
			} else {
				String message = "";
				for(String s : errors) {
					message += s + "\n";
				}
				MessageDialog.openError(getShell(), "Error During Documentation Preprocessing", 
						"The following error(s) occurred during documentation preprocessing:\n" + message);
			}
			return true;
		}
		
		//As with the source preprocessor: check for non-critical errors
		if(errors.size() > 0) {
			String message = "";
			for(String s : errors) {
				message += s + "\n";
			}
			MessageDialog.openError(getShell(), "Error During Documentation Preprocessing", 
					"The documentation preprocessor completed successfully, " +
					"but it emitted the following warnings:\n" + message);
		}
		
		//Next: run the selected technique.
		//Initialize...
		RecoveryTechnique technique = techniquePage.getSelectedTechnique();
		RecoveryTechniqueParams params = technique.getParameters();
		params.setGranularity(granularityPage.getSelectedGranularity());
		params.setMinimumSimilarity(resultsPage.getSimilarityThreshold());
		params.setThreshold(resultsPage.getAbsoluteThreshold());
		params.setWhetherThreshold(resultsPage.getWhetherAbsoluteThreshold());
		
		//... then run
		if(!technique.run(manager.getProject())) {
			//Notify the user of an exception upon failure and back out.
			MessageDialog.openError(getShell(), "Error During Recovery", "Traceability links recovery was not successful.");
			return true;
		} 
		
		//Retrieve the recovered links from the technique.
		results = technique.getRecoveredLinks();
		TraceabilityLink [] links = new TraceabilityLink[results.size()];
		int i = 0;
		for(TraceabilityLink l : results) {
			links[i++] = l;
		}
		
		//Sort the links by similarity.
		Arrays.sort(links, new Comparator<TraceabilityLink>() {
			@Override
			public int compare(TraceabilityLink arg0, TraceabilityLink arg1) {
				//Note that we want a sort in descending order
				if(arg0.getSimilarity() < arg1.getSimilarity()) {
					return 1;
				} else if(arg0.getSimilarity() > arg1.getSimilarity()) {
					return -1;
				}
				return 0;
			}
			
		});
		
		//Drop the links into the set of results.
		results.clear();
		for(i = 0; i < links.length; i++) {
			results.add(links[i]);
			manager.addLink(links[i]);
		}
		
		//Save the results; the user may remove incorrect candidate links.
		manager.setGenerationResults(results);
		try {
			manager.saveSet();
		} catch (ManagerException e) {
			e.printStackTrace();
		}
		
		return true;
	}
	
	/**
	 * Retrieves the most recent results of traceability links recovery.
	 * @return most recent results, or null if recovery has not been performed
	 */
	public Set<TraceabilityLink> getResults() {
		return results;
	}

	//TODO implement canFinish()
}
