package traceability.wizard;

import java.util.Vector;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;

import traceability.recovery.Preprocessor;
import traceability.util.TraceabilitySelectionListener;

/**
 * 
 * Wizard page used to select options for corpus preprocessing.
 * 
 * @author Sam Klock
 *
 */
public class PreprocessingWizardPage extends WizardPage {

	private static final int FILE_TEXT_WIDTH = 200;
	private static final String EXTENSION_POINT_ID = "traceability.preprocessor";
	private static final String PREPROCESSOR_ID = "impl";
	private static final int DEFAULT_SELECTED_INDEX = 0;
	
	private enum PreprocessingOption {
		NonLiterals,
		SplitIdentifiers,
		StemWords,
		StemAllWords,
		StemNounsOnly,
		StopWords;
	}

	/* These store user-selected options. */
	private boolean nonLiterals = true;
	private boolean splitIDs    = true;
	private boolean stemming    = true;
	private boolean allWords    = true;
	private boolean stopWords   = true;
	
	/* These store the user-selected preprocessor and those available, respectively. */
	private Preprocessor selected;
	private Preprocessor [] preprocessors;

	/* These store widgets used when handling events. */
	private Text fileText;
	private Button allWordsButton;
	private Button nounsOnlyButton;
	private Composite composite;
	
	/**
	 * Constructor.
	 */
	public PreprocessingWizardPage() {
		super("Corpus Preprocessing");
		setDescription("Select parameters for preprocessing the corpus.");
	}
	
	public Preprocessor getPreprocessor() {
		return selected;
	}
	
	/**
	 * Returns whether the user decided to remove non-literals.
	 * @return user's decision
	 */
	public boolean removeNonLiterals() {
		return nonLiterals;
	}
	
	/**
	 * Returns whether the user decided to split identifiers.
	 * @return user's decision
	 */
	public boolean splitIdentifiers() {
		return splitIDs;
	}
	
	/**
	 * Returns whether the user decided to stem words.
	 * @return user's decision
	 */
	public boolean stemWords() {
		return stemming;
	}
	
	/**
	 * Returns whether the user decided to stem all words (if the user decided to stem words).  Falsity 
	 * implies the user wants to stem nouns only.
	 * @return user's decision
	 */
	public boolean stemAllWords() {
		return allWords;
	}
	
	/**
	 * Returns whether the user decided to remove stop words.
	 * @return user's decision
	 */
	public boolean removeStopWords() {
		return stopWords;
	}
	
	/**
	 * Returns the user-selected stop words file.
	 * @return custom stop words file
	 */
	public String getStopWordsFile() {
		return fileText.getText();
	}
	
	/**
	 * Selection listener for check/radio buttons on this page.
	 */
	private TraceabilitySelectionListener buttonListener = new TraceabilitySelectionListener() {

		@Override
		public void widgetSelected(SelectionEvent e) {
			Button source = (Button)e.getSource();
			PreprocessingOption opt = (PreprocessingOption)source.getData();
			
			switch(opt) {
			case NonLiterals:
				nonLiterals = source.getSelection();
				break;
			case SplitIdentifiers:
				splitIDs = source.getSelection();
				break;
			case StemWords:
				stemming = source.getSelection();
				allWordsButton.setEnabled(stemming);
				nounsOnlyButton.setEnabled(stemming);
				break;
			case StemAllWords:
				allWords = source.getSelection();
				break;
			case StemNounsOnly:
				allWords = !source.getSelection();
				break;
			case StopWords:
				stopWords = source.getSelection();
				break;
			}
		}
		
	};
	
	/*
	 * (non-Javadoc)
	 * @see org.eclipse.jface.dialogs.IDialogPage#createControl(org.eclipse.swt.widgets.Composite)
	 */
	@Override
	public void createControl(Composite parent) {
		composite = new Composite(parent, SWT.NONE);
		GridLayout layout = new GridLayout();
		layout.numColumns = 1;
		composite.setLayout(layout);
		
		Composite childComp = new Composite(composite, SWT.BORDER);
		childComp.setLayout(new GridLayout());
		addPreprocessors(childComp);
		
		composite = new Composite(composite, SWT.BORDER);
		layout = new GridLayout();
		layout.numColumns = 3;
		composite.setLayout(layout);
		
		Label title = new Label(composite, SWT.NONE);
		title.setText("Select Parameters:");
		generateEmptyLabels(composite);

		Button litButton = new Button(composite, SWT.CHECK);
		litButton.setText("Remove non-literals");
		litButton.setData(PreprocessingOption.NonLiterals);
		litButton.setSelection(nonLiterals);
		litButton.addSelectionListener(buttonListener);
		generateEmptyLabels(composite);
		
		Button splitIDButton = new Button(composite, SWT.CHECK);
		splitIDButton.setText("Split identifier names");
		splitIDButton.setData(PreprocessingOption.SplitIdentifiers);
		splitIDButton.setSelection(splitIDs);
		splitIDButton.addSelectionListener(buttonListener);
		generateEmptyLabels(composite);
		
		Button stemButton = new Button(composite, SWT.CHECK);
		stemButton.setText("Word stemming:");
		stemButton.setData(PreprocessingOption.StemWords);
		stemButton.setSelection(stemming);
		stemButton.addSelectionListener(buttonListener);
		generateEmptyLabels(composite);
		
		allWordsButton = new Button(composite, SWT.RADIO);
		allWordsButton.setText("\tAll words");
		allWordsButton.setData(PreprocessingOption.StemAllWords);
		allWordsButton.setSelection(allWords);
		allWordsButton.addSelectionListener(buttonListener);
		generateEmptyLabels(composite);
		
		nounsOnlyButton = new Button(composite, SWT.RADIO);
		nounsOnlyButton.setText("\tNouns only");
		nounsOnlyButton.setData(PreprocessingOption.StemNounsOnly);
		nounsOnlyButton.setSelection(!allWords);
		nounsOnlyButton.addSelectionListener(buttonListener);
		generateEmptyLabels(composite);
		
		Button stopWordsButton = new Button(composite, SWT.CHECK);
		stopWordsButton.setText("Remove stop-words");
		stopWordsButton.setData(PreprocessingOption.StopWords);
		stopWordsButton.setSelection(stopWords);
		stopWordsButton.addSelectionListener(buttonListener);
		generateEmptyLabels(composite);
		
		Label locationLabel = new Label(composite, SWT.NONE);
		locationLabel.setText("Custom stop-word list:");
		fileText = new Text(composite, SWT.SINGLE | SWT.BORDER);
		GridData bounds = new GridData();
		bounds.minimumWidth = FILE_TEXT_WIDTH;
		bounds.widthHint = FILE_TEXT_WIDTH;
		fileText.setLayoutData(bounds);
		Button browse = new Button(composite, SWT.PUSH);
		browse.setText("Browse...");
		browse.addSelectionListener(new TraceabilitySelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				FileDialog fd = new FileDialog(composite.getShell());
				fileText.setText(fd.open());
			}
			
		});
		
		setControl(composite.getParent());
	}
	
	/*
	 * Private helper method for layout
	 */
	private static void generateEmptyLabels(Composite composite) {
		Label label = new Label(composite, SWT.NONE);
		label.setText("");
		label = new Label(composite, SWT.NONE);
		label.setText("");
	}
	
	/**
	 * Selection listener for techniques.
	 */
	private TraceabilitySelectionListener preprocessorListener = new TraceabilitySelectionListener() {

		@Override
		public void widgetSelected(SelectionEvent e) {
			Button source = (Button)e.getSource();
			Preprocessor preprocessor = (Preprocessor)source.getData();
			
			selected = preprocessor;
		}
	};
	
	private void addPreprocessors(Composite composite) {
		IConfigurationElement [] config = Platform.getExtensionRegistry()
							.getConfigurationElementsFor(EXTENSION_POINT_ID);

		Vector<Preprocessor> preprocs = new Vector<Preprocessor>();
		for(IConfigurationElement e : config) {
			Preprocessor preproc = null;
			Object elem = null;
			try {
				elem = e.createExecutableExtension(PREPROCESSOR_ID);
			} catch (CoreException e1) {
				System.err.println("WARNING: error loading technique extension");
				e1.printStackTrace();
				continue;
			}
			
			if(elem instanceof Preprocessor) {
				preproc= (Preprocessor)elem;
			} else {
				continue;
			}
			
			preprocs.add(preproc);
		}
		
		preprocessors = preprocs.toArray(new Preprocessor[0]);
		selected = preprocessors[DEFAULT_SELECTED_INDEX];
		
		Label title = new Label(composite, SWT.NONE);
		title.setText("Select Preprocessor:");
		
		for(Preprocessor p : preprocessors) {
			Button radioButton = new Button(composite, SWT.RADIO);
			radioButton.setText(p.id());
			radioButton.setSelection(p.equals(selected));
			radioButton.addSelectionListener(preprocessorListener);
			radioButton.setData(p);
		}
	}

}
