package traceability.wizard;

import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.VerifyEvent;
import org.eclipse.swt.events.VerifyListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;

import traceability.util.TraceabilitySelectionListener;

/**
 * 
 * Wizard page enabling the user to select criteria for proposed links (either a specified number 
 * of links or links with similarity above a given threshold).
 * 
 * @author Sam Klock
 *
 */
public class ResultsWizardPage extends WizardPage {

	/* Default values */
	private static final int TEXT_WIDTH = 100;
	private static final int DEFAULT_K = 100;
	private static final double DEFAULT_SIM = 0.8;
	private static final boolean TOP_K_SELECTED = true;
	
	/* Used to identify widgets in listeners */
	private static final char TOP_K_ID = 'K';
	private static final char SIM_ID = 'S';
	
	/* Widgets used in listeners */
	private Button absoluteThreshold;
	private Text topK;
	private Text simThreshold;
	
	/**
	 * Constructor.
	 */
	public ResultsWizardPage() {
		super("Select Thresholds for Results");
		setDescription("Select criteria for proposed traceability links.");
	}
	
	/**
	 * Gets whether the user decided to use an absolute threshold.
	 * @return true if absolute threshold, false if similarity threshold
	 */
	public boolean getWhetherAbsoluteThreshold() {
		return absoluteThreshold.getSelection();
	}
	
	/**
	 * Gets the user-specified number of links to return.
	 * @return the number of links, or -1 if invalid
	 */
	public int getAbsoluteThreshold() {
		try {
			return Integer.parseInt(topK.getText());
		} catch (NumberFormatException e) {
			return -1;
		}
	}
	
	/**
	 * Gets the user-specified similarity threshold.
	 * @return the threshold, or -1.0 if invalid
	 */
	public double getSimilarityThreshold() {
		try {
			return Double.parseDouble(simThreshold.getText());
		} catch(NumberFormatException e) {
			return -1.0;
		}
	}
	
	/**
	 * Verification listener for the number inputs.
	 */
	private VerifyListener vListener = new VerifyListener() {

		@Override
		public void verifyText(VerifyEvent e) {
			if(e.text.equals("")) {
				e.doit = true;
				return;
			}
			Text source = (Text)e.getSource();
			
			if(source.getData().equals(new Character(TOP_K_ID))) {
				if(e.character >= '0' && e.character <= '9') {
					e.doit = true;
				} else if(e.character != SWT.BS && e.character != SWT.DEL) {
					e.doit = false;
				}
				return;
			}
			
			try {
				String currentText = simThreshold.getText();
				String start = currentText.substring(0, e.start);
				String finish = currentText.substring(e.end);
				String full = start + e.text + finish;
				double real = Double.parseDouble(full);
				if(real < 0.0 || real > 1.0) {
					e.doit = false;
				}
			} catch (NumberFormatException exc) {
				e.doit = false;
			}
		}
	};
	
	/**
	 * Selection listener for the radio buttons.
	 */
	private TraceabilitySelectionListener sListener = new TraceabilitySelectionListener() {

		@Override
		public void widgetSelected(SelectionEvent e) {
			Button source = (Button)e.getSource();
			if(source.getData().equals(SIM_ID) && source.getSelection()) {
				simThreshold.setEnabled(true);
				topK.setEnabled(false);
			} else {
				simThreshold.setEnabled(false);
				topK.setEnabled(true);
			}
		}
		
	};
	
	/*
	 * (non-Javadoc)
	 * @see org.eclipse.jface.dialogs.IDialogPage#createControl(org.eclipse.swt.widgets.Composite)
	 */
	@Override
	public void createControl(Composite parent) {
		Composite composite = new Composite(parent, SWT.NONE);
		composite.setLayout(new GridLayout(3, true));
		
		/* Top k/absolute threshold */
		absoluteThreshold = new Button(composite, SWT.RADIO);
		absoluteThreshold.setText("Top k links");
		absoluteThreshold.setData(TOP_K_ID);
		absoluteThreshold.addSelectionListener(sListener);
		absoluteThreshold.setSelection(TOP_K_SELECTED);
		Label label = new Label(composite, SWT.NONE);
		label.setText("Enter k:");
		topK = new Text(composite, SWT.BORDER);
		topK.setText(Integer.toString(DEFAULT_K));
		topK.setData(TOP_K_ID);
		topK.addVerifyListener(vListener);
		topK.setEnabled(TOP_K_SELECTED);
		GridData data = new GridData();
		data.widthHint = TEXT_WIDTH;
		topK.setLayoutData(data);
		
		/* Similarity threshold */
		Button simThresholdButton = new Button(composite, SWT.RADIO);
		simThresholdButton.setText("Similarity threshold");
		simThresholdButton.setData(SIM_ID);
		simThresholdButton.addSelectionListener(sListener);
		simThresholdButton.setSelection(!TOP_K_SELECTED);
		label = new Label(composite, SWT.NONE);
		label.setText("Enter similarity:");
		simThreshold = new Text(composite, SWT.BORDER);
		simThreshold.setText(Double.toString(DEFAULT_SIM));
		simThreshold.setData(SIM_ID);
		simThreshold.addVerifyListener(vListener);
		simThreshold.setEnabled(!TOP_K_SELECTED);
		data = new GridData();
		data.widthHint = TEXT_WIDTH;
		simThreshold.setLayoutData(data);
		
		setControl(composite);
	}

}
