package traceability.wizard;

import java.util.Hashtable;
import java.util.Map;
import java.util.Vector;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StackLayout;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.VerifyEvent;
import org.eclipse.swt.events.VerifyListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;

import traceability.recovery.ParameterType;
import traceability.recovery.RecoveryTechnique;
import traceability.recovery.RecoveryTechniqueParams;
import traceability.util.TraceabilitySelectionListener;

/**
 * 
 * Wizard page for selecting a recovery technique.
 * 
 * There are two main composites: on the left, the user may select a recovery technique; on the right, 
 * the user parameterizes the selected technique.
 * 
 * New techniques may be contributed by other plugins by using the traceability.technique extension point 
 * supplied by this plugin.
 * 
 * @author Sam Klock
 *
 */
public class TechniqueWizardPage extends WizardPage {

	private static final String EXTENSION_POINT_ID = "traceability.technique";
	
	private static final String TECHNIQUE_ID = "technique";
	
	/**
	 * Index of default-selected technique.
	 */
	private static final int DEFAULT_TECHNIQUE_INDEX = 0;
	/**
	 * Default width for text boxes.
	 */
	private static final int TEXT_WIDTH = 100;
	
	/**
	 * Maps recovery techniques to composites for their parameters.
	 */
	private Map<RecoveryTechnique, Composite> techniqueCompositeMap;
	/**
	 * The selected recovery technique.
	 */
	private RecoveryTechnique selected = null;
	/**
	 * Stack layout for parameter composites.
	 */
	private StackLayout stack;
	/**
	 * Composite containing parameter composites; laid out by stack.
	 */
	private Composite parameterCompositeChild;
	
	private RecoveryTechnique [] techniques = null;
	
	/**
	 * Constructor.
	 */
	public TechniqueWizardPage() {
		super("Select Technique");
		setDescription("Select a traceability links recovery technique and parameterize it.");
	}
	
	/**
	 * Retrieves the user-selected, parameterized traceability links recovery technique.
	 * @return the technique
	 */
	public RecoveryTechnique getSelectedTechnique() {
		return selected;
	}
	
	/**
	 * Selection listener for techniques.
	 */
	private TraceabilitySelectionListener techniqueListener = new TraceabilitySelectionListener() {

		@Override
		public void widgetSelected(SelectionEvent e) {
			Button source = (Button)e.getSource();
			RecoveryTechnique technique = (RecoveryTechnique)source.getData();
			
			selected = technique;
			stack.topControl = techniqueCompositeMap.get(selected);
			parameterCompositeChild.layout();
		}
		
	};
	
	/**
	 * Modification listener for text boxes.
	 */
	private ModifyListener valueChangedListener = new ModifyListener() {

		@Override
		public void modifyText(ModifyEvent e) {
			Text source = (Text)e.getSource();
			String id = (String)source.getData();
			
			selected.getParameters().setValueForParameter(id, source.getText());
		}
	};
	
	/**
	 * Verification listener for text boxes.
	 */
	private VerifyListener verifyListener = new VerifyListener() {

		@Override
		public void verifyText(VerifyEvent e) {
			Text source = (Text)e.getSource();
			char newChar = e.character;
			ParameterType type = selected.getParameters().getTypeForParameter((String)source.getData());
			
			/* Always accept backspace/delete */
			if(newChar == SWT.BS || newChar == SWT.DEL) {
				e.doit = true;
				return;
			}
			
			/* Otherwise, whether we accept depends on type */
			switch(type) {
			case Integer:
				e.doit = true;
				if(newChar < '0' || newChar > '9') {		//Must be digit...
					if(newChar != '-' || e.start != 0) { 	//... or must be negative sign
						e.doit = false;
					}
				}
				break;
			case Real:
				e.doit = false;
				if(newChar == '.' && source.getText().indexOf(newChar) == -1) {			//One decimal is okay
					e.doit = true;
				} else if(newChar >= '0' && newChar <= '9') {							//Digits always okay
					e.doit = true;
				} else if(newChar == '-' && source.getText().indexOf(newChar) == -1 &&	//Negative sign is okay
						e.start == 0) {													//if properly placed
					e.doit = true;
				}
				break;
			default:
				e.doit = true;
				break;
			}
		}
		
	};
	
	/**
	 * Selection listener for checkboxes and radio buttons in parameter composites.
	 */
	private TraceabilitySelectionListener parameterButtonListener = new TraceabilitySelectionListener() {

		@Override
		public void widgetSelected(SelectionEvent e) {
			Button source = (Button)e.getSource();
			String id = (String)source.getData();
			
			RecoveryTechniqueParams params = selected.getParameters();
			
			params.setValueForParameter(id, source.getSelection());
			
			if(params.getTypeForParameter(id) == ParameterType.Radio) {
				for(String p : params.getParameterIDs()) {
					if(params.getTypeForParameter(p) == ParameterType.Radio &&
							!p.equals(id)) {
						params.setValueForParameter(p, !source.getSelection());
					}
				}
			}
		}
		
	};
	
	/*
	 * (non-Javadoc)
	 * @see org.eclipse.jface.dialogs.IDialogPage#createControl(org.eclipse.swt.widgets.Composite)
	 */
	@Override
	public void createControl(Composite parent) {
		Composite composite = new Composite(parent, SWT.NONE);
		GridLayout layout = new GridLayout(2, true);
		composite.setLayout(layout);
		initRecoveryTechniques();
		
		/* Composite for technique selection */
		Composite techniqueComposite = new Composite(composite, SWT.BORDER);
		techniqueComposite.setLayout(new GridLayout());
		GridData techniqueBounds = new GridData();
		techniqueBounds.verticalAlignment = SWT.FILL;
		techniqueBounds.horizontalAlignment = SWT.FILL;
		techniqueComposite.setLayoutData(techniqueBounds);
		Label techniqueCompositeTitle = new Label(techniqueComposite, SWT.NONE);
		techniqueCompositeTitle.setText("Select Technique:");
		
		/* Composite for parameter selection */
		Composite parameterComposite = new Composite(composite, SWT.BORDER);
		parameterComposite.setLayout(new GridLayout());
		GridData parameterBounds = new GridData();
		parameterBounds.verticalAlignment = SWT.FILL;
		parameterBounds.horizontalAlignment = SWT.FILL;
		parameterComposite.setLayoutData(parameterBounds);
		Label parameterCompositeTitle = new Label(parameterComposite, SWT.NONE);
		parameterCompositeTitle.setText("Select Parameters:");
		
		/* Child composite in which parameters are actually selected */
		stack = new StackLayout();
		techniqueCompositeMap = new Hashtable<RecoveryTechnique, Composite>();
		parameterCompositeChild = new Composite(parameterComposite, SWT.NONE);
		parameterCompositeChild.setLayout(stack);
		
		/* Populate the technique and parameter child composites */
		for(RecoveryTechnique t : techniques) {
			Button radioButton = new Button(techniqueComposite, SWT.RADIO);
			radioButton.setText(t.getName());
			radioButton.setSelection(t.equals(selected));
			radioButton.addSelectionListener(techniqueListener);
			radioButton.setData(t);
			generateCompositeFor(t);
		}
		
		stack.topControl = techniqueCompositeMap.get(selected);
		parameterCompositeChild.layout();
		
		setControl(composite);
	}
	
	private void initRecoveryTechniques() {
		IConfigurationElement [] config = Platform.getExtensionRegistry()
				.getConfigurationElementsFor(EXTENSION_POINT_ID);
		
		Vector<RecoveryTechnique> techniques = new Vector<RecoveryTechnique>();
		for(IConfigurationElement e : config) {
			RecoveryTechnique technique = null;
			Object elem = null;
			try {
				elem = e.createExecutableExtension(TECHNIQUE_ID);
			} catch (CoreException e1) {
				System.err.println("WARNING: error loading technique extension");
				e1.printStackTrace();
				continue;
			}
			
			if(elem instanceof RecoveryTechnique) {
				technique = (RecoveryTechnique)elem;
			} else {
				continue;
			}
			
			techniques.add(technique);
		}
		
		this.techniques = techniques.toArray(new RecoveryTechnique[0]);
		selected = this.techniques[DEFAULT_TECHNIQUE_INDEX];
	}
	
	/*
	 * Helper method that generates parameter composites for each technique.
	 */
	private void generateCompositeFor(RecoveryTechnique technique) {
		Composite child = new Composite(parameterCompositeChild, SWT.NONE);
		GridLayout layout = new GridLayout();
		layout.numColumns = 2;
		child.setLayout(layout);
		RecoveryTechniqueParams params = technique.getParameters();
		for(String id : params.getParameterIDs()) {
			switch(params.getTypeForParameter(id)) {
			case Text:
			case Integer:
			case Real:
				Label paramLabel = new Label(child, SWT.NONE);
				paramLabel.setText(params.getLabelForParameter(id));
				Text paramText = new Text(child, SWT.BORDER);
				paramText.setText(params.getParameterValue(id).toString());
				paramText.addModifyListener(valueChangedListener);
				paramText.setData(id);
				paramText.addVerifyListener(verifyListener);
				GridData bounds = new GridData();
				bounds.minimumWidth = TEXT_WIDTH;
				bounds.widthHint = TEXT_WIDTH;
				paramText.setLayoutData(bounds);
				break;
			case Check:
				Button checkButton = new Button(child, SWT.CHECK);
				checkButton.setText(params.getLabelForParameter(id));
				checkButton.setData(id);
				checkButton.addSelectionListener(parameterButtonListener);
				break;
			case Radio:
				Button radioButton = new Button(child, SWT.RADIO);
				radioButton.setText(params.getLabelForParameter(id));
				radioButton.setData(id);
				radioButton.addSelectionListener(parameterButtonListener);
				break;
			}
		}
		techniqueCompositeMap.put(technique, child);
	}

}
