package traceability.recovery;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Scanner;
import java.util.Set;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.util.Version;
import org.eclipse.core.resources.IProject;

import traceability.links.Granularity;
import traceability.links.TraceabilityLink;
import traceability.management.Manager;

/**
 * 
 * Proof-of-concept implementation of traceability links recovery using Lucene.
 * 
 * @author Sam Klock
 * @since 0.1.0
 * @see RecoveryTechnique
 * @see RecoveryTechniqueParams
 */
public class LuceneRecoveryTechnique implements RecoveryTechnique {

	//TODO refactor getDirectoryFor...() methods to use static methods given by Preprocessor
	
	/**
	 * Name for this technique.
	 */
	private static final String NAME = "Lucene";
	
	/**
	 * Limit on search results to obtain from Lucene for any query. 
	 */
	private static final int UPPER_LIMIT = 300; 	//May be smaller than user-specified k, but that's okay
	
	/**
	 * Directory (relative to [project_home]/.trace) where Lucene's index is stored.
	 */
	public static final String INDEX_DIRECTORY = "lucene_index";
	
	/**
	 * Parameters for this technique.
	 */
	private LuceneRecoveryTechniqueParams parameters;
	
	/**
	 * Set of links produced via Lucene.
	 */
	private Set<TraceabilityLink> links;
	
	/**
	 * Project for which to recover links.
	 */
	private IProject project;
	
	/**
	 * Constructor.
	 */
	public LuceneRecoveryTechnique() {
		parameters = new LuceneRecoveryTechniqueParams();
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#getParameters()
	 */
	@Override
	public RecoveryTechniqueParams getParameters() {
		return parameters;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#getRecoveredLinks()
	 */
	@Override
	public Set<TraceabilityLink> getRecoveredLinks() {
		return links;
	}

	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#run(org.eclipse.core.resources.IProject)
	 */
	@Override
	public boolean run(IProject project) {
		this.project = project;
		Analyzer analyzer = new StandardAnalyzer(Version.LUCENE_30);
		Directory dir = null;
		File indexDir = new File(getDirectoryForIndex());
		links = new HashSet<TraceabilityLink>();
		
		
		if(!indexDir.isDirectory() && !indexDir.mkdir()) {
			return false;
		}
		try {
			dir = FSDirectory.open(indexDir);
		} catch (IOException e) {
			//TODO improving handling
			e.printStackTrace();
			return false;
		}
		int maximumFieldLength = (Integer)parameters.getParameterValue(parameters.getParameterIDs()[0]);
		IndexWriter iWriter = null;
		try {
			iWriter = new IndexWriter(dir, analyzer, 
					new IndexWriter.MaxFieldLength(maximumFieldLength));
			iWriter.deleteAll();
			
			File docDir = new File(getDirectoryForDocs());
			for(String s : docDir.list()) {
				//Ignore the mapper
				if(s.equals(Preprocessor.MAPPER_NAME)) {
					continue;
				}
				
				//Add document to index
				String filename = docDir.getAbsolutePath() + File.separator + s;
				File docFile = new File(filename);
				Document document = new Document();
				document.add(new Field("FILENAME", s, Field.Store.YES, Field.Index.NOT_ANALYZED));
				
				Reader reader = new FileReader(docFile);
				document.add(new Field("CONTENTS", reader));
				
				iWriter.addDocument(document);
			}
			iWriter.optimize();
			iWriter.close();
		} catch (Exception e) {
			//TODO improve
			e.printStackTrace();
			return false;
		}
		
		Scanner scanner;
		String text;
		Query query;
		try {
			IndexSearcher iSearcher = new IndexSearcher(dir);
			QueryParser queryParser = new QueryParser(Version.LUCENE_30, "CONTENTS", analyzer);
			
			File sourceDir = new File(getDirectoryForSource());
			for(String s : sourceDir.list()) {
				if(s.equals(Preprocessor.MAPPER_NAME)) {
					continue;
				}
				
				//String filename = docDir.getAbsolutePath() + File.separator + s;
				String filename = sourceDir.getAbsolutePath() + File.separator + s;
				scanner = new Scanner(new File(filename));
				text = "";
				while(scanner.hasNext()) {
					text += scanner.next() + " ";
				}
				scanner.close();
				try {
					query = queryParser.parse(text);
				} catch(ParseException e) {
					//Ignore the query
					//TODO do better than this
					continue;
				}
				
				TopDocs results = iSearcher.search(query, UPPER_LIMIT);
				String source;
				String documentation;
				Granularity gran = parameters.getGranularity();
				double minScore = parameters.getMinimumSimilarity();
				for(ScoreDoc scoreDoc : results.scoreDocs) {
					if(!parameters.getWhetherThreshold() && scoreDoc.score < minScore) {
						//Ignore results with too low a score
						continue;
					}
					Document doc = iSearcher.doc(scoreDoc.doc);
					documentation = getDocument(doc.get("FILENAME"));
					source = getSource(s);
					TraceabilityLink link = new TraceabilityLink(source, 
							documentation, scoreDoc.score, gran);
					links.add(link);	
				}
			}
		} catch (Exception e) {
			//TODO improve (again)
			e.printStackTrace();
			return false;
		}
		
		//Pare down the set of results to the number specified by the threshold, if applicable
		if(parameters.getWhetherThreshold()) {
			int threshold = parameters.getThreshold();
			TraceabilityLink [] theLinks = new TraceabilityLink[links.size()];
			int i = 0;
			for(TraceabilityLink l : links) {
				theLinks[i++] = l;
			}
			
			//Sort the links by similarity.
			Arrays.sort(theLinks, new Comparator<TraceabilityLink>() {
				@Override
				public int compare(TraceabilityLink arg0, TraceabilityLink arg1) {
					//Note that we want a sort in descending order
					if(arg0.getSimilarity() < arg1.getSimilarity()) {
						return 1;
					} else if(arg0.getSimilarity() > arg1.getSimilarity()) {
						return -1;
					}
					return 0;
				}
				
			});
			
			links.clear();
			for(int j = 0; j < threshold && j < theLinks.length; j++) {
				links.add(theLinks[j]);
			}
		}
		
		return true;
	}
	
	/*
	 * (non-Javadoc)
	 * @see traceability.recovery.RecoveryTechnique#getName()
	 */
	@Override
	public String getName() {
		return NAME;
	}
	
	private String getDirectoryForIndex() {
		Manager manager = Manager.getInstance();
		String loc = manager.getDirPath(project);
		
		return loc + File.separator + INDEX_DIRECTORY;
	}
	
	private String getDirectoryForSource() {
		Manager manager = Manager.getInstance();
		String loc = manager.getDirPath(project);
		
		return loc + File.separator + Preprocessor.SOURCE_DIR;
	}
	
	private String getDirectoryForDocs() {
		Manager manager = Manager.getInstance();
		String loc = manager.getDirPath(project);
		
		return loc + File.separator + Preprocessor.DOC_DIR;
	}

	private String getDocument(String file) {
		String docID = file.replace(Preprocessor.DOC_PREFIX, "");
		docID = docID.replace(Preprocessor.SUFFIX, "");
		int docIDnum = Integer.parseInt(docID);
		
		Scanner scanner = null;
		try {
			scanner = new Scanner(new File(getDirectoryForDocs() + File.separator + 
											Preprocessor.MAPPER_NAME));
		} catch(Exception e) {
			e.printStackTrace();
			return null;
		}
		
		String ret = null;
		for(int i = 0; i < docIDnum; i++) {
			ret = scanner.nextLine();
		}
		scanner.close();
		return ret;
	}
	
	private String getSource(String file) {
		String srcID = file.replace(Preprocessor.SOURCE_PREFIX, "");
		srcID = srcID.replace(Preprocessor.SUFFIX, "");
		int srcIDnum = Integer.parseInt(srcID);
		
		Scanner scanner = null;
		try {
			scanner = new Scanner(new File(getDirectoryForSource() + File.separator + 
											Preprocessor.MAPPER_NAME));
		} catch(Exception e) {
			e.printStackTrace();
			return null;
		}
		
		String ret = null;
		for(int i = 0; i < srcIDnum; i++) {
			if(!scanner.hasNextLine()) {
				break;
			}
			ret = scanner.nextLine();
		}
		scanner.close();
		return ret;
	}
}
